
/* t2_aio.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/26/97
 * Some typedefs for asynchronous I/O under T2.
 */

#ifndef _T2_AIO_H_
#define _T2_AIO_H_

/*
 * Each asynchronous I/O inteface must define the following:
 *  struct T2_LIOCB	-- this is the struct of the control blocks for
 *			   individual asynchornous I/O requests
 *  struct T2_LIOAux	-- a struct to store interface-dependent auxiliary
 *			   data (T2_LIO::initAux() is invoked by the
 *			   constructor of T2_LIO to initialize this struct,
 *			   and T2_LIO::deleteAux() is invoked by the
 *			   destructor of T2_LIO to delete any pointer used by
 *			   this struct)
 */


/**
 ** POSIX
 **/
#ifdef POSIX_AIO
	#include <aio.h>

	typedef aiocb		T2_LIOCB;
	typedef struct {}	T2_LIOAux;

	/*
	 * macro definition for the generic field names
	 */
	#define t2_liocb_opcode		aio_lio_opcode
	#define t2_liocb_fildes		aio_fildes
	#define t2_liocb_buf		aio_buf
	#define t2_liocb_nbytes		aio_nbytes
	#define t2_liocb_offset		aio_offset

#endif	/* POSIX_AIO */


/**
 ** AIX-version of POSIX
 **/
#ifdef AIX_AIO
	#include <fcntl.h>
	#include <aio.h>

	typedef liocb		T2_LIOCB;
	typedef struct {}	T2_LIOAux;

	/*
	 * macro definition for the generic field names
	 */
	#define t2_liocb_opcode		lio_opcode
	#define t2_liocb_fildes		lio_fildes
	#define t2_liocb_buf		lio_aiocb.aio_buf
	#define t2_liocb_nbytes		lio_aiocb.aio_nbytes
	#define t2_liocb_offset		lio_aiocb.aio_offset

#endif	/* AIX_AIO */


/**
 ** SUN, SunOS 5.5 (Solaris)
 **/
#ifdef SUNOS_AIO
	#include <sys/asynch.h>
	#include <sys/time.h>

	#ifndef LIO_NOP
		#define LIO_NOP		0
		#define LIO_READ	1
		#define LIO_WRITE	2
	#endif

	typedef struct {
			int	aio_whence;
			off_t	aio_offset;
			char*	aio_buf;
			size_t	aio_nbytes;
		} aiocb;

	typedef struct {
			int	lio_opcode;
			int	lio_fildes;
//			aiocb	lio_aiocb;
			aio_result_t	resultp;

			int	aio_whence;
			off_t	aio_offset;
			char*	aio_buf;
			size_t	aio_nbytes;
		} T2_LIOCB;

	typedef struct {
			int	back_diff;	/* const to store the diff
						   between &(T2_LIOCB.resultp)
						   and &T2_LIOCB */
		} T2_LIOAux;

	/*
	 * macro definition for the generic field names
	 */
	#define t2_liocb_opcode		lio_opcode
	#define t2_liocb_fildes		lio_fildes
	#define t2_liocb_buf		aio_buf
	#define t2_liocb_nbytes		aio_nbytes
	#define t2_liocb_offset		aio_offset

#endif	/* SUNOS_AIO */


#endif	/* !_T2_AIO_H_ */
