
/* $Id: t2_commfunc.h,v 1.5 2000/03/16 00:43:05 chialin Exp $ */

/* t2_commfunc.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 8/19/98
 * Some communication functions among the back-end.
 */


#ifndef _T2_COMMFUNC_H_
#define _T2_COMMFUNC_H_


#include	"t2_header.h"
#include	"t2_acomm.h"


void t2_be_gsync();			// synchronization


// reduction over a single number
u_int t2_be_gsum(u_int i);		// all sum reduction
int t2_be_gsum(int i);
long t2_be_gsum(long i);
u_long t2_be_gsum(u_long i);

// reduction over an array of numbers
void
t2_be_gsum(u_int* in, u_int* out, const u_int cnt),	// all sum reduction
t2_be_gsum(int* in, int* out, const u_int cnt),
t2_be_gsum(long* in, long* out, const u_int cnt),
t2_be_gsum(u_long* in, u_long* out, const u_int cnt),
t2_be_gmax(u_int* in, u_int* out, const u_int cnt),	// all max reduction
t2_be_gmax(int* in, int* out, const u_int cnt),
t2_be_gmax(long* in, long* out, const u_int cnt),
t2_be_gmax(u_long* in, u_long* out, const u_int cnt),
t2_be_gmin(u_int* in, u_int* out, const u_int cnt),	// all in reduction
t2_be_gmin(int* in, int* out, const u_int cnt),
t2_be_gmin(long* in, long* out, const u_int cnt),
t2_be_gmin(u_long* in, u_long* out, const u_int cnt);


// blocking sends and blocking recvs
bool t2_be_send(const char* buf, size_t nbytes, T2_ProcID dest, int tag),
     t2_be_send(const int* buf, size_t count, T2_ProcID dest, int tag),
     t2_be_send(const u_int* buf, size_t count, T2_ProcID dest, int tag),
     t2_be_send(const long* buf, size_t count, T2_ProcID dest, int tag),
     t2_be_send(const u_long* buf, size_t count, T2_ProcID dest, int tag),
     t2_be_recv(char* buf, size_t buflen, int& source, int& tag,
		size_t& nbytes),
     t2_be_recv(int* buf, size_t bufcount, int& source, int& tag,
		size_t& count),
     t2_be_recv(u_int* buf, size_t bufcount, int& source, int& tag,
		size_t& count),
     t2_be_recv(long* buf, size_t bufcount, int& source, int& tag,
		size_t& count),
     t2_be_recv(u_long* buf, size_t bufcount, int& source, int& tag,
		size_t& count);


// a blocking probe for src, tag and msgsize
bool t2_be_probe(int& src, int& tag, size_t& msgsize);

// some inline functions
template <class T>
inline bool
t2_be_send(const T& k, T2_ProcID dest, int tag)
{
  return t2_be_send(&k, 1, dest, tag);
}

template <class T>
inline bool
t2_be_recv(T& k, int& source, int& tag)
{
  size_t count;
  bool r = t2_be_recv(&k, 1, source, tag, count);
  if (count != 1)
     return false;
  return r;
}



#endif	/* ! _T2_COMMFUNC_H_ */
