
/* t2_ohmc.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 8/13/98
 * classes for returning the final output arrays through meta-chaos
 */


#ifndef _T2_OHMC_H_
#define _T2_OHMC_H_


#include	"t2_header.h"
#include	"mc_system.H"
#include	"mc_offbandcomm.H"
#include	"mc_schedule.H"
#include	"t2mc_user.H"


///////////////////////////////////////////////////////////////////////
//
// T2_MCHelloPacket
//
// This class is implemented just to enforce the meta-chaos oob protocol
// between the client and the back-ends.
//
///////////////////////////////////////////////////////////////////////
class T2_MCHelloPacket {
public:
  T2_MCHelloPacket() {};
  ~T2_MCHelloPacket() {};

  // meta-chaos oob broadcast
  bool bcast(MC_OffBandComm& offband, MC_Client& client,
	     const char* pname, int numnodes, int iters, int tag);

protected:
private:
};


inline bool
T2_MCHelloPacket::bcast(MC_OffBandComm& offband, MC_Client& client,
		        const char* pname, int numnodes, int iters, int tag)
{
  t2_Assert (pname != NULL);
  if (offband.bcast(client, (void *) pname, sizeof(char)*(strlen(pname)+1),
		    DATA_STRING, tag) == 0)
     return false;
  if (offband.bcast(client, (void *)&numnodes, sizeof(int), DATA_INT, tag) == 0)
     return false;
  if (offband.bcast(client, (void *)&iters, sizeof(int), DATA_INT, tag) == 0)
     return false;

  return true;
}



///////////////////////////////////////////////////////////////////////
//
// T2_OHMetaChaos
//
///////////////////////////////////////////////////////////////////////
class T2_OHMetaChaos: public T2_OutputHandle {
friend class T2_RTS;
public:
  T2_OHMetaChaos(T2_QEnv* p, T2_RTS& rts, T2_QSpecOutput& qspec);
  virtual ~T2_OHMetaChaos();

  // is the output-handling object blocking and therefore must wait
  // until all pending isends for the global-combine phase complete?
  // (eg blocking Meta-Chaos)
  bool isBlocking() const { return true; };

  // initialization at the beginning of query execution; returns true
  // if initialization succeeds, false if initialization fails and 
  // query should not be executed
  virtual bool initBeforeQueryExecution(T2_RTS& rts);

  // handle the output, returns true if the output handling phase is
  // needed, false if the output handling phase is not needed
  virtual bool handleOutput(const T2_System& system, T2_Output& output,
			    T2_Accumulator& acc, T2_Iteration i);

  // method invoked when a non-blocking write completes;
  // return true if another write is issued, false if all writes had been
  // issued
  virtual bool issueNextOutputWrite();

  // method invoked when the query is completely done
  virtual void finalize();

  // accessing MC_Server
  static MC_Server& getMCServer() { return mc_server; };

protected:
  MC_Client mc_client;			// meta-chaos client
  int ofbtag,				// msg tag for off-band communication
      datatag;				// msg tag for communication data arrays

  T2_MetaChaosUser* mcu;		// meta-chaos user object
  MC_Schedule* sched;			// meta-chaos schedule

  T2_OBuf obuf;
  T2_MCOutputDataPtrList ptrs;		// pointer list
  u_int cur_ptr;			// index for current pointer

  void cleanupAfterAllWritesDoneForIteration();
  bool issueOneWrite();

  static MC_Server mc_server;		// meta-chaos server
  static bool initMCServer(const char* mc_pname, int numnodes);
  static bool initMCServer(const char* mc_pname, int numnodes, int me);
  static void exitMCServer();

private:
};




#endif	/* _T2_OHMC_H_ */
