
/* $Id: comm.C,v 1.15 2000/12/02 02:10:50 chialin Exp $ */

/* comm.C */


#include	"t2_comm.h"
#include	"t2_commtags.h"


extern int _g_mynode;

//////////////////////////////////////////////////////////////////////
//
// include interface-dependent functions
//
//////////////////////////////////////////////////////////////////////
#ifdef T2_MPI_COMM
  #include	"comm_mpi.C"
#endif

//////////////////////////////////////////////////////////////////////
//
// T2_Comm methods
//
//////////////////////////////////////////////////////////////////////
T2_Comm::T2_Comm(u_int n): cbs(n), freel(n), pendingl(n), donel(n)
{
  for (u_int i=0; i<n; i++) {
    t2_Assert (t2_setSource(cbs[i], -1) == -1);
    t2_Assert (t2_setMsgTag(cbs[i], -1) == -1);
    freel.in(&(cbs[i]));
  }
}


T2_Comm::~T2_Comm()
{
}


T2_AsyncResult
T2_Comm::isend(char* buf, size_t nbytes, int dest, int tag, T2_CommRequest* key)
{
  t2_Assert (nbytes == 0 || buf != NULL);

  t2_AssertOrDo (!freel.isEmpty(), printStatus(cout));
  T2_CommCB* cb = freel.out();
  t2_Assert (cb != NULL);

  cb->key = key;
  T2_AsyncResult r = t2_issueISend(*cb, buf, nbytes, dest, tag);
  switch (r) {
     case aSUCCEED:
		break;

     case aFAIL_SYS:
		t2_err_sys(__FILE__, __LINE__,
			   "t2_issueISend returns aFAIL_SYS");
		return aFAIL_SYS;

     case aFAIL_USR:
		t2_err_ret(__FILE__, __LINE__,
			   "t2_issueISend returns aFAIL_USR");
		return aFAIL_USR;

     default:
		t2_err_quit(__FILE__, __LINE__,
			   "t2_issueISend returns unrecognized result %d",
			   (int) r);
		return aFAIL_USR;
  }

  pendingl.in(cb);
  return aSUCCEED;
}


T2_AsyncResult
T2_Comm::irecv(char* buf, size_t nbytes, int src, int tag, T2_CommRequest* key)
{
  t2_Assert (nbytes == 0 || buf != NULL);

  t2_AssertOrDo (!freel.isEmpty(), printStatus(cout));
  T2_CommCB* cb = freel.out();
  t2_Assert (cb != NULL);

  cb->key = key;
  T2_AsyncResult r = t2_issueIRecv(*cb, buf, nbytes, src, tag);
  switch (r) {
     case aSUCCEED:
		break;

     case aFAIL_SYS:
		t2_err_sys(__FILE__, __LINE__,
			   "t2_issueIRecv returns aFAIL_SYS");
		return aFAIL_SYS;

     case aFAIL_USR:
		t2_err_ret(__FILE__, __LINE__,
			   "t2_issueIRecv returns aFAIL_USR");
		return aFAIL_USR;

     default:
		t2_err_quit(__FILE__, __LINE__,
			   "t2_issueIRecv returns unrecognized result %d",
			   (int) r);
		return aFAIL_USR;
  }

  pendingl.in(cb);
  return aSUCCEED;
}


bool
T2_Comm::iprobe(int& src, int& tag, int& nbytes)
{
  T2_CommCB cb;
  if (t2_issueIProbe(cb, src, tag) == false)
     return false;

  src = t2_getSource(cb);
  tag = t2_getMsgTag(cb);
  nbytes = t2_getMsgSize(cb);
  return true;
}


u_int
T2_Comm::poll()
{
  u_int count = 0;
  for (u_int i=0; i<getNumberPendingCB();) {
     T2_CommCB* cb = pendingl[i];
     t2_Assert (cb != NULL);

     if (t2_commTestCompletion(*cb) == true) {	// comm operation completes
	t2_Assert (!donel.isFull());
	donel.in(pendingl.out(i));
	count++;
     }
     else {
	i++;
     }
  }

  return count;
}


T2_CommRequest*
T2_Comm::getDoneRequest(int& msgsize, int& source, int& tag)
{
  if (donel.isEmpty() == true)
     return NULL;

  T2_CommCB* cb = donel.out();
  t2_Assert (cb != NULL);

  msgsize = t2_getMsgSize(*cb);
  source = t2_getSource(*cb);
  tag = t2_getMsgTag(*cb);
  T2_CommRequest *r = cb->key;

  t2_Assert ((cb->key = NULL) == NULL);
  t2_AssertOrDo (!freel.isFull(), printStatus(cout));
  cb->nbytes = UINT_MAX;
  freel.in(cb);

  return r;
}


void
T2_Comm::printStatus(ostream& os) const
{
  os << "T2_Comm status:" << endl;
  os << "    # T2_CommCB: total=" << cbs.getNumberElements()
     << ", free=" << getNumberFreeCB()
     << ", pending=" << getNumberPendingCB()
     << ", done=" << getNumberDoneCB();
}



//////////////////////////////////////////////////////////////////////
//
// T2_SendManager methods
//
//////////////////////////////////////////////////////////////////////
void
T2_SendManager::registerISend(T2_CommRequest* req, T2_Comm* comm)
{
  t2_Assert (req != NULL);
  t2_Assert (req->op == t2_cISend || req->op == t2_cISendKillQuery);

  // if req->op is t2_cISend, then req->msgbuf must not be NULL
  t2_Assert (req->op != t2_cISend || req->msgbuf != NULL);

  // if req->op is t2_cISendKillQuery, then req->msgbuf must be NULL
  t2_Assert (req->op != t2_cISendKillQuery || req->msgbuf == NULL);

  if (readyToIssueISend() == true) {	// yes, issue isend immediately
     t2_Assert (comm != NULL);
     issueISend(req, comm);

     t2_say4(cout, "just called isend to proc " << req->proc 
		    << " with nbytes = " << req->msgbuf->getDataSize()
		    << ", npending = " << npending << endl);
  }
  else {				// don't issue immediately
     assert (readyl.isFull() == false);
#    ifdef T2_ASSERTION
        bool e = 
#    endif
     readyl.enqueue(req);	// add to readySends
     t2_Assert (e == true);
  }
}


bool
T2_SendManager::readyToIssueISend()
{
  // not sure what's the right policy is;
  // for now, just make sure that each processor has max_num_pending_isends
  // of isends
  const u_int max_num_pending_isends = 5;
  if (getNumberPendingISends() < max_num_pending_isends)
     return true;
  return false;
}


void
T2_SendManager::issueISend(T2_CommRequest* req, T2_Comm* comm)
{
  t2_Assert (req != NULL);
  t2_Assert (comm != NULL);
  t2_Assert (readyToIssueISend() == true);
  t2_Assert (req->op == t2_cISend || req->op == t2_cISendKillQuery);

  // if req->op is t2_cISend, then req->msgbuf must not be NULL
  t2_Assert (req->op != t2_cISend || req->msgbuf != NULL);

  // if req->op is t2_cISendKillQuery, then req->msgbuf must be NULL
  t2_Assert (req->op != t2_cISendKillQuery || req->msgbuf == NULL);

  T2_AccMsgSendBuffer* msgbuf = req->msgbuf;
  const char *send_buf = NULL;
  size_t send_buf_size = 0;

  if (req->op == t2_cISend)
     msgbuf->getSendBufferInfo(send_buf, send_buf_size);

# ifdef T2_ASSERTION
     T2_AsyncResult r =
# endif
  comm->isend((char *) send_buf, send_buf_size, req->proc, req->tag, req);
  t2_Assert (r == aSUCCEED);
  npending++;
}


void
T2_SendManager::issueReadyISend(T2_Comm* comm)
{
  t2_Assert (comm != NULL);
  t2_Assert (readyToIssueISend() == true);

  if (readyl.isEmpty() == true)		// no ready request in queue
     return;

  T2_CommRequest* req = readyl.dequeue();
  t2_Assert (req != NULL);
  issueISend(req, comm);
}



//////////////////////////////////////////////////////////////////////
//
// T2_CommManager methods
//
//////////////////////////////////////////////////////////////////////
T2_CommManager::T2_CommManager(u_int nprocs, u_int nreqs, T2_Comm* c):
			senders(nprocs),
			requests(nreqs), freeRequests(nreqs),
			nPreRecvs(nprocs),
			nRecvExpect(0), nRecvsPending(0), nSendsCurrent(0),
			nRecvsIssued(nprocs), nSendsIssued(nprocs),
			comm(c)
{
  for (u_int p=0; p<nprocs; p++) {
    senders[p].setMaxNumberReadyISends(nreqs);
    nRecvsIssued[p] = nSendsIssued[p] = 0;
  }

  for (u_int i=0; i<nreqs; i++) {
    freeRequests.in(&(requests[i]));
  }

  // not sure what is the best policy for deciding the max number of pending
  // recvs at any time;
  // for now, just try to keep as many as number of processors, but no more
  // than a threshold, say 16
  max_number_irecvs = t2_min(nprocs, (u_int) 16);
}


T2_CommManager::~T2_CommManager()
{
//  cout << _g_mynode << ": ~T2_CommManager:: npendings=";
//  for (u_int i=0; i<senders.getNumberElements(); i++)
//     cout << " " << senders[i].getNumberPendingISends();
//  cout << "   nready=";
//  for (u_int ii=0; ii<senders.getNumberElements(); ii++)
//     cout << " " << senders[ii].getNumberReadyISends();
//  cout << endl;
  t2_Assert (nRecvExpect == 0);
  t2_Assert (nRecvsPending == 0);
  t2_Assert (nSendsCurrent == 0);
  for (u_int iii=0; iii<senders.getNumberElements(); iii++) {
     t2_AssertOrDo (senders[iii].getNumberReadyISends() == 0,
		    cout << "comm-manger.sender has "
			 << senders[iii].getNumberReadyISends()
			 << " ready sends for node " << iii << endl);
     t2_AssertOrDo (senders[iii].getNumberPendingISends() == 0,
		    cout << "comm-manger.sender has "
			 << senders[iii].getNumberPendingISends()
			 << " pending sends for node " << iii << endl);
  }
}


void
T2_CommManager::initBeforeQueryExecution()
{
  // allow comm_manager to post some number of recvs before any query
  // in the query batch reaches its global combine phase
  nRecvExpect = nPreRecvs;
}


void
T2_CommManager::postISends()
{
  static u_int cur_p = 0;
  const u_int min_free_cb = 1;
  const u_int numnodes = senders.getNumberElements();
  t2_Assert (cur_p < numnodes);
  if (comm->getNumberFreeCB() <= min_free_cb)	// not enough free cb's
     return;

  // For now, I'm sending one isend per processor in a round-robin fashion,
  // and will stop as soon as 
  //	(1) not enough free comm cb's are available, or
  //	(2) nothing was sent after polling each of the send-managers
  // The implementation uses a temporary stop_p to point to the first 
  // send-manager that did not issue any send, while cur_p scans through
  // the send-managers one by one. If a send-manager issues a send, then
  // stop_p is advanced to the next send-manager after that send-manager
  // that issued a send. This effectively reset the ending point to that
  // next send-manager and therefore allows cur_p to go around all the 
  // send-managers at least one more time. Note that this implementation
  // only works for the case where numnodes > 1, so have to treate
  // the case numnodes == 1 differently.
  t2_Assert (comm != NULL);
  t2_Assert (numnodes > 0);
  if (numnodes == 1) {				// keep issuing sends while can
     T2_SendManager& sender = senders[0];
     while (sender.getNumberReadyISends() > 0
	 && sender.readyToIssueISend() == true
	 && comm->getNumberFreeCB() > min_free_cb)
	sender.issueReadyISend(comm);
  }
  else {
     u_int stop_p = cur_p;
     while (1) {
	T2_SendManager& sender = senders[cur_p];
	if (sender.getNumberReadyISends() > 0 &&
	    sender.readyToIssueISend()==true)
	{
	   sender.issueReadyISend(comm);

	   if (++cur_p == numnodes)			// next node
	      cur_p = 0;

	   if (comm->getNumberFreeCB() <= min_free_cb)	// not enough free cb's
	      break;

	   stop_p = cur_p;				// advance stop_p
	}
	else {
	   if (++cur_p == numnodes)			// next node
	      cur_p = 0;
	   if (cur_p == stop_p)				// complete a round
	      break;
	}

     }			// while (1)
  }			// if (numnodes == 1) else
}


void
T2_CommManager::postIRecvs()
{
  // maintain at most max_number_irecvs recvs pending recvs
  for (; nRecvsPending < max_number_irecvs; nRecvsPending++) {
     t2_Assert (nRecvsPending <= nRecvExpect);
     if (nRecvsPending == nRecvExpect) 		// enough irecvs posted
	break;

     int tag = T2_AnyTag;
     int src = T2_AnySource;
     int nbytes;

     t2_Assert (comm != NULL);
     if (comm->iprobe(src, tag, nbytes) == false)	// no message arrived
	break;

     // issue a non-blocking recv for the arrived msg
     t2_Assert (!freeRequests.isEmpty());
     T2_CommRequest* req = freeRequests.out();
     t2_Assert (req != NULL);
     t2_Assert (req->msgbuf == NULL);

     req->op = t2_cIRecv;
     req->proc = src;
     req->tag = tag;

     // only need to allocate buffer if it is an acc msg
     if (req->tag >= T2_AccMsgDataTag) {
        req->msgbuf = new T2_AccMsgSendBuffer;	// though I'm allocating a
						// T2_AccMsgSendBuffer, but
						// I'm only using the base
						// class T2_AccMsgRecvBuffer
        t2_Assert (req->msgbuf != NULL);

#       ifdef T2_ASSERTION
           size_t s = 
#       endif
        req->msgbuf->allocBuffer(nbytes);
        t2_Assert (s >= (size_t) nbytes);
        req->msgbuf->setDataSize(nbytes);

#       ifdef T2_ASSERTION
           T2_AsyncResult r =
#       endif
		comm->irecv((char *) req->msgbuf->getStartOfData(),
			    req->msgbuf->getBufferSize(), src, tag, req);

	   t2_Assert (r == aSUCCEED);
     }
     else {		// if (req->tag >= T2_AccMsgDataTag)

	t2_Assert (nbytes == 0);

#       ifdef T2_ASSERTION
           T2_AsyncResult r =
#       endif
		comm->irecv(NULL, nbytes, src, tag, req);

	   t2_Assert (r == aSUCCEED);
     }			// if (req->tag >= T2_AccMsgDataTag)

     incNumberRecvsIssued(src);
  }
}


void
T2_CommManager::registerISends(T2_QEnv& qenv)
{
  const u_int qid = qenv.getQEnvID();
  T2_Array<T2_AccMsgSendBuffer>& msgbuf = qenv.getAccMsgSendBuffer();
  t2_Assert (msgbuf.getNumberElements() == senders.getNumberElements());

  const u_int numnodes = senders.getNumberElements();
  for (u_int p=0; p<numnodes; p++) {
     T2_AccMsgSendBuffer& buf = msgbuf[p];

     assert (!freeRequests.isEmpty());
     T2_CommRequest* req = freeRequests.out();
     t2_Assert (req != NULL);
     t2_Assert (req->msgbuf == NULL);

     req->op = t2_cISend;
     req->qid = qid;
     req->proc = p;
     req->tag = T2_AccMsgDataTag + qid;
     req->msgbuf = &buf;
     senders[p].registerISend(req, comm);

     t2_say5(cout, "send for query " << qid << " acc msg to prc " << p
		   << ", nbytes=" << buf.getUserDataSize() << endl);

     incNumberSendsIssued(p);
  }

  nSendsCurrent += numnodes;
}


void
T2_CommManager::poll(T2_Array<T2_QEnv>& qenvs, T2_WorkList& wlist)
{
  t2_Assert (comm != NULL);
  comm->poll();

  T2_CommRequest *req;
  int source, msgsize, tag;

  while (!wlist.isFull() &&
	 (req = comm->getDoneRequest(msgsize, source, tag)) != NULL)
  {
     switch (req->op) {
	case t2_cUnknown:
		t2_err_quit(__FILE__, __LINE__,
			    "encounters unexpected t2_cUnknown op");
		return;

	case t2_cISend:
		t2_Assert (req->proc >= 0);
		t2_Assert ((u_int) req->proc < senders.getNumberElements());
		t2_Assert (senders[req->proc].getNumberPendingISends() > 0);
		senders[req->proc].decrementNumberPendingISends();
		actionISendCompletes(req, qenvs, wlist);
		t2_Assert (nSendsCurrent > 0);
		nSendsCurrent--;
		break;

	case t2_cISendKillQuery:
		t2_Assert (req->proc >= 0);
		t2_Assert ((u_int) req->proc < senders.getNumberElements());
		t2_Assert (senders[req->proc].getNumberPendingISends() > 0);
		t2_Assert (req->qid < qenvs.getNumberElements());
		t2_Assert (qenvs[req->qid].isValid() == false);
		t2_Assert (req->msgbuf == NULL);
		senders[req->proc].decrementNumberPendingISends();
		t2_Assert (nSendsCurrent > 0);
		nSendsCurrent--;
		break;

	case t2_cIRecv:
		t2_Assert (source == req->proc);
		t2_Assert (source != -1);
		t2_Assert (tag != -1);
		actionIRecvCompletes(req, qenvs, wlist);
		break;

	default:
		t2_err_quit(__FILE__, __LINE__,
			    "encounters unexpected op = %d", (int) req->op);
		return;
     }

     t2_Assert (!freeRequests.isFull());
     freeRequests.in(req);
  }

  postIRecvs();
  postISends();
}


void
T2_CommManager::actionISendCompletes(T2_CommRequest* req,
				     T2_Array<T2_QEnv>& qenvs,
				     T2_WorkList& wlist)
{
  t2_Assert (req->qid < qenvs.getNumberElements());
  T2_QEnv& qenv = qenvs[req->qid];

  T2_Array<T2_AccMsgSendBuffer>& msgbuf = qenv.getAccMsgSendBuffer();
  msgbuf[req->proc].releaseBuffer();


  t2_say4(cout, "### one isend for query id=" << req->qid 
		<< " completes, dest=" << req->proc 
		<< " size=" << req->msgbuf->getDataSize()
		<< ", #acc_sends_yet_complete = "
		<< qenv.getQueryStatus().getNumberAccSendsYetComplete() - 1 
		<< endl);

  t2_Assert ((req->msgbuf = NULL) == NULL);

  if (qenv.isValid() == true) {
     T2_QState s;
     switch (s = qenv.decNumberAccSendsYetCompleteAndCheckEndOfStatus()) {
       case t2_qGlobal:	// global combine phase done
			{
#			   ifdef T2_ASSERTION
			      bool r = 
#			   endif
			   wlist.enqueueOutputHandle(&qenv);
			   t2_Assert (r == true);
			}
			break;
       case t2_qOutput:	// output handling phase done
			{
#			   ifdef T2_ASSERTION
			      bool r = 
#			   endif
			   wlist.enqueueEndIteration(&qenv);
			   t2_Assert (r == true);
			}
			break;
       case t2_qIdle:	// no phase is done, do nothing
			break;
       default:		// error !!
			t2_err_quit(__FILE__, __LINE__,
				   "unexpected return value %d", (int) s);
     }	   // switch (qenv.decNumberAccSendsYetCompleteAndCheckEndOfStatus())
  }	   // if (qenv.isValid() == true)
}


void
T2_CommManager::actionIRecvCompletes(T2_CommRequest* req,
				     T2_Array<T2_QEnv>& qenvs,
				     T2_WorkList& wlist)
{
  t2_Assert (req != NULL);
  t2_Assert (req->proc >= 0);

  // req->proc is enclosed in () simply due to a seemingly bug in g++2.7.2 
  // parser
  t2_Assert ((u_int) (req->proc) < senders.getNumberElements());

  // is it a message to kill a query ?
  if (req->tag < T2_AccMsgDataTag) {
        u_int qid = T2_AccMsgDataTag - 1 - req->tag;
        t2_Assert (qid < qenvs.getNumberElements());
        t2_say2(cout, "**** recv a kill msg for query " << qid
		     << " from node " << req->proc << endl);

        T2_QEnv& qenv = qenvs[qid];
        if (qenv.isValid() == true) {   // query still valid, so kill it !
           // Need to update the expected number of sends and recvs for 
           // T2_CommManager. 
           // I am assuming here that qenv.qstatus has its numbers of acc 
           // msg sends and recvs set to 0 before the query reaches global 
           // combine phase.
           // I will decrement the number of recvs here. Later, if
           // T2_CommManager recvs another acc msg for this query, or if an
           // acc msg send for this query completes, I will not decrement
           // the number of recvs for T2_CommManager.
           // Note that one must decrement the counters before killing the 
	   // query, because killing a query would reset those counters in 
	   // qenv.qstatus to 0.
	   T2_QStatus& qstatus = qenv.getQueryStatus();
	   if (qstatus.getState() == t2_qGlobal)
	      decNumberExpectedRecvsWhenKillQuery(qstatus);

           qenv.killQueryNow();
        }

	t2_Assert (req->msgbuf == NULL);
     }
  else {		// if (req->tag < T2_AccMsgDataTag)
	// so this is an acc msg for the global combind phase
	// find out the query env id first
        u_int qid = req->tag - T2_AccMsgDataTag;
        t2_Assert (qid < qenvs.getNumberElements());
	t2_Assert (req->msgbuf != NULL);

        t2_say5(cout, "recv for query " << qid << " acc msg from " << req->proc 
			<< ", nbytes=" << req->msgbuf->getDataSize()
			<< ", msgbuf=" << (u_int) req->msgbuf << endl);

        T2_QEnv& qenv = qenvs[qid];
        if (qenv.isValid() == false) {	// invalid query, so drop the msg
           delete req->msgbuf;
        }
        else if (qenv.deferAccMsgProcess(req->proc, req->msgbuf) == false) {
           // this query is already in its global-combine phase, so add the msg
           // to the work list so that it can be processed by the query
           t2_Assert (!wlist.isFull());
#          ifdef T2_ASSERTION
              bool r = 
#          endif
           wlist.enqueueAccMsg(req->proc, req->msgbuf, &qenv);
           t2_Assert (r == true);

           // only decrement the number of expected recvs when the recv'd msg
           // actually belongs to the current iteration; acc msgs for later
           // iterations that arrive early do not count for the quota for the
           // number of expected msgs for the current iteration
           decNumberExpectedRecvs();
	   qenv.getQueryStatus().decNumberAccMsgsYetRecv();

        }		// else if (qenv.deferAccMsgProcess() == false)

	t2_say4(cout, ": ### one irecv for query id = " << qid
		     << " from " << req->proc << " completes"
		     << " " << req->msgbuf->getDataSize() << " bytes"
		     << " nRecvsPending = " << nRecvsPending
		     << " nRecvExpect = " << nRecvExpect
		     << endl);

  }			// if (req->tag < T2_AccMsgDataTag) else

  t2_Assert ((req->msgbuf = NULL) == NULL);

  nRecvsPending--;
}


void
T2_CommManager::killRemoteQuery(const u_int qid,
				const u_int numnodes, const T2_ProcID mynode)
{
  t2_Assert (mynode >= 0);
  for (u_int p=0; p<numnodes; p++) {
     if (p == (u_int) mynode)
	continue;

     assert (!freeRequests.isEmpty());
     T2_CommRequest* req = freeRequests.out();
     t2_Assert (req != NULL);
     t2_Assert (req->msgbuf == NULL);

     req->op = t2_cISendKillQuery;
     req->qid = qid;
     req->proc = p;
     req->tag = T2_AccMsgDataTag - qid - 1;
     req->msgbuf = NULL;
     senders[p].registerISend(req, comm);
     incNumberSendsIssued(p);
  }
  nSendsCurrent += numnodes - 1;
}


void
T2_CommManager::decNumberExpectedRecvsWhenKillQuery(T2_QStatus& qstatus)
{
  t2_Assert (qstatus.getState() == t2_qGlobal);
  t2_AssertOrDo (getNumberExpectedRecvs() >=
			qstatus.getNumberAccMsgsYetRecv(),
		 cout << _g_mynode << "#-expect-recvs="
		      << getNumberExpectedRecvs()
		      << ", #-acc-msg-yet-recv=" 
		      << qstatus.getNumberAccMsgsYetRecv()
		      << endl);
  decNumberExpectedRecvs(qstatus.getNumberAccMsgsYetRecv());
}


void
T2_CommManager::drainMessages(T2_ProcID mynode, T2_Array<T2_QEnv>& qenvs)
{
  // There should be no more pending recvs and no more msgs expected to recv
  // To drain the network, each processor must first go and cancel all
  // the sends that are registered but not yet issued.
  // Then, each processor needs to learn how many more messages (both kill
  // and acc msgs) it needs to receive to drain the network.
  // Here is the strategy. Every processor keeps track of how many recvs
  // it has posted and how many sends it has issued.

  // qglobal (ie T2_QStatus::getGlobalStatus()) to keep track of the number
  // of queries that the local processor kills before being informed by
  // the others (counter local), and the number of kill messages that it 
  // has received from other processors (counter msg).
  // After the main while loop above, all processors do the following:
  //	tot_num_kill_msg = t2_be_gsum(local);
  //	t2_Assert (tot_num_kill_msg >= local + msg);
  //	num_kill_msg_yet_arrive = tot_num_kill_msg - (local + msg);
  //	block on num_kill_msg_yet_arrive msgs;
  //	t2_be_gsync();
  //
  // The global sum at the beginning on one hand makes sure that all nodes
  // have exited the main while loop and therefore cannot send any more
  // kill messages. On the other hand, it finds out the total number of
  // kill messages that each processor should have received, should it had
  // sent a kill message to itself whenever it sended a kill message to some
  // other processor. Therefore, (local+msg) is the total number of kill
  // messages that the local processor has received, and the difference
  // bewteen tot_num_kill_msg and (local+msg) is the number of kill messages
  // that have yet to be received.

  t2_Assert (comm != NULL);
  t2_AssertOrDo (nRecvExpect == nPreRecvs,
		 cout << _g_mynode << ": nRecvExpect=" << nRecvExpect
		      << ", nPreRecvs=" << nPreRecvs << endl);
  nRecvExpect = 0;

  const u_int numnodes = senders.getNumberElements();
  t2_Assert(nSendsIssued.getNumberElements() == numnodes);

  t2_say4(cout, "gonna cancel all queued but not issued sends" << endl);
  t2_say4(cout, "freeRequests has " << freeRequests.getNumberElements()
		<< " of " << freeRequests.getMaxSetSize() << endl);

  // cancel all sends that are queued in the send-managers but not yet
  // issued
  for (u_int i=0; i<numnodes; i++) {
     T2_SendManager& sender = senders[i];
     while (sender.readyl.isEmpty() == false) {
	T2_CommRequest* req = sender.readyl.dequeue();
	t2_Assert (req != NULL);
	switch (req->op) {
	   case t2_cISend:	// need to free memory
		   {
		      t2_Assert (req->qid < qenvs.getNumberElements());
		      T2_QEnv& qenv = qenvs[req->qid];

		      T2_Array<T2_AccMsgSendBuffer>& msgbuf = 
						qenv.getAccMsgSendBuffer();
		      msgbuf[req->proc].releaseBuffer();
		   }
		   break;

	   case t2_cISendKillQuery:
		   // do nothing
		   break;

	   case t2_cIRecv:
	   case t2_cUnknown:
		   // impossible op
		   t2_err_quit(__FILE__, __LINE__,
			"ERROR: unexpected op = %d", req->op);
		   break;
	}

	t2_Assert ((req->msgbuf = NULL) == NULL);
	t2_Assert(req->proc >= 0);
	t2_Assert((u_int) req->proc < numnodes);
	decNumberSendsIssued(req->proc);
	t2_Assert (!freeRequests.isFull());
	freeRequests.in(req);
     }
  }

  t2_say4(cout, "gonna do global sum" << endl);
  t2_say4(cout, "freeRequests has " << freeRequests.getNumberElements()
		<< " of " << freeRequests.getMaxSetSize() << endl);
  t2_say4(cout, "nRecvsPending= " << nRecvsPending << endl);

  // find out the total number of msgs that have been sent to me by other
  // nodes
  T2_Array<u_int> sum_nsends(numnodes);
  t2_be_gsum(nSendsIssued.getElements(), sum_nsends.getElements(), numnodes);
  u_int nmsgs_to_recv = sum_nsends[mynode];
  for (u_int p=0; p<numnodes; p++) {
     t2_Assert (nmsgs_to_recv >= nRecvsIssued[p]);
     nmsgs_to_recv -= nRecvsIssued[p];
  }
  
# if (T2_VERBOSE > 3)
  {
	u_int p;
	t2_say_label(cout);
	cout << "in drainMessages nrecvs= ";
	for (p=0; p<numnodes; p++)
	   cout << " " << nRecvsIssued[p];
	cout << endl;
	t2_say_label(cout);
	cout << "in drainMessages nsends= ";
	for (p=0; p<numnodes; p++)
	   cout << " " << nSendsIssued[p];
	cout << endl;
  }
# endif

  t2_say4(cout, "nmsgs_to_recv = " << nmsgs_to_recv
		<< ", gonna wait for " << nRecvsPending
		<< " pending recvs to complete" << endl);
  t2_say4(cout, "freeRequests has " << freeRequests.getNumberElements()
		<< " of " << freeRequests.getMaxSetSize() << endl);

  // wait for all pending recvs in comm to complete
  while (nRecvsPending > 0) {
     comm->poll();

     T2_CommRequest *req;
     int source, msgsize, tag;

     while ((req = comm->getDoneRequest(msgsize, source, tag)) != NULL)
     {
       switch (req->op) {
	case t2_cUnknown:
		t2_err_quit(__FILE__, __LINE__,
			    "encounters unexpected t2_cUnknown op");
		return;

	case t2_cISend:
		t2_Assert (req->proc >= 0);
		t2_Assert ((u_int) req->proc < senders.getNumberElements());
		t2_Assert (senders[req->proc].getNumberPendingISends() > 0);
		senders[req->proc].decrementNumberPendingISends();
		{
		   t2_Assert (req->qid < qenvs.getNumberElements());
		   T2_QEnv& qenv = qenvs[req->qid];

		   T2_Array<T2_AccMsgSendBuffer>& msgbuf = 
						qenv.getAccMsgSendBuffer();
		   msgbuf[req->proc].releaseBuffer();
		}
		t2_Assert (nSendsCurrent > 0);
		nSendsCurrent--;
		break;

	case t2_cISendKillQuery:
		t2_Assert (req->proc >= 0);
		t2_Assert ((u_int) req->proc < senders.getNumberElements());
		t2_Assert (senders[req->proc].getNumberPendingISends() > 0);
		t2_Assert (req->qid < qenvs.getNumberElements());
		t2_Assert (qenvs[req->qid].isValid() == false);
		t2_Assert (req->msgbuf == NULL);
		senders[req->proc].decrementNumberPendingISends();
		t2_Assert (nSendsCurrent > 0);
		nSendsCurrent--;
		break;

	case t2_cIRecv:
		t2_Assert (source == req->proc);
		t2_Assert (source != -1);
		t2_Assert (tag != -1);
		delete req->msgbuf;
		nRecvsPending--;
		break;

	default:
		t2_err_quit(__FILE__, __LINE__,
			    "encounters unexpected op = %d", (int) req->op);
		return;
       }		// switch (req->op)

       t2_Assert (!freeRequests.isFull());
       freeRequests.in(req);

     }			// while ((req = comm->getDoneRequest()) != NULL)
  }			// while (nRecvsPending > 0)

  t2_say4(cout, "gonna drain all " << nmsgs_to_recv << " msgs" << endl);
  t2_say4(cout, "freeRequests has " << freeRequests.getNumberElements()
		<< " of " << freeRequests.getMaxSetSize() << endl);

  // drain all incoming msgs first
  size_t bufsize = 0;
  char* buf = NULL;
  for (u_int n=0; n<nmsgs_to_recv; n++) {
     int src = T2_AnySource,
	 tag = T2_AnyTag;
     size_t msgsize;
     bool r = t2_be_probe(src, tag, msgsize);
     t2_Assert (r == true);
     t2_Assert (src >= 0);

     if (msgsize > bufsize) {	// make sure buffer large enough
	delete [] buf;
	t2_Alloc (buf, msgsize, char);
	bufsize = msgsize;
     }

     size_t nbytes;
     r = t2_be_recv(buf, msgsize, src, tag, nbytes);	    // blocking recv
     t2_Assert (r == true);
  }
  delete [] buf;

  t2_say4(cout, "gonna wait for " << nSendsCurrent << " sends to complete" 
		<< endl);
  t2_say4(cout, "freeRequests has " << freeRequests.getNumberElements()
		<< " of " << freeRequests.getMaxSetSize() << endl);

  // wait for all pending sends to complete
  t2_Assert (nSendsCurrent == comm->getNumberPendingCB());
  while (nSendsCurrent > 0) {
     comm->poll();

     T2_CommRequest *req;
     int source, msgsize, tag;

     while ((req = comm->getDoneRequest(msgsize, source, tag)) != NULL)
     {
       switch (req->op) {
	case t2_cUnknown:
		t2_err_quit(__FILE__, __LINE__,
			    "encounters unexpected t2_cUnknown op");
		return;

	case t2_cISend:
		t2_Assert (req->proc >= 0);
		t2_Assert ((u_int) req->proc < senders.getNumberElements());
		t2_Assert (senders[req->proc].getNumberPendingISends() > 0);
		senders[req->proc].decrementNumberPendingISends();
		{
		   t2_Assert (req->qid < qenvs.getNumberElements());
		   T2_QEnv& qenv = qenvs[req->qid];

		   T2_Array<T2_AccMsgSendBuffer>& msgbuf = 
						qenv.getAccMsgSendBuffer();
		   msgbuf[req->proc].releaseBuffer();
		}
		t2_Assert (nSendsCurrent > 0);
		nSendsCurrent--;
		break;

	case t2_cISendKillQuery:
		t2_Assert (req->proc >= 0);
		t2_Assert ((u_int) req->proc < senders.getNumberElements());
		t2_Assert (senders[req->proc].getNumberPendingISends() > 0);
		t2_Assert (req->qid < qenvs.getNumberElements());
		t2_Assert (qenvs[req->qid].isValid() == false);
		t2_Assert (req->msgbuf == NULL);
		senders[req->proc].decrementNumberPendingISends();
		t2_Assert (nSendsCurrent > 0);
		nSendsCurrent--;
		break;

	default:
		t2_err_quit(__FILE__, __LINE__,
		       "ERROR: encounters unexpected op = %d", (int) req->op);
		return;
       }		// switch (req->op)

       t2_Assert (!freeRequests.isFull());
       freeRequests.in(req);

     }			// while ((req = comm->getDoneRequest()) != NULL)
  }			// while (nSendsCurrent > 0)

  t2_say4(cout, "done draining, about to return" << endl);
  t2_say4(cout, "freeRequests has " << freeRequests.getNumberElements()
		<< " of " << freeRequests.getMaxSetSize() << endl);
}

