
/* lio_posix.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/6/98
 * Method implementation for asynchronous I/O under the POSIX interface.
 */


#include	<iostream.h>
#include	<string.h>

/////////////////////////////////////
//
// T2_LIOCB methods
//
/////////////////////////////////////
void
initLIOCB(T2_LIOCB* lcb, u_int n)
{
  t2_Assert (lcb != NULL);
  for (u_int i=0; i<n; i++) {
     T2_LIOCB& cb = lcb[i];
     memset(&(cb.aio_sigevent), 0, sizeof(cb.aio_sigevent));
     cb.aio_reqprio = 0;
  }
}


void
setMiscValues(T2_LIOCB& cb, int whence)
{
  // POSIX lio_listio interface only supports whence=SEEK_SET
  t2_Assert (whence == SEEK_SET);
}


// method to extract the return value (ie number of bytes read/written)
// of completed asynchronous I/O operation
int
getReturnValue(T2_LIOCB& cb)
{
  return aio_return(&cb);
}


// method to check if a control block completes
//bool
//opCompletes(T2_LIOCB& cb)
//{
//  aiocb *acb = &(cb.lio_aiocb);
//  return (aio_error(acb) == 0)?	true : false;
//}


/////////////////////////////////////
//
// T2_LIO methods
//
/////////////////////////////////////
void
T2_LIO::initAux()
{
}


void
T2_LIO::deleteAux()
{
}


T2_AsyncResult
T2_LIO::issueReadyRequests()
{
  const u_int nr = readyl.getNumberElements();
  if (nr == 0)
     return aSUCCEED;


#if (T2_TIME_EXECUTE > 1)
{
  double tsnap = t2_dclock();
#endif

  if (lio_listio(LIO_NOWAIT, readyl.getAllElements(), nr, &lio_sigevent) != 0)
  {
     t2_err_sys(__FILE__, __LINE__, "lio_listio returns error");
     return aFAIL_SYS;
  }

#if (T2_TIME_EXECUTE > 1)
  t2_Assert (exetimer != NULL);
  exetimer->incLiolistioTime(t2_dclock() - tsnap);
}
#endif


  // lio_listio succeeds, so move all liocb in readyl to pendingl
  bool r = pendingl.in(readyl);
  t2_Assert(r == true);
  readyl.emptySet();

  return aSUCCEED;
}


u_int
T2_LIO::poll()
{
  u_int count = 0;
  for (u_int i=0; i<pendingl.getNumberElements();) {
     T2_LIOCB *cb = pendingl[i];
     t2_Assert (cb != NULL);

     aiocb *acb = cb;
     if (aio_error(acb) == 0) {			// this async I/O completes


//#ifdef T2_TIME
//{
//      int n = cb - lcb;
//      lio_elapse_t += t2_dclock() - start_t[n];
//      lio_elapse_cnt++;
//      t2_Assert((start_t[n] = -1.) < 0.);
//}
//#endif


	t2_Assert(!donel.isFull());
	donel.in(pendingl.out(i));
	count++;
     }
     else					// this async I/O is not done
	i++;					// go check the next T2_LIOCB
  }

  return count;
}

