
/* $Id: pproc.C,v 1.16 2000/11/14 15:39:46 chialin Exp $ */

/* pproc.C */

#include	<iostream.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<limits.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_qspec.h"
#include	"t2_rts.h"
#include	"t2_qenv.h"
#include	"t2_blockreq.h"
#include	"t2_clusterreq.h"
#include	"t2_qplan.h"
#include	"t2_clusterinfo.h"
#include	"t2_comm.h"


extern int _g_mynode;


/////////////////////////////////////////////////////////////////////
//
// declaration of some function prototypes
//
/////////////////////////////////////////////////////////////////////
void
init_qenvs(T2_Array<T2_QEnv>& qenv, T2_RTS& rts, T2_QBatch& qbatch,
	   T2_QPlan& qplan);
void
gen_all_requests(T2_QBatch& qbatch, T2_RTS& rts,
		 T2_Array<T2_QEnv>& qenv, T2_QPlan& qplan);
bool
gen_requests(T2_QSpec& qspec, T2_IndexCatalog& indexcat, T2_RTSDSList& dslist,
	     T2_Array<T2_DatasetEnv>& datasetenv,
	     T2_BlockRequestListMap& blockreqmap, const T2_System& system);
u_int
add_disk_requests_to_list(T2_VArray<T2_BlockRequest>& reqs,
			  T2_DSFileMap& dfmap,
			  T2_ClusterRequestList& clusterreq,
			  T2_PhyBlockRequestList& blockreq,
			  const T2_Box& mbr, T2_UsrArg& cinfo_buf);
void
strip_mine_all(T2_RTS& rts, T2_QPlan& qplan, T2_Array<T2_QEnv>& qenv,
	       size_t mem);
bool
strip_mine(T2_QQPlan& qqplan, T2_QEnv& qenv, const T2_System& system,
	   const size_t mem);
void
execute_qplan(T2_QPlan& qplan, T2_Array<T2_QEnv>& qenv, T2_RTS& rts,
	      bool verbose);


/////////////////////////////////////////////////////////////////////
//
// close the file descriptors stored from fd[s] through fd[e-1]
//
/////////////////////////////////////////////////////////////////////
inline void
closeFiles(u_int s, u_int e, const T2_Array<int>& fd)
{
  for (u_int j=s; j<e && j<fd.getNumberElements(); j++) {
     if (fd[j] >= 0)
        close(fd[j]);
  }
}



/////////////////////////////////////////////////////////////////////
//
// query processing function
//
/////////////////////////////////////////////////////////////////////
void query_processing(T2_RTS& rts, T2_QBatch& qbatch, size_t mem, bool verbose)
{
# ifdef T2_GETRUSAGE
    struct rusage rr1, rr2;
    getrusage(RUSAGE_SELF, &rr1);
# endif

# if (T2_VERBOSE > 2)
     if (rts.getSystem().getMyNode() == 0)
        cout << "---- query batch received ----" << endl
	     << qbatch << endl;
     t2_be_gsync();
# endif

# ifdef T2_TIME_PLANNING
    double tsnaps[6];
    tsnaps[0] = t2_dclock();
# endif

  u_int q;
  const u_int numqueries = qbatch.getNumberQueries();

  // initialize query plan, which has a T2_QQPlan for each query
  T2_QPlan qplan(numqueries);

  // initialize a query environment for each given query in query batch
  t2_say(cout, "initializing a query env for each given query"<<endl);
  T2_Array<T2_QEnv> qenv(numqueries);
  init_qenvs(qenv, rts, qbatch, qplan);

# ifdef T2_TIME_PLANNING
     tsnaps[1] = t2_dclock();
# endif

  // generate disk requests from qenv
  t2_say(cout, "generating disk requests for all queries"<<endl);
  gen_all_requests(qbatch, rts, qenv, qplan);

# ifdef T2_TIME_PLANNING
     tsnaps[2] = t2_dclock();
# endif

  // strip-mining
  t2_say(cout, "performing strip-mining" << endl);
  strip_mine_all(rts, qplan, qenv, mem);

# ifdef T2_TIME_PLANNING
     tsnaps[3] = t2_dclock();
# endif

  // communicate with all back-end nodes to make sure that for each query,
  // all back-end nodes come up with the same number of strip-mined iterations
  // (ie output tiles);
  // for now, just invalidate those queries that have different number of
  // iterations on different processors;
  // well, not sure how to do this efficiently now, so I'll just do for each
  // query a min and max number of iterations across the processors and 
  // check it against the local number of iterations for that query;
  // if the min and max are not identical, then invalidate the query;
  // if the query is invalid, then use UINT_MAX as the number of iterations,
  // hoping that no query would ever need that many iterations
  {
    T2_Array<u_int> my_ntiles(numqueries),
		    min_ntiles(numqueries),
		    max_ntiles(numqueries);
    for (q=0; q<numqueries; q++) {
       if (qenv[q].isValid() == true)
	  my_ntiles[q] = qplan.getQQPlan(q).getNumberSubplans();
       else
	  my_ntiles[q] = UINT_MAX;		// invalid query
    }
    t2_be_gmax(my_ntiles.getElements(), max_ntiles.getElements(), numqueries);
    t2_be_gmin(my_ntiles.getElements(), min_ntiles.getElements(), numqueries);
    for (q=0; q<numqueries; q++) {
       if (qenv[q].isValid() == false)
	  continue;
       if (max_ntiles[q] != min_ntiles[q]) {
	  qenv[q].invalidate();
	  if (rts.getSystem().getMyNode() == 0)
	     t2_say4(cout, "WARNING: invalidating query " << q 
			<< " because inconsistent # iterations: min = " <<
			min_ntiles[q] << " max = " << max_ntiles[q] << endl);
       }
    }
  }

# ifdef T2_TIME_PLANNING
     tsnaps[4] = t2_dclock();
# endif

  // sort the subplans here, if decided to
  t2_Assert (qenv.getNumberElements() == qplan.getNumberQQPlans());
  for (q=0; q<numqueries; q++)
     if (qenv[q].isValid() == true)
	qplan.getQQPlan(q).sortClusterRequests();

# ifdef T2_TIME_PLANNING
     tsnaps[5] = t2_dclock();
# endif

# if (defined T2_TIME_EXECUTE || defined T2_TIME_QUERY)
     // sync before entering execution service
     t2_say2(cout,
	     "synchronizing back-end nodes before entering execution service"
	     << endl);
     t2_be_gsync();
# endif

  // now execute the query plan
  t2_say(cout, "executing query plan" << endl);
  execute_qplan(qplan, qenv, rts, verbose);


  // deactivate all dataset entries in rts
  rts.deactivateAllDatasets();


  // if one wants to force T2 to close all the fd's for datasets used
  // by the queries, then do the following
  // rts.getRTSDatasetInfo().killAll();


  t2_say2(cout, "done query processing" << endl);


# ifdef T2_TIME_PLANNING
     t2_say_label(cout);
     cout << "total-plan-time= "	<< tsnaps[5] - tsnaps[0]
	  << " (initqenv="		<< tsnaps[1] - tsnaps[0]
	  << " genreq="			<< tsnaps[2] - tsnaps[1]
	  << " stripmine="		<< tsnaps[3] - tsnaps[2]
	  << " checkntiles="		<< tsnaps[4] - tsnaps[3]
	  << " sort="			<< tsnaps[5] - tsnaps[4]
	  << " )" << endl;
# endif


# ifdef T2_GETRUSAGE
	getrusage(RUSAGE_SELF, &rr2);
	t2_say_label(cout);
	cout << "major page faults = " << rr2.ru_majflt - rr1.ru_majflt << endl;
	t2_say_label(cout);
	cout << "user time = "
           << (double) rr2.ru_utime.tv_sec + rr2.ru_utime.tv_usec/1000000. -
              ((double) rr1.ru_utime.tv_sec + rr1.ru_utime.tv_usec/1000000.)
           << ", system time = "
           << (double) rr2.ru_stime.tv_sec + rr2.ru_stime.tv_usec/1000000. -
              ((double) rr1.ru_stime.tv_sec + rr1.ru_stime.tv_usec/1000000.)
           << endl;
# endif
}


/////////////////////////////////////////////////////////////////////
//
// initializing query environment
//
/////////////////////////////////////////////////////////////////////
void
init_qenvs(T2_Array<T2_QEnv>& qenv, T2_RTS& rts, T2_QBatch& qbatch,
	   T2_QPlan& qplan)
{
  t2_Assert (qenv.getNumberElements() == qbatch.getNumberQueries());
  t2_Assert (qenv.getNumberElements() == qplan.getNumberQQPlans());

  const u_int nqry = qbatch.getNumberQueries();
  for (u_int i=0; i<nqry; i++) {
     T2_QEnv& qe = qenv[i];
//     if (!qe) {
//	t2_say2(cout, "query " << i << " is invalid entering init_qenvs,"
//			<< " so skipping its init" << endl);
//	continue;
//     }

     if (qe.init(i, qbatch.getQuerySpec(i), rts) == false) {	// init failed
	t2_say2(cout, "query " << i << " init returns false, so set invalid"
			<< endl);
	qe.invalidate();
	continue;
     }
     qe.setQQPlan(qplan.getQQPlan(i).getAddress());
  }
}


/////////////////////////////////////////////////////////////////////
//
// generating disk requests for all queries
//
/////////////////////////////////////////////////////////////////////
void
gen_all_requests(T2_QBatch& qbatch, T2_RTS& rts,
		 T2_Array<T2_QEnv>& qenv, T2_QPlan& qplan)
{
  t2_Assert (qenv.getNumberElements() == qbatch.getNumberQueries());

  const u_int nqry = qbatch.getNumberQueries();
  for (u_int i=0; i<nqry; i++) {
     T2_QEnv& qe = qenv[i];
     if (!qe) { 			// an invalid query env
	t2_say2(cout, "query " << i << " is invalid entering gen_all_requests"
			<< endl);
	continue;
     }
     bool b = gen_requests(qbatch.getQuerySpec(i),
		rts.getIndexCatalog(), rts.getRTSDatasetInfo(),
		qe.getDatasetEnvList(),	    // actually just need clusterreq
		qplan.getBlockRequestListMap(), rts.getSystem());

     if (b == false) {			// gen_requests failed
	t2_say2(cout, "query " << i 
			<< " gen_requests returns false, so set to invalid"
			<< endl);
	qe.invalidate();
     }
  }

  qplan.initBlockRequests();		// initialize the block requests
}


/////////////////////////////////////////////////////////////////////
//
// generating disk requests
//
/////////////////////////////////////////////////////////////////////
bool
gen_requests(T2_QSpec& qspec, T2_IndexCatalog& indexcat,
	     T2_RTSDSList& dslist, T2_Array<T2_DatasetEnv>& datasetenv,
	     T2_BlockRequestListMap& blockreqmap, const T2_System& system)
{
  const u_int nds = qspec.getNumberDatasets();
  for (u_int i=0; i<nds; i++) {		// step through each dataset
     u_int j;				// loop index
     T2_UDFRet rv = T2_UDFRet_OK;

     // identify all relevant objects
     T2_QSpecDataset& dsspec = qspec.getDatasetSpec(i);
     T2_DSID ds_id = dsspec.getDatasetID();
     T2_IndexID index_id = dsspec.getIndexID();
     T2_DatasetEnv& dsenv = datasetenv[i];
     T2_ClusterRequestList& clusterreq = dsenv.getClusterRequestList();
     T2_PhyBlockRequestList& blockreq =
				blockreqmap.touchBlockRequestList(ds_id);

     // get the dataset entry from rts
     bool flag;
     T2_RTSDSEntry& dsentry = dslist.search(ds_id, flag);
     if (flag == false) {
        t2_err_ret(__FILE__, __LINE__,
	   "ERROR: cannot find in T2_RTSDSList dataset with id = %d", ds_id);
	return false;
     }

     T2_DSFileMap& dfmap = dsentry.getDSFileMap();
     const u_int ndsfiles = dfmap.getNumberElements();
     if (ndsfiles == 0) {		// no assigned data files
	// assume no local data
	continue;
     }

     // get the index entry from the index catalog
     T2_IndexCatEntry index_catentry;
     if (indexcat.getIndexCatalogEntry(index_id, index_catentry) == false) {
        t2_err_ret(__FILE__, __LINE__,
		 "ERROR: cannot read from catalog info for index %d", index_id);
        return false;
     }

     // set the bitmap that marks the index files required to support the
     // assigned data files for the given dataset
     T2_BitVector my_idxfiles(0);
     u_int n_my_idxfiles =
	index_catentry.support(dsentry.getDSFileIDBitVector(), my_idxfiles);
     if (n_my_idxfiles == 0) {          // no index file !!
	t2_err_ret(__FILE__, __LINE__,
	   "ERROR: no assigned index file while there is assigned data files.");
	return false;
     }

     t2_Assert (clusterreq.getNumberClusterRequests() == 0);

     // open index files;
     // fd[i] would store the file descriptior of the i-th index file
     // needed to support all local data files, and
     // idx_fileid[i] would store the position of the i-th index file
     // as in the order all index files are stored in the index catalog
     // (ie, idx_fileid[2] = 5 means the 5th index file listed in the index
     // catalog is the 2nd index file needed to support the local data files)
     const u_int nfiles = index_catentry.getNumberFiles();
     T2_Array<int> fd(n_my_idxfiles);
     T2_Array<int> idx_fileid(n_my_idxfiles);
     u_int rj;
     for (j=0, rj=0; j<nfiles; j++) {
	if (my_idxfiles.isSet(j) == true) {
	  const u_int num_tries = 10;
	  u_int nnn;
	  for (nnn=0; nnn<num_tries; nnn++) {
	      t2_say2(cout, "openning index file "
				<< index_catentry.getFileName(j) << endl);

	     if ((fd[rj] = open(index_catentry.getFileName(j), O_RDONLY)) > 0)
		break;

//	     t2_err_ret(__FILE__,__LINE__,
//			"cannot open index file %s, will retry",
//			index_catentry.getFileName(j));
	     usleep(1000);
	  }

	  if (nnn == num_tries) {		// all read attempts failed
	     t2_err_ret(__FILE__,__LINE__,
			"cannot open index file %s after %d tries",
			 index_catentry.getFileName(j), num_tries);
	     closeFiles(0, j, fd);
	     return false;
	  }

	  idx_fileid[rj] = j;
	  rj++;
	}
     }

     // find the file descriptors for the data files of the dataset
     T2_Index *index = NULL;

     {
        // construct the index from its constructor
        T2_IndexConstructorPtr index_cons =
	    T2_RTS::getIndexConstructorPtr(index_catentry.getConstructorID());
        if (index_cons) {
	     T2_UsrArgReader reader(dsspec.getIndexConstructorArg());
	     rv = index_cons(system, reader, fd, index);
        }
        if (index == NULL) {
	     t2_err_ret(__FILE__, __LINE__,
		"ERROR: index constructor for index %d returns NULL.",
		index_id);
	     closeFiles(0, nfiles, fd);
	     return false;
        }
	if (rv == T2_UDFRet_ERROR) {
	     t2_err_ret(__FILE__, __LINE__,
		"ERROR: index constructor for index %d returns ERROR.",
		index_id);
	     closeFiles(0, nfiles, fd);
	     return false;
	}
     }


     // now generate the list of disk requests

     t2_say3(cout, "gonna call index.fetchInit()" << endl);

     // first let the index initialize itself before search
     const T2_Box& qbox = dsspec.getQueryBox();
     index->fetchInit(system, qbox);

     // about ready to do an index lookup to find all logical chunks
     // (ie clusters) that intersect the query box
     u_int wc = 0;				// just a counter
     T2_UsrArg cinfo_buf;			// buffer for aux cluster info
     T2_ClusterAuxInfoWriter cinfo(cinfo_buf);
     bool eod = true;				// end-of-data flag
     T2_Box mbr;				// an empty bounding box
     T2_VArray<T2_BlockRequest> reqs(20);	// physical block requests

     t2_say3(cout, "gonna call index.fetch()" << endl);

     // now retrieve the logical chunks and their corresponding physical
     // chunk requests
     while ((rv = index->fetch(system, qbox, eod, mbr, cinfo, reqs))
	     == T2_UDFRet_OK)
     {
	if (eod == true)
	   break;				// all done

	wc++;
	cinfo.close();
	const u_int nn = reqs.getNumberElements();

	t2_say5(cout, "index fetch() called " << wc << "-th time, returns "
			<< nn << " block requests" << endl);

	if (nn == 0) {			// no chunk request !!
//	   t2_err_ret(__FILE__, __LINE__,
//		     "No chunk request was generated when index fetch() called %d-th time",
//		     wc);
	   eod = true;
	   mbr.setNumberDimensions(0);
	   cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	   cinfo.open(cinfo_buf);	// must rebind cinfo with cinfo_buf
	   reqs.clear();
	   continue;
	}

	// for now, I am assuming that for a given logical chunk id, either
	// all its actual chunk requests go to local disks, or none of them do;
	// ie, no joins between processors can happen;
	// this makes scheduling much easier  :)

	// check the first chunk request to see if local reads are needed
	int fd_t;
	u_int diskno_t;
	if (dfmap.getFileInfo(reqs[0].getFileID(), fd_t, diskno_t) == false) {
	   eod = true;
	   mbr.setNumberDimensions(0);
	   cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	   cinfo.open(cinfo_buf);	// must rebind cinfo with cinfo_buf
	   reqs.clear();
	   continue;			// not a local data chunk
	}

	// so this is a local data cluster request, needs to add to req list
	t2_say5(cout, "this is my local cluster, add to local req list"
			<< endl);
	u_int cnt =
		add_disk_requests_to_list(reqs, dfmap, clusterreq, blockreq,
					  mbr, cinfo_buf);
	// make sure that the buffer in info has been "stolen" and put into
	// the cluster list by add_disk_requests_to_list()
	t2_Assert (cinfo_buf.getBufferSize() == 0);
	t2_Assert (cinfo_buf.getDataSize() == 0);

	if (cnt != reqs.getNumberElements()) {
//	   t2_err_ret(__FILE__, __LINE__,
//		      "Only %d out of %d block requests are on local disks\nCurrent implementation assumes that all block requests of a cluster returned by\nan index must reside on the same processor",
//		      cnt, reqs.getNumberElements());

	   delete index;
	   closeFiles(0, nfiles, fd);
	   return false;
	}

	// get ready to call fetch again
	eod = true;
	mbr.setNumberDimensions(0);
	cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	cinfo.open(cinfo_buf);		// must rebind cinfo with cinfo_buf
	reqs.clear();

     }		// while ((rv = index->fetch()) == T2_UDFRet_OK)

     t2_say2(cout, "index.fetch sets eod=true when called the " << wc+1
		   << " times" << endl);
     t2_say2(cout, "total # local cluster reqs = "
		   << clusterreq.getNumberClusterRequests() << endl);

     // make each cluster request in clusterreq point to dsenv, and init
     // its reference count
     for (j=0; j<clusterreq.getNumberClusterRequests(); j++) {
	T2_ClusterRequest& cls = clusterreq[j];
	cls.setDatasetEnv(dsenv.getAddress());
	cls.initRefCount();
     }

     // release memory
     delete index;
     closeFiles(0, nfiles, fd);

     if (rv != T2_UDFRet_OK) 			// some thing bad happened
	return false;

  }		// for (u_int i=0; i<nds; i++) // step thru each dataset

  return true;
}



/////////////////////////////////////////////////////////////////////
//
// adding local disk requests into the appropriate lists,
// and returning the number of block requests added to list blockreq
//
/////////////////////////////////////////////////////////////////////
u_int
add_disk_requests_to_list(T2_VArray<T2_BlockRequest>& reqs,
			  T2_DSFileMap& dfmap,
			  T2_ClusterRequestList& clusterreq,
			  T2_PhyBlockRequestList& blockreq,
			  const T2_Box& mbr, T2_UsrArg& cinfo_buf)
{
  const u_int nn = reqs.getNumberElements();

  // insert a new cluster request into clusterreq
  T2_ClusterRequest& cluster = clusterreq.insert();
  cluster.setNumberBlockRequests(nn);
  cluster.setMBR(mbr);
  cluster.getMetaData().steal(cinfo_buf);

  u_int cnt = 0;
  for (u_int jj=0; jj<nn; jj++) {
      // get file descriptor and disk number
      int fdesc;
      u_int diskno;
      T2_BlockRequest& chk = reqs[jj];
      if (!dfmap.getFileInfo(chk.getFileID(), fdesc, diskno)) {	// not local !!
	 t2_err_quit(__FILE__, __LINE__,
	    "ERROR: this shouldn't happen, since I have checked this before");
	 continue;
      }

      // insert a new block request into blockreq
      T2_PhyBlockRequest& blockr = blockreq.touch(fdesc, chk.getOffset(),
					          chk.getNumberBytes());

      // now set the cross pointers between cluster request and block request
      // first make cluster request point to the new block request
      cluster.setBlockRequest(jj, blockr.getAddress());
      // now have the block request point back to the cluster request
      blockr.addStaticOwner(cluster.getAddress());

      cnt++;
  }

  return cnt;
}


/////////////////////////////////////////////////////////////////////
//
// strip-mining all queries
//
/////////////////////////////////////////////////////////////////////
void
strip_mine_all(T2_RTS& rts, T2_QPlan& qplan, T2_Array<T2_QEnv>& qenv,
	       size_t mem)
{
  const u_int nqry = qenv.getNumberElements();
  for (u_int i=0; i<nqry; i++) {
     if (qenv[i]) {
	if (strip_mine(qplan.getQQPlan(i), qenv[i], rts.getSystem(), mem)
	    == false)
	{
	   // strip_mine failed
	   qenv[i].invalidate();
	   continue;
	}

	t2_say2(cout, "query " << i << " strip_mine generates "
		  << qplan.getQQPlan(i).getNumberSubplans() << " iterations"
		  << endl);

	if (qplan.getQQPlan(i).getNumberSubplans() == 0) {
	   // hmmm....., zero iteration?
	   // don't know what to do, kill the query for now
	   t2_err_ret(__FILE__, __LINE__,
		     "warning: query %d generates 0 strip-mined iterations",
		     qenv[i].getQEnvID());
	   qenv[i].invalidate();
	}
     }
     else {
	t2_Assert (qplan.getQQPlan(i).getNumberSubplans() == 0);

	t2_say2(cout, "query " << i << " is invalid, so skipping its strip_mine"
		      << endl);

     }
  }
}


/////////////////////////////////////////////////////////////////////
//
// strip-mining
//
/////////////////////////////////////////////////////////////////////
bool
strip_mine(T2_QQPlan& qqplan, T2_QEnv& qenv, const T2_System& system,
	   const size_t mem)
{
  // get the accumulator meta object
  t2_Assert (qenv.getAccMetaPtr() != NULL);
  T2_AccMetaObj& accmeta = *(qenv.getAccMetaPtr());

  T2_VArray<T2_Region>& qwindows = qenv.getQueryWindows();
  t2_Assert (qenv.getNumberQueryWindows() == 0);
  if (accmeta.stripMine(qenv.getQueryInfo(), mem, T2_ClusterInfoList(qenv),
			qwindows) != T2_UDFRet_OK)
  {
      t2_say2(cout, "T2_AccMetaObj::stripMine() FAILED." << endl);
      return false;
  }

  t2_say2(cout, "T2_AccMetaObj::stripMine() returns " 
		<< qwindows.getNumberElements() << " tiles" << endl);

  for (u_int j=0; j<qwindows.getNumberElements(); j++)
      t2_say4(cout, "tile " << j << ": " << qwindows[j] << endl);

  // partition the disk requests by iterations
  const u_int nn = qenv.getNumberDatasets();
  for (u_int i=0; i<nn; i++) {
     T2_DatasetEnv& dsenv = qenv.getDatasetEnv(i);
     qqplan.insertClusterRequests(dsenv.getClusterRequestList(), qwindows,
				  *(dsenv.getProjectFuncObjPtr()),
				  system);
  }

  return true;
}

