
/* $Id: work.C,v 1.12 2000/03/16 00:44:46 chialin Exp $ */

/* work.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 2/1/98
 * T2 work list
 */

#include	"t2_work.h"

extern T2_Iteration niter;

extern int _g_mynode;

/////////////////////////////////////////////////
//
// static variables
//
/////////////////////////////////////////////////
T2_RTS* T2_WorkEntry::rts = NULL;
T2_WorkList* T2_WorkEntry::wlist = NULL;
T2_CommManager* T2_WorkEntry::comm = NULL;


/////////////////////////////////////////////////
//
// a helper function to remove all acc msgs
// that arrive before the query enters the
// global combine phase and are thus buffered
// by the qenv;
// this method should be called when error
// occurs and the query corresponding to qenv
// is to be killed;
//
/////////////////////////////////////////////////
//inline void
//delete_recved_acc_msgs(T2_QEnv& qenv)
//{
//  const u_int nrecvbufs = qenv.getNumberAccMsgRecvBuffers();
//  for (u_int j=0; j<nrecvbufs; j++) {
//    T2_ProcID src;
//    T2_AccMsgRecvBuffer* msg;
//
//#   ifdef T2_ASSERTION
//	bool q =
//#   endif
//    qenv.getAccMsgRecvBuffer(j, src, msg);
//    t2_Assert (q == true);
//    t2_Assert (msg != NULL);
//
//    delete msg;
//  }
//}


/////////////////////////////////////////////////
//
// methods for T2_ClusterWork
//
/////////////////////////////////////////////////
void
T2_ClusterWork::handler()
{
  t2_say4(cout, "T2_ClusterWork::handler for cluster "
		<< (int) cluster << " is invoked ##################" << endl);

  t2_Assert (cluster != NULL);
  t2_Assert (cluster->getRefCount() == 0);

  T2_DatasetEnv* dsenv = cluster->getDatasetEnv();
  t2_Assert (dsenv != NULL);

  T2_QEnv* qenv = dsenv->getQEnv();
  t2_Assert (qenv != NULL);
  if (qenv->isValid() == false) {
     // reset ref count
     cluster->initRefCount();

     // release all block buffers
     const u_int nblocks = cluster->getNumberBlockRequests();
     for (u_int i=0; i<nblocks; i++)
        cluster->getBlockRequest(i)->release();

     return;
  }

  const T2_QQPlan* qqplan = qenv->getQQPlan();
  t2_Assert (qqplan != NULL);
  T2_Iteration iter = qqplan->getCurrentIteration();
  t2_Assert (iter < qenv->getNumberQueryWindows());
  const T2_Region& tilereg = qenv->getQueryWindow(iter);

  T2_Cluster data(*cluster);
  T2_ClusterAuxInfoReader meta(cluster->getMetaData());

  T2_Dataset* dataset = dsenv->getDatasetPtr();
  t2_Assert (dataset != NULL);
  T2_Accumulator *accptr = qenv->getAccumulator();
  t2_Assert (accptr != NULL);
  T2_ProjectFuncObj *proj = dsenv->getProjectFuncObj();
  t2_Assert (proj != NULL);
  T2_AggregateFuncObj* aggr = qenv->getAggregateFuncObject();
  t2_Assert (aggr != NULL);
  t2_Assert (rts != NULL);
  T2_UDFRet rv;

# if (T2_TIME_QUERY > 1)
  {
     T2_QQPlan* qqp = qenv->getQQPlan();
     t2_Assert (qqp != NULL);
     t2_Assert (qqp->getNumberSubplans() > 0);
     T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
     double tsnap = t2_dclock();
# endif

  // process input data chunk
  rv = aggr->dafAcc(qenv->getQueryInfo(), data, cluster->getMBR(),
			meta, *dataset, dsenv->getDatasetID(),
			dsenv->getIteratorID(), dsenv->getQueryRegion(),
			tilereg, *proj, qenv->getAccNavigatorID(), *accptr);

# if (T2_TIME_QUERY > 1)
     qtimer.incDafAccTime(t2_dclock() - tsnap);
  }
# endif


  // reset ref count
  cluster->initRefCount();

  // release all block buffers
  const u_int nblocks = cluster->getNumberBlockRequests();
  for (u_int i=0; i<nblocks; i++)
     cluster->getBlockRequest(i)->release();

  // if the aggregation function failed earlier, invalidate query on local
  // and remote nodes
  if (rv != T2_UDFRet_OK) {
     t2_Assert (comm != NULL);
     const T2_System& sys = rts->getSystem();
     comm->killRemoteQuery(qenv->getQEnvID(),
			   sys.getNumberNodes(), sys.getMyNode());
     qenv->killQueryNow();
     return;
  }

  // check if this is the last cluster to process for the current iteration
  t2_Assert (qenv->getQueryStatus().getNumberClustersYetProcess() > 0);
  if (qenv->decNumberClustersYetToProcessAndCheckEndLocalReduction() == true) {
     // this is the last cluster to process for the current iteration,
     // so advance to global combine phase
     wlist->enqueueGlobalCombine(qenv);
  }


//cout << "qenv " << (int) qenv << " ncluster_yetprocessed becomes "
//<< qenv->getQueryStatus().getNumberClustersYetProcess() << endl;


}



/////////////////////////////////////////////////
//
// methods for T2_GlobalCombineWork
//
/////////////////////////////////////////////////
void
T2_GlobalCombineWork::handler()
{
  t2_say4(cout, "T2_GlobalCombineWork::handler for qenv "
		<< (int) qenv << " is invoked ##################" << endl);

  t2_Assert (qenv != NULL);
  t2_Assert (rts != NULL);


# ifdef T2_TIME_QUERY
     T2_QQPlan* qqp = qenv->getQQPlan();
     t2_Assert (qqp != NULL);
     t2_Assert (qqp->getNumberSubplans() > 0);
     T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
     qtimer.setStartCombine();
# endif


  t2_say4(cout, "query " << qenv->getQEnvID()
		<< " enter global-combine phase at iter " << niter << endl);

  T2_AggregateFuncObj* aggr = qenv->getAggregateFuncObject();
  t2_Assert (aggr != NULL);
  T2_Accumulator *acc = qenv->getAccumulator();
  t2_Assert (acc != NULL);

  if (qenv->isValid() == false) {
     // remove all acc msgs that had arrived earlier
//     delete_recved_acc_msgs(*qenv);
     return;
  }

  if (aggr->needGlobalCombine(qenv->getQueryInfo(), *acc) == false) {
     // no need for global combine phase

     // check if any acc msg had arrived earlier
     const u_int nrecvbufs = qenv->getNumberAccMsgRecvBuffers();
     if (nrecvbufs > 0) {		// error, should kill query
	// remove all acc msgs buffered by qenv
//	delete_recved_acc_msgs(*qenv);

	// kill query 
	t2_Assert (comm != NULL);
	const T2_System& sys = rts->getSystem();
	comm->killRemoteQuery(qenv->getQEnvID(),
			      sys.getNumberNodes(), sys.getMyNode());
	qenv->killQueryNow();
	return;
     }

     // proceed to output-handling phase
     T2_OutputHandleWork outhandle(qenv);
     outhandle.handler();

     return;
  }

  // need to go through global combine phase
  T2_System& system = rts->getSystem();
  const u_int numnodes = system.getNumberNodes();

  t2_say4(cout, "T2_GlobalCombineWork calling initGlobalCombine(" << numnodes
         << ")" << endl);
  qenv->initGlobalCombine(numnodes, numnodes);

  // generate acc messages
  T2_Array<T2_AccMsgSendBuffer>& msgbuf = qenv->getAccMsgSendBuffer();
  T2_Array<T2_AccMsgBufferWriter> writers(numnodes);

  for (u_int i=0; i<numnodes; i++) {
     t2_Assert (msgbuf[i].getBufferSize() == 0);
     t2_Assert (msgbuf[i].getUserDataPtr() == NULL);
     t2_Assert (msgbuf[i].getUserDataSize() == 0);
     writers[i].open(msgbuf[i]);
  }

  T2_UDFRet rv;

# if (T2_TIME_QUERY > 1)
  {
     double tsnap = t2_dclock();
# endif


  // fill msg buffer
  rv = aggr->fillAccMsgBuffer(qenv->getQueryInfo(), *acc, writers);


# if (T2_TIME_QUERY > 1)
     qtimer.incFillAccMsgBufferTime(t2_dclock() - tsnap);
  }
# endif


  for (u_int ii =0; ii<numnodes; ii++)		// close the writers
     writers[ii].close();

  if (rv != T2_UDFRet_OK) {			// error, so kill the query
     // remove all acc msgs buffered by qenv
//     delete_recved_acc_msgs(*qenv);

     // update expected number of recvs in T2_CommManager
     t2_Assert (comm != NULL);
     comm->killRemoteQuery(qenv->getQEnvID(),
			   system.getNumberNodes(), system.getMyNode());

     // kill query
     qenv->killQueryNow();
     return;
  }

  // enqueue all the acc msgs that had arrived earlier into worklist
  t2_Assert (wlist != NULL);
  const u_int nrecvbufs = qenv->getNumberAccMsgRecvBuffers();
  for (u_int j=0; j<nrecvbufs; j++) {
     T2_ProcID src;
     T2_AccMsgRecvBuffer* msg;

#    ifdef T2_ASSERTION
	bool q =
#    endif
     qenv->getAccMsgRecvBuffer(j, src, msg);
     t2_Assert (q == true);

     t2_Assert (! wlist->isFull());

#    ifdef T2_ASSERTION
        bool r =
#    endif
     wlist->enqueueAccMsg(src, msg, qenv);
     t2_Assert (r == true);
  }

  t2_say5(cout, "in T2_GlobalCombineWork::handler(), nrecvbufs = " 
		<< nrecvbufs << endl);

  // tell comm about the sends and recvs
  t2_Assert (comm != NULL);
  t2_Assert (nrecvbufs <= numnodes);
  comm->registerISends(*qenv);
  // since comm_manager does not decrement the number of expected recvs
  // when an acc msg arrives before the query reaches the global combine
  // phase, therefore only need to tell comm_manager to expect
  // (numnodes-nrecvbufs) more messages
  comm->incNumberExpectedRecvs(numnodes - nrecvbufs); 	// only this many more
							// recvs are expected
}



/////////////////////////////////////////////////
//
// methods for T2_AccMsgWork
//
/////////////////////////////////////////////////
void
T2_AccMsgWork::handler()
{
  t2_say4(cout, "T2_AccMsgWork::handler for qenv "
		<< (int) qenv << " is invoked" << endl);

  t2_Assert (msg != NULL);
  t2_Assert (qenv != NULL);
  t2_Assert (rts != NULL);

  if (qenv->isValid() == false) {
     delete msg;				// release acc msg buffer
     t2_Assert ((msg = NULL) == NULL);
     return;
  }


  T2_AggregateFuncObj* aggr = qenv->getAggregateFuncObject();
  t2_Assert (aggr != NULL);
  T2_Accumulator *acc = qenv->getAccumulator();
  t2_Assert (acc != NULL);

  T2_UDFRet rv = T2_UDFRet_OK;
  {
     T2_AccMsgBufferReader reader(*msg);

     t2_say(cout, "acc msg from src = " << src << " has real data = "
		  << reader.getDataSize() << " bytes, msg=" << (u_int) msg
		  << endl);

     if (reader.getDataSize() > 0) {		// not an empty message

#	if (T2_TIME_QUERY > 1)
	{
	   T2_QQPlan* qqp = qenv->getQQPlan();
	   t2_Assert (qqp != NULL);
	   t2_Assert (qqp->getNumberSubplans() > 0);
	   T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
	   double tsnap = t2_dclock();
#	endif

	rv = aggr->processAccMsg(qenv->getQueryInfo(), src, reader, *acc);

#	if (T2_TIME_QUERY > 1)
	   qtimer.incProcessAccMsgTime(t2_dclock() - tsnap);
	}
#	endif

     }			// if (reader.getDataSize() != 0)
  }

  delete msg;					// release acc msg buffer
  t2_Assert ((msg = NULL) == NULL);
  
  if (rv != T2_UDFRet_OK) {			// error, so kill query
     // update expected number of recvs in T2_CommManager
     t2_Assert (comm != NULL);
     comm->decNumberExpectedRecvsWhenKillQuery(qenv->getQueryStatus());
     const T2_System& sys = rts->getSystem();
     comm->killRemoteQuery(qenv->getQEnvID(),
			   sys.getNumberNodes(), sys.getMyNode());

     // kill query
     qenv->killQueryNow();
     return;
  }

  if (qenv->decNumberAccMsgsYetProcessAndCheckEndGlobalCombine() == true) {
     t2_Assert (wlist != NULL);
     qenv->finalizeGlobalCombine();
     t2_Assert (qenv->getQueryStatus().getState() == t2_qPostGlobal);
#    ifdef T2_ASSERTION
        bool r = 
#    endif
     wlist->enqueueOutputHandle(qenv);
     t2_Assert (r == true);
  }

  t2_say4(cout, "T2_AccMsgWork for query " << qenv->getQEnvID()
		<< " was fired, # acc_msg_yet_process = "
		<< qenv->getQueryStatus().getNumberAccMsgsYetProcess() << endl);
}



/////////////////////////////////////////////////
//
// methods for T2_OutputHandleWork
//
/////////////////////////////////////////////////
void
T2_OutputHandleWork::handler()
{
  t2_Assert (qenv != NULL);

# ifdef T2_TIME_QUERY
     T2_QQPlan* qqp = qenv->getQQPlan();
     t2_Assert (qqp != NULL);
     t2_Assert (qqp->getNumberSubplans() > 0);
     T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
     qtimer.setStartOHandle();
# endif

  t2_say4(cout,  " ******* T2_OutputHandleWork:: handler for qenv "
		<< (int) qenv << " is fired" << endl);

  if (qenv->isValid() == false)
     return;

  T2_QStatus& qstatus = qenv->getQueryStatus();
  T2_OutputEnv& outenv = qenv->getOutputEnv();

  t2_Assert (rts != NULL);

# ifdef T2_ASSERTION
     T2_QState s = 
# endif
  qstatus.getState();
  t2_AssertOrDo (s == t2_qLocal || s == t2_qPostGlobal,
		 cout << __FILE__ << " at line " << __LINE__
		      << ": s = " << (int) s << endl);

  // finalize output
  qenv->finalizeOutput();
  T2_Output* output = qenv->getOutput();
  t2_Assert (output != NULL);
  T2_Accumulator* acc = qenv->getAccumulator();
  t2_Assert (acc != NULL);

  const T2_QQPlan* qqplan = qenv->getQQPlan();
  t2_Assert (qqplan != NULL);

  if (outenv.handleOutput(*output, *acc, qqplan->getCurrentIteration())
      == false)				// output handling is done
  {

     t2_say4(cout, "outputhandling work is done, status = " << 
		    (int) (qstatus.getState()) << endl);

     if (qenv->endOutputHandle() == true) { // output handle phase has ended
	t2_Assert (wlist != NULL);
#	ifdef T2_ASSERTION
	   bool r = 
#	endif
	wlist->enqueueEndIteration(qenv);
	t2_Assert (r == true);
     }
  }
}



/////////////////////////////////////////////////
//
// methods for T2_EndIterationWork
//
/////////////////////////////////////////////////
void
T2_EndIterationWork::handler()
{
  t2_Assert (qenv != NULL);

# ifdef T2_TIME_QUERY
     T2_QQPlan* qqp = qenv->getQQPlan();
     t2_Assert (qqp != NULL);
     t2_Assert (qqp->getNumberSubplans() > 0);
     T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
     qtimer.setStartEndIteration();
# endif

  t2_say2(cout,  "T2_EndIterationWork::handler for qenv " << (int) qenv
		<< " is invoked" << endl);

  if (qenv->isValid() == false)		// invalid query, do no more
     return;

# ifdef T2_ASSERTION
     T2_QState s = 
# endif
  qenv->getQueryStatus().getState();
  t2_Assert (s == t2_qLocal || s == t2_qGlobal || s == t2_qOutput);

  // release memory in qenv
  qenv->finalizeIteration();

  // advance cur_iter and reset cur_cluster in T2_QQPlan
  T2_QQPlan *qqplan = qenv->getQQPlan();
  t2_Assert (qqplan != NULL);
  t2_Assert (qqplan->getCurrentIteration() < qqplan->getNumberSubplans());
  t2_Assert (qqplan->getCurrentCluster() ==
             qqplan->getSubplan(qqplan->getCurrentIteration()).
                     getNumberClusterRequests());
  qqplan->advanceCurrentIteration();

  t2_Assert (qenv->getQueryStatus().setState(t2_qEndIteration) ==
	     t2_qEndIteration);

  if (qqplan->allIterationsDone() == true) {	// query completely done
     qenv->initQueryDone();
     T2_QGlobalStatus& global = qenv->getQueryStatus().getGlobalStatus();
     t2_Assert (global.getNumberQueriesNotDone() > 0);
     global.decNumberQueriesNotDone();

#    ifdef T2_TIME_QUERY
        qtimer.setEndEndIteration();
#    endif

     return;
  }


# ifdef T2_TIME_QUERY
     qtimer.setEndEndIteration();

     // get the timer for the new iteration
     T2_QueryIterTimer& qtimer_next = qqplan->getCurrentSubplan();
     qtimer_next.setStartInitIteration();	// start of initIteration()
# endif


  // query not done, start the next iteration
  t2_Assert (qenv->getQueryStatus().setState(t2_qIdle) == t2_qIdle);
  t2_Assert (wlist != NULL);
  t2_Assert (rts != NULL);

  if (qenv->initIteration(*wlist) == false) {	// error, kill query
     t2_Assert (comm != NULL);
     const T2_System& sys = rts->getSystem();
     comm->killRemoteQuery(qenv->getQEnvID(),
			   sys.getNumberNodes(), sys.getMyNode());
     qenv->killQueryNow();
     return;
  }

# ifdef T2_TIME_QUERY
     qtimer_next.setEndInitIteration();		// end of initIteration()
# endif
}



/////////////////////////////////////////////////
//
// methods for T2_WorkList
//
/////////////////////////////////////////////////
void
T2_WorkList::init(T2_RTS* r, T2_CommManager* c)
{
  t2_Assert (r != NULL);
  t2_Assert (T2_WorkEntry::rts == NULL);
  T2_WorkEntry::rts = r;

  t2_Assert (T2_WorkEntry::wlist == NULL);
  T2_WorkEntry::wlist = this;

  t2_Assert (c != NULL);
  t2_Assert (T2_WorkEntry::comm == NULL);
  T2_WorkEntry::comm = c;
}

