
/* $Id: t2_charptr.h,v 1.1 2000/02/23 22:08:07 chialin Exp $ */

/* t2_charptr.h */

#ifndef _T2_CHARPTR_H_
#define _T2_CHARPTR_H_

#include	<string.h>
#include	"t2_header.h"

///////////////////////////////////////////////////////////////////////////
//
// T2_CharPtr
//
// Just a wrapper for char*. This class is best used in place of char*
// when one needs an array of char* in the form of T2_Array<char*> or
// T2_VArray<char*>. This is because T2_CharPtr has constructor to
// initialize the pointer for NULL, and destructor to free the memory
// it points to. Other than that, it is exactly the same as char*.
//
///////////////////////////////////////////////////////////////////////////
class T2_CharPtr {
public:
  T2_CharPtr(): ptr(NULL) {};
  ~T2_CharPtr() { delete [] ptr; };

  // operators
  operator char*& () { return ptr; };
  operator char* const& () const { return ptr; };
  char& operator [] (int i) { return ptr[i]; };

  // copy constructor and operator =
  T2_CharPtr(const T2_CharPtr& c);
  T2_CharPtr& operator =(const T2_CharPtr& c);

  char* ptr;
};


inline
T2_CharPtr::T2_CharPtr(const T2_CharPtr& c)
{
  if (c.ptr == NULL)
     ptr = NULL;
  else {
     size_t len = strlen(c.ptr) + 1;
     t2_Alloc(ptr, len, char);
     memcpy(ptr, c.ptr, len);
  }
}

inline
T2_CharPtr&
T2_CharPtr::operator =(const T2_CharPtr& c)
{
  delete [] ptr;
  if (c.ptr == NULL)
     ptr = NULL;
  else {
     size_t len = strlen(c.ptr) + 1;
     t2_Alloc(ptr, len, char);
     memcpy(ptr, c.ptr, len);
  }
  return *this;
}



#endif	/* ! _T2_CHARPTR_H_ */
