
/* t2_chunkreq.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/9/98
 * a chunk request
 */


#ifndef _T2_CHUNKREQ_H_
#define _T2_CHUNKREQ_H_


#include	"t2_header.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_ChunkRequest
//
// A read request for a contiguous chunk.
//
/////////////////////////////////////////////////////////////////////////
class T2_ChunkRequest {
public:
  T2_ChunkRequest(): fid(0), offset(0), nbytes(0) {};
  T2_ChunkRequest(T2_DSFileID i,off_t f,size_t s): fid(i),offset(f),nbytes(s){};
  T2_ChunkRequest(const T2_ChunkRequest& r);

  // const access methods
  T2_DSFileID getFileID() const { return fid; };
  off_t getOffset() const { return offset; };
  size_t getNumberBytes() const { return nbytes; };

  // non-const access methods
  T2_DSFileID& getFileID() { return fid; };
  off_t& getOffset() { return offset; };
  size_t& getNumberBytes() { return nbytes; };

  T2_ChunkRequest& operator =(const T2_ChunkRequest& r);

protected:
  T2_DSFileID fid;			// data file id, as in the catalog
  off_t offset;				// offset into the data file
  size_t nbytes;			// number of bytes to read

private:
};


inline
T2_ChunkRequest::T2_ChunkRequest(const T2_ChunkRequest& r):
			fid(r.fid), offset(r.offset), nbytes(r.nbytes)
{
}

inline T2_ChunkRequest&
T2_ChunkRequest::operator =(const T2_ChunkRequest& r)
{
  fid = r.fid;
  offset = r.offset;
  nbytes = r.nbytes;
  return *this;
}



#endif	/* _T2_CHUNKREQ_H_ */
