
/* $Id: t2_dsreg.h,v 1.3 2000/02/23 22:08:07 chialin Exp $ */

/* t2_dsreg.h */

#ifndef _T2_DSREG_H_
#define _T2_DSREG_H_


#include	<iostream.h>
#include	<fstream.h>
#include	<strings.h>
#include	<string.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_txtfile.h"
#include	"t2_idrange.h"
#include	"t2_charptr.h"


///////////////////////////////////////////////////////////////////////////
//
// class T2_NameIDPair
//
// A class for a pair of a name and an id.
//
///////////////////////////////////////////////////////////////////////////
class T2_NameIDPair {
public:
  T2_NameIDPair(): name(NULL), id(0) {};
  ~T2_NameIDPair() { delete [] name; };

  char *name;
  u_int id;

  // do not use copy constructor and operator =
  T2_NameIDPair(const T2_NameIDPair&);
  T2_NameIDPair& operator =(const T2_NameIDPair&);
};


inline
T2_NameIDPair::T2_NameIDPair(const T2_NameIDPair& p)
			: name(NULL), id(0)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_NameIDPair copy constructor");
}

inline
T2_NameIDPair&
T2_NameIDPair::operator =(const T2_NameIDPair& p)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_NameIDPair copy constructor");
  return *this;
}



///////////////////////////////////////////////////////////////////////////
//
// class T2_NameIDRangePair
//
// A class for a pair of name and a list of ranges of id's.
//
///////////////////////////////////////////////////////////////////////////
class T2_NameIDRangePair {
public:
  T2_NameIDRangePair(): name(NULL), ranges(8) {};
  ~T2_NameIDRangePair() { delete [] name; };

  char *name;
  T2_VArray<T2_IDRange> ranges;

  // do not use copy constructor and operator =
  T2_NameIDRangePair(const T2_NameIDRangePair&);
  T2_NameIDRangePair& operator =(const T2_NameIDRangePair&);
};


inline
T2_NameIDRangePair::T2_NameIDRangePair(const T2_NameIDRangePair& p)
			: name(NULL), ranges(0)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_NameIDRangePair copy constructor");
}

inline
T2_NameIDRangePair&
T2_NameIDRangePair::operator =(const T2_NameIDRangePair& p)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_NameIDRangePair copy constructor");
  return *this;
}



///////////////////////////////////////////////////////////////////////////
//
// class T2_IndexRegistrationEntry
//
///////////////////////////////////////////////////////////////////////////
class T2_IndexRegistrationEntry {
public:
  T2_IndexRegistrationEntry();
  ~T2_IndexRegistrationEntry();

  char *index_name,
       *index_constructor_name,
       *description;
  T2_VArray<T2_NameIDRangePair> idxfiles;	// a list of index files

  // do not use copy constructor and operator =
  T2_IndexRegistrationEntry(const T2_IndexRegistrationEntry&);
  T2_IndexRegistrationEntry& operator =(const T2_IndexRegistrationEntry&);

protected:
private:
};


inline
T2_IndexRegistrationEntry::T2_IndexRegistrationEntry()
		: index_name(NULL), index_constructor_name(NULL),
		  description(NULL), idxfiles(5)
{
}

inline
T2_IndexRegistrationEntry::~T2_IndexRegistrationEntry()
{
  delete [] index_name;
  delete [] index_constructor_name;
  delete [] description;
}

inline
T2_IndexRegistrationEntry::
T2_IndexRegistrationEntry(const T2_IndexRegistrationEntry& e)
		: index_name(NULL), index_constructor_name(NULL),
		  description(NULL), idxfiles(5)
{
  t2_err_quit(__FILE__, __LINE__,
		"do not use T2_IndexRegistrationEntry copy constructor");
}

inline
T2_IndexRegistrationEntry&
T2_IndexRegistrationEntry::operator =(const T2_IndexRegistrationEntry& e)
{
  t2_err_quit(__FILE__, __LINE__,
		"do not use T2_IndexRegistrationEntry operator =");
  return *this;
}



///////////////////////////////////////////////////////////////////////////
//
// class T2_DSRegistrationEntry
//
///////////////////////////////////////////////////////////////////////////
class T2_DSRegistrationEntry {
public:
  T2_DSRegistrationEntry();
  ~T2_DSRegistrationEntry();

  char *dataset_name,			// name for dataset
       *dsconstructor_name,		// name for dataset constructor
       *blob_fname,			// filename for associated blob
       *description;			// a one-line description

  T2_VArray<T2_NameIDPair> datafiles;	// data filenames and assigned nodes
  T2_VArray<T2_CharPtr> auxfiles;	// aux filenames
  T2_VArray<T2_IndexRegistrationEntry> indexes;	// index entries
  
  // read from a file
  bool read(T2_TxtFile& f);

  // check if the content is correct
  bool check_content_correct(u_int min_invalid_log_did) const;

  // clean up all entries
  void clear();

  // output to ostream
  ostream& write(ostream& os) const;

protected:
  // allocate space and copy a string into another string
  static void str_alloc_cpy(char*& dest, const char* src);

  // read the part for the iterators, data files, aux files,
  // optionally allowing/disallowing file redirection
  bool
  read_iterator_datafile_auxfile_part(T2_TxtFile& dfile, const char*& string,
				      bool disable_file_redirect);

  // read details about data files from the given file;
  // return true if no error is encountered;
  bool read_dataset_details_file(const char* fname);

  // read a single record for a data file
  bool read_data_file_line(T2_TxtFile& dfile, u_int lineno);

  // methods similar to read_dataset_details_file() and 
  // read_iterator_datafile_auxfile_part(), but for indexes;
  // when returns, read_index_details would leave the last string that is 
  // not an index data file in the argument 'string'
  bool read_index_details_file(const char* fname,
			       T2_VArray<T2_NameIDRangePair>& files);
  bool read_index_details(T2_TxtFile& ifile,
			  T2_VArray<T2_NameIDRangePair>& files,
			  const char*& string, bool disable_file_redirect);

private:
};


inline
T2_DSRegistrationEntry::T2_DSRegistrationEntry()
			: dataset_name (NULL), dsconstructor_name (NULL),
			  blob_fname(NULL), description (NULL),
			  datafiles(32), auxfiles(5), indexes(3)
{
}

inline
T2_DSRegistrationEntry::~T2_DSRegistrationEntry()
{
  delete [] dataset_name;
  delete [] dsconstructor_name;
  delete [] blob_fname;
  delete [] description;
}

inline void
T2_DSRegistrationEntry::clear()
{
  delete [] dataset_name;
  delete [] dsconstructor_name;
  delete [] blob_fname;
  delete [] description;
  dataset_name = NULL;
  dsconstructor_name = NULL;
  blob_fname = NULL;
  description = NULL;

  datafiles.clear();
  auxfiles.clear();
  indexes.clear();
}

inline void
T2_DSRegistrationEntry::str_alloc_cpy(char*& dest, const char* src)
{
  delete [] dest;
  size_t len = strlen(src) + 1;
  t2_Alloc(dest, len, char);
  memcpy(dest, src, len);
}

inline ostream&
operator << (ostream& os, const T2_DSRegistrationEntry& e)
{
  return e.write(os);
}



///////////////////////////////////////////////////////////////////////////
//
// T2_DSRegistrationFileReader
//
// A class that reads a dataset registration flie.
//
///////////////////////////////////////////////////////////////////////////
class T2_DSRegistrationFileReader {
public:
  T2_DSRegistrationFileReader(const char* fname= NULL);
  ~T2_DSRegistrationFileReader() { close(); };

  // open and close file
  void open(const char*);
  void close() { if (file) file.close(); };

  // query methods
  operator const void*() const { return (const void *) file; };

  // get next entry from file;
  // if eof, returns NULL and set eflag to 0
  // if error, returns NULL and set eflag to 1
  bool getNextEntry(T2_DSRegistrationEntry& e);

  // return true if s is a end-of-record marker
  static bool check_eor_mark(const char* s);

protected:
  T2_TxtFile file;

private:
};



#endif	/* ! _T2_DSREG_H_ */
