
/* t2_error.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/22/97
 * The header file for error handlers, assertion verifier and memory allocation
 * verifier.
 */

#ifndef _T2_ERROR_H_
#define _T2_ERROR_H_


/* include some stardard header files */
#include	<stdio.h>
#include	<stdarg.h>
/* #include	<varargs.h> */
#include	<assert.h>


/*
 * some definitions for the C functions
 */
#ifdef __cplusplus
   extern "C" {
#endif


#ifdef _ANSI_PROTO_
   void	t2_err_quit(char* filename, int lineno, ...),
	t2_err_sys(char* filename, int lineno, ...),
	t2_err_ret(char* filename, int lineno, ...),
	t2_err_dump(char* filename, int lineno, ...),
	/* t2_perror(char* filename, int lineno, ...), */
	t2_perror(),
	t2_halt (char*, char*, int);
   char *t2_search_env(char *envp[], const char *target);
   char *t2_sys_err_str();

#else	/* _ANSI_PROTO_ */

   void	t2_err_quit(),
	t2_err_sys(),
	t2_err_ret(),
	t2_err_dump(),
	t2_perror(),
	t2_halt();
   char *t2_search_env();
   char *t2_sys_err_str();

#endif /* _ANSI_PROTO_ */

#ifdef __cplusplus
   }
#endif



/*
 * assertion and memory allocation macros
 * including: t2_Assert, t2_AssertOrDo, t2_Alloc
 */
#ifdef T2_ASSERTION

   /*
    * assertions
    */
   #define t2_Assert(c_) \
		do {\
		   if (!(c_)) t2_halt("t2_Assert", __FILE__, __LINE__);\
		} while (0)
   #define t2_AssertOrDo(c_, a_) \
		do {\
		   if (!(c_)) {a_; t2_halt("t2_AssertOrDo",__FILE__,__LINE__);}\
		} while (0)

   /*
    * memory allocation
    */
   #ifdef __cplusplus

	# define t2_Alloc(ptr_, num_, type_)\
		do {\
		   if ((ptr_ = new type_[num_]) == NULL)\
		      t2_halt("t2_Alloc", __FILE__, __LINE__);\
		} while (0)

   #else	/* __cplusplus */

	# define t2_Alloc(ptr_, num_, type_)\
		do {\
		   if ((ptr_ = (type_ *)calloc(num_, sizeof(type_))) == NULL)\
		      t2_halt("t2_Alloc", __FILE__, __LINE__);\
		} while (0)

   #endif	/* __cplusplus */

#else	/* T2_ASSERTION */

   /*
    * assertions
    */
   # define t2_Assert(c_)		for(;0;)
   # define t2_AssertOrDo(c_, a_)	for(;0;)

   /*
    * memory allocations
    */
   #ifdef __cplusplus

	# define t2_Alloc(ptr_, num_, type_)	ptr_ = new type_[num_]

   #else	/* __cplusplus */

	# define t2_Alloc(ptr_, num_, type_)\
		ptr_ = (type_ *)calloc(num_, sizeof(type_))

   #endif	/* __cplusplus */

#endif	/* T2_ASSERTION */


#endif	/* _T2_ERROR_H_ */
