
/* t2_tqueue.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/2/98
 * Template classes for queues.
 */

#ifndef _T2_TQUEUE_H_
#define _T2_TQUEUE_H_

#include	<iostream.h>
#include	"t2_header.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_TQueue
// A template for a queue of objects, of type Type.
// The element objects must have a constructor that takes no arguments,
// a copy constructor, an assignment operator (ie operator =), and
// can be written out to ostream (ie operator<<(ostream&, const Type&).
//
/////////////////////////////////////////////////////////////////////////
template <class Type>
class T2_TQueue {
 private:
   u_int NEXT(u_int i) const { t2_Assert(len != 0); return ((i+1) % len); };

   // do not use copy constructor and operator =
   T2_TQueue(const T2_TQueue<Type>& q);
   T2_TQueue<Type>& operator=(const T2_TQueue<Type>& t);

 protected:
   u_int head, tail,	// head and tail of the queue
         len;		// total length of array data[]
   Type *data;

 public:
   T2_TQueue(u_int l = 0);
   ~T2_TQueue() { delete [] data; };

   // query methods
   bool isFull() const { return (head == NEXT(tail))? true:false; };
   bool isEmpty() const { return (head == tail)? true:false; };
   u_int getMaxQueueSize() const { return len-1; };
   u_int getNumberElements() const
	{ return (tail>=head)? tail-head : tail+len-head; };

   // access methods
   const Type& operator[](u_int i) const;

   // modify the queue
   Type dequeue();
   bool enqueue(const Type& p);

   // change the max size of the queue and discard all elements
   void setMaxQueueSize(u_int n);

   ostream& write(ostream& os) const;
};


template <class Type>
inline
ostream& operator << (ostream& os, const T2_TQueue<Type>& tq)
{ return tq.write(os); }


template <class Type>
inline
T2_TQueue<Type>::T2_TQueue(u_int l) : head(0), tail(0), len(l+1)
{ t2_Alloc(data, len, Type); }


template <class Type>
inline
T2_TQueue<Type>::T2_TQueue(const T2_TQueue<Type>& q)
				: head(q.head), tail(q.tail), len(q.len)
{
  t2_Alloc(data, len, Type);
  for (u_int i=head; i!=tail; i=NEXT(i))
     data[i] = q.data[i];
  t2_err_quit(__FILE__,__LINE__,"do not use T2_TQueue copy constructor");
}

template <class Type>
inline T2_TQueue<Type>&
T2_TQueue<Type>::operator=(const T2_TQueue<Type>& t)
{
  t2_err_quit(__FILE__,__LINE__,"do not use T2_TQueue operator=");
  return *this;
}

template <class Type>
inline const Type&
T2_TQueue<Type>::operator[](u_int i) const
{
  t2_Assert(i<getNumberElements());
  u_int k = head + i;
  return data[(k<len)? k: k-len];
}

template <class Type>
inline Type
T2_TQueue<Type>::dequeue()
{
  t2_Assert (!isEmpty());
  u_int i = head;
  head = NEXT(head);
  return data[i];
}

template <class Type>
inline bool
T2_TQueue<Type>::enqueue(const Type& p)
{
  if (isFull())
     return false;
  data[tail] = p;
  tail = NEXT(tail);
  return true;
}

template <class Type>
inline void
T2_TQueue<Type>::setMaxQueueSize(u_int n)
{
  delete [] data;
  len = n + 1;
  head = tail = 0;
  t2_Alloc(data, len, Type);
}

template <class Type>
inline ostream&
T2_TQueue<Type>::write(ostream& os) const
{
  os << "FQueueInt :- head = " << head << ",  tail = " << tail;
  os << endl;
  return os;
}


#endif	/* _T2_TQUEUE_H_ */
