
/* t2_tsetbstr.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/8/98
 * inputing/outputing a tset from/into a binary string
 */


#ifndef _T2_TSETBSTR_H_
#define _T2_TSETBSTR_H_


#include	"t2_header.h"
#include	"t2_bstr.h"
#include	"t2_tset.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_TSetBStrWriter and T2_TSetBStrReader
//
// A template for outputing and inputing a T2_TSet. These sets work
// together. The class OutType is assumed to have a bunch of operator <<
// and the class InType is assumed to have a bunch of opertor >> .
// The element class is assumed to have the appropriate operator << and
// operator >> as well.
//
/////////////////////////////////////////////////////////////////////////
template <class Type>
class T2_TSetBStrWriter {
public:
  T2_TSetBStrWriter(const T2_TSet<Type>& s): tset(s) {};

  // output method
  T2_FOBStr& bwrite(T2_FOBStr& os) const;

protected:
  const T2_TSet<Type>& tset;

private:
};

template <class Type>
class T2_TSetBStrReader {
public:
  T2_TSetBStrReader(T2_TSet<Type>& s): tset(s) {};

  // input methods
  // bread creates exactly the same set that was written into T2_FOBStr
  // (ie with the original length, which could be larger than the number of
  // elements in the set), while breadCompact only allocates space
  // enough to hold the set (ie tset.ndata always equal to tset.len after
  // breadCompact invoked)
  T2_FIBStr& bread(T2_FIBStr& is);
  T2_FIBStr& breadCompact(T2_FIBStr& is);

protected:
  T2_TSet<Type>& tset;

private:
};


template <class Type>
inline T2_FOBStr&
T2_TSetBStrWriter<Type>::bwrite(T2_FOBStr& os) const
{
  os << tset.len << tset.ndata;
  for (u_int i=0; i<tset.ndata; i++)
     os << tset[i];
  return os;
}

template <class Type>
inline T2_FIBStr&
T2_TSetBStrReader<Type>::bread(T2_FIBStr& is)
{
  t2_Assert(is);
  u_int l;
  is >> l;
  tset.emptySet();
  tset.growMaxSetSize(l);
  is >> tset.ndata;
  for (u_int i=0; i<tset.ndata; i++)
     is >> tset.data[i];
  return is;
}

template <class Type>
inline T2_FIBStr&
T2_TSetBStrReader<Type>::breadCompact(T2_FIBStr& is)
{
  t2_Assert(is);
  u_int l;
  is >> l >> l;				// throw away the array length
  tset.emptySet();
  tset.growMaxSetSize(l);
  tset.ndata = l;
  for (u_int i=0; i<tset.ndata; i++)
     is >> tset.data[i];
  return is;
}


#endif	/* _T2_TSETBSTR_H_ */
