
/* $Id: t2_tshape.h,v 1.11 2000/02/04 00:18:26 chialin Exp $ */

/* t2_tshape.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/2/98
 * Adapted from part of the Titan code written by Cater T. Shock for
 * multi-dimensional space
 */


#ifndef _T2_TSHAPE_H_
#define _T2_TSHAPE_H_

#include	<iostream.h>
#include	"t2_header.h"
#include	"t2_template.h"
#include	"t2_tset.h"
#include	"t2_tsetbstr.h"
#include	"t2_tarray.h"
#include	"t2_bstr.h"
#include	"t2_bstream.h"

#ifdef X11
#include	"XGraphics.H"
#endif


// forward declarations
template <class Type> class T2_TBox;
template <class Type> class T2_TRegion;

////////////////////////////////////////////////////////////
//
// T2_Shape
//
// Virtual base class for shapes.
//
////////////////////////////////////////////////////////////
class T2_Shape {
 public:
  T2_Shape() {};
  T2_Shape(const T2_Shape&) {};
  virtual ~T2_Shape() {};
  virtual u_int getNumberDimensions() const = 0;
  virtual ostream & print(ostream &) const = 0;

  bool compatible(const T2_Shape& s) const;

#ifdef _T2_BSTREAM_H_
  virtual T2_OBStream &  bwrite(T2_OBStream &) const = 0;
  virtual T2_IBStream &  bread(T2_IBStream &) = 0;
#endif

#ifdef _T2_BSTR_H_
  virtual T2_FOBStr& bwrite(T2_FOBStr& s) const = 0;
  virtual T2_FIBStr& bread(T2_FIBStr& s) = 0;
#endif

#ifdef X11
  virtual void draw(const GWorld &gw) const = 0;
#endif

};


inline bool
T2_Shape::compatible(const T2_Shape& s) const
{
  return (getNumberDimensions() == s.getNumberDimensions())? true:false;
}

inline ostream&
operator << (ostream &os, const T2_Shape& s)
{  return s.print(os); }



////////////////////////////////////////////////////////////
//
// T2_TPoint
//
////////////////////////////////////////////////////////////
template <class Type>
class T2_TPoint : public T2_Shape {
public:
  T2_TPoint(u_int d=0): T2_Shape(), coord(d) {};
//  T2_TPoint(const u_int d, ...);		// d is # dimensions
  T2_TPoint(const T2_Array<Type>& c);
  T2_TPoint(const u_int l, const Type* c);
  T2_TPoint(const T2_TPoint<Type>& pt);

  // virtual methods
  u_int getNumberDimensions() const { return coord.getNumberElements(); };
  ostream & print(ostream &) const;

  // access methods
  const Type& operator[](const u_int i) const;
  Type& operator[](const u_int i);

  // change the number of dimensions of the point
 void setNumberDimensions(u_int d) { coord.resize(d); };

  // move up and right
  T2_TPoint<Type>& offset(const T2_TPoint<Type>& p);

  // some operators
  T2_TPoint<Type>& operator = (const T2_TPoint<Type> &p);
  bool operator == (const T2_TPoint<Type>& p) const;
  bool operator != (const T2_TPoint<Type>& p) const;
  bool operator ^(const T2_TPoint<Type>& p) const { return (*this == p); };

#ifdef _T2_BSTREAM_H_
  virtual T2_OBStream &  bwrite(T2_OBStream &) const;
  virtual T2_IBStream &  bread(T2_IBStream &);
#endif

#ifdef _T2_BSTR_H_
  virtual T2_FOBStr& bwrite(T2_FOBStr& s) const;
  virtual T2_FIBStr& bread(T2_FIBStr& s);
#endif

#ifdef X11
  virtual void draw(const GWorld &gw) const;
#endif

protected:
  T2_Array<Type> coord;

private:
};



//
// this does not work well when Type is other than int or float
//
//template <class Type>
//inline
//T2_TPoint<Type>::T2_TPoint(const u_int d, ...): T2_Shape(), coord(d)
//{
//  va_list ap;
//  va_start(ap, d);
//  for (u_int i=0; i<d; i++)
//     coord[i] = va_arg(ap, Type);
//  va_end(ap);
//}


template <class Type>
inline
T2_TPoint<Type>::T2_TPoint(const T2_Array<Type>& c): T2_Shape(), coord(c)
{
}

template <class Type>
inline
T2_TPoint<Type>::T2_TPoint(const u_int l, const Type* c): T2_Shape(), coord(l)
{
  t2_Assert(l == 0 || c != NULL);
  for (u_int i=0; i<l; i++)
    coord[i] = c[i];
}

template <class Type>
inline
T2_TPoint<Type>::T2_TPoint(const T2_TPoint<Type>& pt):
		T2_Shape(pt), coord(pt.getNumberDimensions())
{
  for (u_int i=0; i<pt.getNumberDimensions(); i++)
    coord[i] = pt[i];
}

template <class Type>
inline ostream &
T2_TPoint<Type>::print(ostream & os) const
{

  os << "T2_TPoint<";
  if (getNumberDimensions() > 0) {
     const u_int last = getNumberDimensions()-1;
     for (u_int i=0; i<last; i++)
        os << coord[i] << ", ";
     os << coord[last];
  }
  os << ">";
  return os;
}

template <class Type>
inline const Type&
T2_TPoint<Type>::operator[](const u_int i) const
{
  t2_Assert(i<getNumberDimensions());
  return coord[i];
}

template <class Type>
inline Type&
T2_TPoint<Type>::operator[](const u_int i)
{
  t2_AssertOrDo (i<getNumberDimensions(),
	 cout << "i=" << i << ", # dim = " << getNumberDimensions() << endl);
  return coord[i];
}

template <class Type>
inline T2_TPoint<Type>&
T2_TPoint<Type>::offset(const T2_TPoint<Type>& p)
{
  t2_Assert(compatible(p));
  for (u_int i=0; i<getNumberDimensions(); i++)
     (*this)[i] += p[i];
  return *this;
}

template <class Type>
inline T2_TPoint<Type>&
T2_TPoint<Type>::operator = (const T2_TPoint<Type> &p)
{
  t2_Assert(getNumberDimensions() == 0 || compatible(p));
  coord = p.coord;
  return *this;
}

template <class Type>
inline bool
T2_TPoint<Type>::operator == (const T2_TPoint<Type> &p) const
{
  t2_Assert(compatible(p));
  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (coord[i] != p[i])
        return false;
  }
  return true;
}

template <class Type>
inline bool
T2_TPoint<Type>::operator != (const T2_TPoint<Type> &p) const
{
  t2_Assert(compatible(p));
  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (coord[i] != p[i])
        return true;
  }
  return false;
}


#ifdef _T2_BSTREAM_H_

template <class Type>
inline T2_OBStream&
T2_TPoint<Type>::bwrite(T2_OBStream & obf) const
{
  const u_int n = coord.getNumberElements();
  obf << n;
  for (u_int i=0; i<n; i++)
     obf << coord[i];
  return obf;
}

template <class Type>
inline T2_IBStream&
T2_TPoint<Type>::bread(T2_IBStream & ibf)
{
  t2_Assert (ibf);
  u_int n;
  ibf >> n;
  coord.clear(n);
  for (u_int i=0; i<n; i++)
     ibf >> coord[i];
  return ibf;
}

template <class Type>
inline T2_OBStream&
operator <<(T2_OBStream & obf, const T2_TPoint<Type>& p)
{
  return p.bwrite(obf);
}

template <class Type>
inline T2_IBStream&
operator >>(T2_IBStream & ibf, T2_TPoint<Type>& p)
{
  return p.bread(ibf);
}

#endif  /* _T2_BSTREAM_H_ */


#ifdef _T2_BSTR_H_

template <class Type>
inline T2_FOBStr&
T2_TPoint<Type>::bwrite(T2_FOBStr & obs) const
{
  const u_int n = coord.getNumberElements();
  obs << n;
  for (u_int i=0; i<n; i++)
     obs << coord[i];
  return obs;
}

template <class Type>
inline T2_FIBStr&
T2_TPoint<Type>::bread(T2_FIBStr & ibs)
{
  t2_Assert (ibs);
  u_int n;
  ibs >> n;
  coord.clear(n);
  for (u_int i=0; i<n; i++)
     ibs >> coord[i];
  return ibs;
}

template <class Type>
inline T2_FOBStr&
operator <<(T2_FOBStr & obs, const T2_TPoint<Type>& p)
{
  return p.bwrite(obs);
}

template <class Type>
inline T2_FIBStr&
operator >>(T2_FIBStr & ibs, T2_TPoint<Type>& p)
{
  return p.bread(ibs);
}

#endif  /* _T2_BSTR_H_ */





////////////////////////////////////////////////////////////
//
// T2_TBox
//
// A class for a multi-dimensional box, which represents
// set of ranges, one along each dimension. It is represented
// by two points, the low point and the high point.
//
////////////////////////////////////////////////////////////
template <class Type>
class T2_TBox : public T2_Shape {
public:
  T2_TBox(u_int d=0): T2_Shape(), lowpt(d), highpt(d) {};
  T2_TBox(const T2_TPoint<Type>& l, const T2_TPoint<Type>& h);
  T2_TBox(const T2_TBox<Type>& b);
  T2_TBox(const T2_TPoint<Type>& p);

  // virtual methods
  u_int getNumberDimensions() const { return lowpt.getNumberDimensions(); };
  ostream & print(ostream &) const;

  // set number of dimensions
  void setNumberDimensions(u_int d);

  // access methods
  const T2_TPoint<Type>& getLow() const { return lowpt; };
  const T2_TPoint<Type>& getHigh() const { return highpt; };
  T2_TPoint<Type>& getLow() { return lowpt; };
  T2_TPoint<Type>& getHigh() { return highpt; };
  void setLow(const T2_TPoint<Type>& p) { lowpt = p; };
  void setHigh(const T2_TPoint<Type>& p) { highpt = p; };

  // returns true if entire ranges of b are bounded by *this
  bool contains(const T2_TBox<Type>& b) const;

  // some operators
  T2_TBox<Type> & operator = (const T2_TPoint<Type>& p);
  T2_TBox<Type> & operator = (const T2_TBox<Type>& b);
  bool operator == (const T2_TBox<Type>& b) const;
  bool operator != (const T2_TBox<Type>& b) const;
  bool operator ^(const T2_TPoint<Type>& p) const;	// test intersection
  bool operator ^(const T2_TBox<Type>& b) const;	// test intersection

  // make sure that lowpt is always smaller than highpt
  bool valid() const;


#ifdef _T2_BSTREAM_H_
  virtual T2_OBStream &  bwrite(T2_OBStream &) const;
  virtual T2_IBStream &  bread(T2_IBStream &);
#endif

#ifdef _T2_BSTR_H_
  virtual T2_FOBStr& bwrite(T2_FOBStr& s) const;
  virtual T2_FIBStr& bread(T2_FIBStr& s);
#endif

#ifdef X11
  virtual void draw(const GWorld &gw) const;
#endif

protected:
  T2_TPoint<Type> lowpt, highpt;

private:
};


template <class Type>
inline
T2_TBox<Type>::T2_TBox(const T2_TPoint<Type>& l, const T2_TPoint<Type>& h)
			: T2_Shape(), lowpt(l), highpt(h)
{
  t2_Assert(valid());
}

template <class Type>
inline
T2_TBox<Type>::T2_TBox(const T2_TBox<Type>& b)
			: T2_Shape(b), lowpt(b.lowpt), highpt(b.highpt)
{
  t2_Assert(valid());
}

template <class Type>
inline
T2_TBox<Type>::T2_TBox(const T2_TPoint<Type>& p)
			: T2_Shape(), lowpt(p), highpt(p)
{
}

template <class Type>
inline ostream &
T2_TBox<Type>::print(ostream & os) const
{
  os << "T2_TBox: low = " << lowpt << ", high = " << highpt;
  return os;
}

template <class Type>
inline void
T2_TBox<Type>::setNumberDimensions(u_int d)
{
  lowpt.setNumberDimensions(d);
  highpt.setNumberDimensions(d);
}

template <class Type>
inline bool
T2_TBox<Type>::contains(const T2_TBox<Type> &b) const
{
  t2_Assert (compatible(b));

  const T2_TPoint<Type>& l = b.getLow();
  const T2_TPoint<Type>& h = b.getHigh();
  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (! t2_between(lowpt[i], l[i], highpt[i]) ||
         ! t2_between(lowpt[i], h[i], highpt[i]))
        return false;
  }
  return true;
}

template <class Type>
inline T2_TBox<Type>&
T2_TBox<Type>::operator = (const T2_TPoint<Type>& p)
{
  t2_Assert (getNumberDimensions() == 0 || compatible(p));
  lowpt = p;
  highpt = p;
  return *this;
}

template <class Type>
inline T2_TBox<Type>&
T2_TBox<Type>::operator = (const T2_TBox<Type>& b)
{
  t2_Assert (getNumberDimensions() == 0 || compatible(b));
  lowpt = b.lowpt;
  highpt = b.highpt;
  return *this;
}

template <class Type>
inline bool
T2_TBox<Type>::operator == (const T2_TBox<Type>& b) const
{
  t2_Assert (compatible(b));
  return (lowpt == b.lowpt && highpt == b.highpt)? true : false;
}

template <class Type>
inline bool
T2_TBox<Type>::operator != (const T2_TBox<Type>& b) const
{
  t2_Assert (compatible(b));
  return (lowpt != b.lowpt || highpt != b.highpt)? true : false;
}

template <class Type>
inline bool
T2_TBox<Type>::operator ^(const T2_TPoint<Type>& p) const
{
  t2_Assert (compatible(p));
  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (! t2_between(lowpt[i], p[i], highpt[i])) {
        return false;
     }
  }
  return true;
}

template <class Type>
inline bool
T2_TBox<Type>::operator ^(const T2_TBox<Type>& b) const
{
  t2_Assert (compatible(b));

  const T2_TPoint<Type>& l = b.getLow();
  const T2_TPoint<Type>& h = b.getHigh();
  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (! t2_between(lowpt[i], l[i], highpt[i]) &&
         ! t2_between(lowpt[i], h[i], highpt[i]) &&
         ! t2_between(l[i], lowpt[i], h[i]) &&
         ! t2_between(l[i], highpt[i], h[i]))
        return false;
  }
  return true;
}

template <class Type>
inline bool
T2_TBox<Type>::valid() const
{
  t2_Assert (lowpt.compatible(highpt));
  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (lowpt[i] > highpt[i])
        return false;
  }
  return true;
}


#ifdef _T2_BSTREAM_H_

template <class Type>
inline T2_OBStream&
T2_TBox<Type>::bwrite(T2_OBStream & obf) const
{
  obf << lowpt << highpt;
  return obf;
}

template <class Type>
inline T2_IBStream&
T2_TBox<Type>::bread(T2_IBStream & ibf)
{
  t2_Assert (ibf);
  ibf >> lowpt >> highpt;
  return ibf;
}

template <class Type>
inline T2_OBStream&
operator <<(T2_OBStream & obf, const T2_TBox<Type> b)
{
  return b.bwrite(obf);
}

template <class Type>
inline T2_IBStream&
operator >>(T2_IBStream & ibf, T2_TBox<Type>& b)
{
  return b.bread(ibf);
}

#endif  /* _T2_BSTREAM_H_ */


#ifdef _T2_BSTR_H_

template <class Type>
inline T2_FOBStr&
T2_TBox<Type>::bwrite(T2_FOBStr & obs) const
{
  obs << lowpt << highpt;
  return obs;
}

template <class Type>
inline T2_FIBStr&
T2_TBox<Type>::bread(T2_FIBStr & ibs)
{
  t2_Assert (ibs);
  ibs >> lowpt >> highpt;
  return ibs;
}

template <class Type>
inline T2_FOBStr&
operator <<(T2_FOBStr & obs, const T2_TBox<Type> b)
{
  return b.bwrite(obs);
}

template <class Type>
inline T2_FIBStr&
operator >>(T2_FIBStr & ibs, T2_TBox<Type>& b)
{
  return b.bread(ibs);
}

#endif  /* _T2_BSTR_H_ */





////////////////////////////////////////////////////////////
//
// T2_TRegion
//
// A class for a list of multi-dimensional boxes, each of
// which represents a set of ranges, one along each dimension.
// A region is effectively a list of point-pairs, which
// usually are interpretated as a list of lower- and higher-
// bounds.
//
////////////////////////////////////////////////////////////
template <class Type>
class T2_TRegion : public T2_Shape {
public:
  T2_TRegion(u_int d=0, u_int l=0);
  T2_TRegion(const T2_TSet< T2_TBox<Type> >& b);
  T2_TRegion(const u_int l, const T2_TBox<Type>* b);
  T2_TRegion(const T2_TRegion<Type>& b);
  T2_TRegion(const T2_TPoint<Type>& p);
  T2_TRegion(const T2_TBox<Type>& b);

  // virtual methods
  u_int getNumberDimensions() const { return ndim; };
  ostream & print(ostream &) const;

  // query methods
  u_int getNumberBoxes() const { return boxes.getNumberElements(); };
  u_int getMaxSetSize() const { return boxes.getMaxSetSize(); };
  bool isFull() const { return (getMaxSetSize()==getNumberBoxes())?true:false;};

  // access methods
  const T2_TBox<Type>& operator[](u_int i) const;
  T2_TBox<Type>& operator[](u_int i);

  // set number of dimensions
  void setNumberDimensions(u_int d);

  // change the max set size of the region
  void growMaxSetSize(u_int l) { boxes.growMaxSetSize(l); };

  // add a box or a point (which will be treated as a box) into the region
  T2_TRegion<Type>& operator +=(const T2_TPoint<Type>& p);
  T2_TRegion<Type>& operator <<(const T2_TPoint<Type>& p) { return (*this)+=p;};
  T2_TRegion<Type>& operator +=(const T2_TBox<Type>& b);
  T2_TRegion<Type>& operator <<(const T2_TBox<Type>& b) { return (*this)+=b; };

  // returns true if entire box of b are bounded by *this
  bool contains(const T2_TPoint<Type>& p) const;
  bool contains(const T2_TBox<Type>& b) const;
  bool contains(const T2_TRegion<Type>& r) const;

  // some operators
  T2_TRegion<Type> & operator = (const T2_TPoint<Type>& p);
  T2_TRegion<Type> & operator = (const T2_TBox<Type>& b);
  T2_TRegion<Type> & operator = (const T2_TRegion<Type>& r);
  bool operator == (const T2_TRegion<Type>& b) const;
  bool operator != (const T2_TRegion<Type>& b) const;
  bool operator ^(const T2_TPoint<Type>& p) const;	// test intersection
  bool operator ^(const T2_TBox<Type>& b) const;	// test intersection
//  bool operator ^(const T2_TRegion<Type>& r) const;	// test intersection

  // get a bounding box for the entire region
  void getMBR(T2_TBox<Type>& box) const;

#ifdef _T2_BSTREAM_H_
  virtual T2_OBStream &  bwrite(T2_OBStream &) const;
  virtual T2_IBStream &  bread(T2_IBStream &);
#endif

#ifdef _T2_BSTR_H_
  virtual T2_FOBStr& bwrite(T2_FOBStr& s) const;
  virtual T2_FIBStr& bread(T2_FIBStr& s);
#endif

#ifdef X11
  virtual void draw(const GWorld &gw) const;
#endif

protected:
  u_int ndim;
  T2_TSet < T2_TBox<Type> > boxes;

private:
};


template <class Type>
inline
T2_TRegion<Type>::T2_TRegion(u_int d, u_int l): T2_Shape(), ndim(d), boxes(l)
{
  for (u_int i=0; i<l; i++) {
     boxes.getPosition(i).setNumberDimensions(ndim);
  }
}

template <class Type>
inline
T2_TRegion<Type>::T2_TRegion(const T2_TSet< T2_TBox<Type> >& b)
		: T2_Shape(), boxes(b.getNumberElements())
{
  const u_int n = b.getNumberElements();
  if (n == 0) {
     ndim = 0;
     return;
  }

  ndim = b[0].getNumberDimensions();
  for (u_int i=0; i<n; i++) {
     boxes[i].setNumberDimensions(ndim);
     t2_Assert (b[i].getNumberDimensions() == ndim);
     boxes += b[i];
  }
}

template <class Type>
inline
T2_TRegion<Type>::T2_TRegion(const u_int l, const T2_TBox<Type>* b)
			    : T2_Shape(), boxes(l)
{
  t2_Assert (l == 0 || b != NULL);
  if (l == 0) {
     ndim = 0;
     return;
  }

  ndim = b[0].getNumberDimensions();
  for (u_int i=0; i<l; i++) {
     boxes[i].setNumberDimensions(ndim);
     t2_Assert (b[i].getNumberDimensions() == ndim);
     boxes += b[i];
  }
}

template <class Type>
inline
T2_TRegion<Type>::T2_TRegion(const T2_TRegion<Type>& b)
				: T2_Shape(b), boxes(b.getNumberBoxes())
{
  ndim = b.getNumberDimensions();
  for (u_int i=0; i<b.getNumberBoxes(); i++) {
     boxes[i].setNumberDimensions(ndim);
     t2_Assert (b[i].getNumberDimensions() == ndim);
     boxes += b[i];
  }
}

template <class Type>
inline
T2_TRegion<Type>::T2_TRegion(const T2_TPoint<Type>& p): T2_Shape(), boxes(1)
{
  ndim = p.getNumberDimensions();
  boxes += T2_TBox<Type>(p);
}

template <class Type>
inline
T2_TRegion<Type>::T2_TRegion(const T2_TBox<Type>& b): T2_Shape(), boxes(1)
{
  ndim = b.getNumberDimensions();
  boxes += b;
}

template <class Type>
inline ostream &
T2_TRegion<Type>::print(ostream& os) const
{
  if (getNumberBoxes() == 0) {
     os << "T2_TRegion = <>" << endl;
     return os;
  }
  os << "T2_TRegion = " << getNumberBoxes() << " boxes";
  for (u_int i=0; i<getNumberBoxes(); i++)
     os << endl << "     " << boxes[i];
  return os;
}

template <class Type>
inline const T2_TBox<Type>&
T2_TRegion<Type>::operator[](u_int i) const
{
  t2_Assert(i < getNumberBoxes());
  return boxes[i];
}

template <class Type>
inline T2_TBox<Type>&
T2_TRegion<Type>::operator[](u_int i)
{
  t2_Assert(i < getNumberBoxes());
  return boxes[i];
}

template <class Type>
inline void
T2_TRegion<Type>::setNumberDimensions(u_int d)
{
  ndim = d;
  for (u_int i=0; i<boxes.getMaxSetSize(); i++)
     boxes.getPosition(i).setNumberDimensions(d);

  // not sure what to do with the elements in the set; empty the set for now
  boxes.emptySet();
}

template <class Type>
inline T2_TRegion<Type>&
T2_TRegion<Type>::operator +=(const T2_TPoint<Type>& p)
{
  t2_Assert (compatible(p));
  if (! boxes.isFull()) {
     boxes += T2_TBox<Type>(p);
  }
  return *this;
}

template <class Type>
inline T2_TRegion<Type>&
T2_TRegion<Type>::operator +=(const T2_TBox<Type>& b)
{
  t2_Assert (compatible(b));
  if (! boxes.isFull()) {
     boxes += b;
  }
  return *this;
}

template <class Type>
inline bool
T2_TRegion<Type>::contains(const T2_TPoint<Type>& p) const
{
  t2_Assert(compatible(p));
  for (u_int i=0; i<getNumberBoxes(); i++) {
     if (boxes[i].contains(p))
        return true;
  }
  return false;
}

template <class Type>
inline bool
T2_TRegion<Type>::contains(const T2_TBox<Type>& b) const
{
  // the following is a simple implementation that assume all boxes in *this
  // are disjoint -- that is, this implementation does not consider the
  // possibility that box b could be wholly contained by boxes[1] and boxes[2],
  // but not by either one

  t2_Assert(compatible(b));
  for (u_int i=0; i<getNumberBoxes(); i++) {
     if (boxes[i].contains(b))
        return true;
  }
  return false;
}

template <class Type>
inline bool
T2_TRegion<Type>::contains(const T2_TRegion<Type>& r) const
{
  t2_Assert(compatible(r));

  const u_int n = getNumberBoxes();
  for (u_int j=0; j<r.getNumberBoxes(); j++) {
     const T2_TBox<Type>& b = r[j];
     u_int i;
     for (i=0; i<n; i++) {
        if (boxes[i].contains(b))
           break;
     }
     if (i == n)	// r[j] is not contained by any box of *this
        return false;
  }
  return true;
}

template <class Type>
inline T2_TRegion<Type>&
T2_TRegion<Type>::operator = (const T2_TPoint<Type>& p)
{
  t2_Assert(getNumberDimensions() == 0 || compatible(p));
  ndim = p.getNumberDimensions();
  boxes.emptySet();
  if (getMaxSetSize() < 1) {
     // not enough space
     boxes.growMaxSetSize(1);
  }
  u_int i;
  bool r = boxes.appendElement(i);
  t2_Assert (r == true);
  boxes[i] = p;
  return *this;
}

template <class Type>
inline T2_TRegion<Type>&
T2_TRegion<Type>::operator = (const T2_TBox<Type>& b)
{
  t2_Assert(getNumberDimensions() == 0 || compatible(b));
  ndim = b.getNumberDimensions();
  boxes.emptySet();
  if (getMaxSetSize() < 1) {
     // not enough space
     boxes.growMaxSetSize(1);
  }
  boxes << b;
  return *this;
}

template <class Type>
inline T2_TRegion<Type>&
T2_TRegion<Type>::operator = (const T2_TRegion<Type>& r)
{
  t2_Assert(getNumberDimensions() == 0 || compatible(r));
  ndim = r.ndim;
  const u_int n = r.getNumberBoxes();
  boxes.emptySet();
  if (getMaxSetSize() < n) {
     // not enough space
     boxes.growMaxSetSize(n);
  }
  for (u_int i=0; i<n; i++)
     boxes << r.boxes[i];
  return *this;
}

template <class Type>
inline bool
T2_TRegion<Type>::operator == (const T2_TRegion<Type>& r) const
{
  // this is a simple implementation that checks the boxes along two lists
  // without reordering; that is, it checks to see if the two lists are
  // identical

  t2_Assert(compatible(r));

  if (getNumberBoxes() != r.getNumberBoxes())
     return false;

  for (u_int i=0; i<getNumberBoxes(); i++) {
     if (boxes[i] != r[i])
        return false;
  }
  return true;
}

template <class Type>
inline bool
T2_TRegion<Type>::operator != (const T2_TRegion<Type>& r) const
{
  // this is a simple implementation that checks the boxes along two lists
  // without reordering; that is, it checks to see if the two lists are
  // identical

  t2_Assert(compatible(r));

  if (getNumberBoxes() != r.getNumberBoxes())
     return true;

  for (u_int i=0; i<getNumberBoxes(); i++) {
     if (boxes[i] != r[i])
        return true;
  }
  return false;
}

template <class Type>
inline bool
T2_TRegion<Type>::operator ^(const T2_TPoint<Type>& p) const
{
  t2_Assert (compatible(p));
  for (u_int i=0; i<getNumberBoxes(); i++) {
     if (boxes[i] ^ p)
        return true;
  }
  return false;
}

template <class Type>
inline bool
T2_TRegion<Type>::operator ^(const T2_TBox<Type>& b) const
{
  t2_Assert (compatible(b));
  for (u_int i=0; i<getNumberBoxes(); i++) {
     if (boxes[i] ^ b)
        return true;
  }
  return false;
}

//template <class Type>
//inline bool
//T2_TRegion<Type>::operator ^(const T2_TRegion<Type>& r) const

template <class Type>
inline void
T2_TRegion<Type>::getMBR(T2_TBox<Type>& box) const
{
  box.setNumberDimensions(ndim);
  T2_TPoint<Type>& low = box.getLow();
  T2_TPoint<Type>& high = box.getHigh();

  const u_int nboxes = getNumberBoxes();
  if (nboxes == 0)
     return;

  for (u_int j=0; j<ndim; j++) {
     low[j] = (boxes[0].getLow())[j];
     high[j] = (boxes[0].getHigh())[j];
  }
  for (u_int i=1; i<nboxes; i++) {
     const T2_TBox<Type>& b = boxes[i];
     for (u_int d=0; d<ndim; d++) {
	low[d] = t2_min(low[d], (b.getLow())[d]);
	high[d] = t2_max(high[d], (b.getHigh())[d]);
     }
  }
}


#ifdef _T2_BSTREAM_H_

template <class Type>
inline T2_OBStream&
T2_TRegion<Type>::bwrite(T2_OBStream & obf) const
{
  obf << ndim;
  T2_TSetBStreamWriter< T2_TBox<Type> > boxwriter(boxes);
  boxwriter.bwrite(obf);
  return obf;
}

template <class Type>
inline T2_IBStream&
T2_TRegion<Type>::bread(T2_IBStream & ibf)
{
  t2_Assert (ibf);
  T2_TSet < T2_TBox<Type> > boxes;
  ibf >> ndim;
  T2_TSetBStreamReader< T2_TBox<Type> > boxreader(boxes);
  boxreader.breadCompact(ibf);
  return ibf;
}

template <class Type>
inline T2_OBStream&
operator <<(T2_OBStream & obf, const T2_TRegion<Type> r)
{
  return r.bwrite(obf);
}

template <class Type>
inline T2_IBStream&
operator >>(T2_IBStream & ibf, T2_TRegion<Type>& r)
{
  return r.bread(ibf);
}

#endif	/* _T2_BSTREAM_H_ */


#ifdef _T2_BSTR_H_

template <class Type>
inline T2_FOBStr&
T2_TRegion<Type>::bwrite(T2_FOBStr & obs) const
{
  obs << ndim;
  T2_TSetBStrWriter< T2_TBox<Type> > boxwriter(boxes);
  boxwriter.bwrite(obs);
  return obs;
}

template <class Type>
inline T2_FIBStr&
T2_TRegion<Type>::bread(T2_FIBStr & ibs)
{
  t2_Assert (ibs);
  T2_TSet < T2_TBox<Type> > boxes;
  ibs >> ndim;
  T2_TSetBStrReader< T2_TBox<Type> > boxreader(boxes);
  boxreader.breadCompact(ibs);
  return ibs;
}

template <class Type>
inline T2_FOBStr&
operator <<(T2_FOBStr & obs, const T2_TRegion<Type> r)
{
  return r.bwrite(obs);
}

template <class Type>
inline T2_FIBStr&
operator >>(T2_FIBStr & ibs, T2_TRegion<Type>& r)
{
  return r.bread(ibs);
}

#endif  /* _T2_BSTR_H_ */




#endif	/* _T2_TSHAPE_H_ */
