
/* $Id: bfile.C,v 1.6 2000/03/16 00:48:11 chialin Exp $ */

/* bfile.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/15/98
 * Adapted from the code in Titan written by Carter T. Shock
 */

#include	<iostream.h>
#include	<unistd.h>
#include	<strings.h>
#include	<fcntl.h>
#include	"t2_bfile.h"
#include	"t2_template.h"


// extern void perror(char *s);
extern int errno;


int T2_BFile::dflt_ib_flg = O_RDONLY;
int T2_BFile::dflt_ob_flg = O_WRONLY | O_CREAT | O_TRUNC;


//////////////////////////////////////////////////////////////////////////////
//
// T2_BFile
//
//////////////////////////////////////////////////////////////////////////////
T2_BFile::T2_BFile(const char *name, int flags, int mode)
{
  fd = open(name,flags,mode);
  errval = errno;
  errno = 0;
  attached = false;
}

T2_BFile & T2_BFile::bopen(const char *name, int flags, int mode)
{
  bclose();

  fd = mode ? open(name,flags,mode) : open(name,flags);
  errval = errno;
  errno = 0;
  attached = false;
  return *this;
}

T2_BFile & T2_BFile::attach(int f)
{
  bclose();

  if (f >= 0) {
     fd = f;
     errval = 0;
     errno = 0;
     attached = true;
  }
  else
     fd = -1;
  return *this;
}

int T2_BFile::bclose()
{
  if(fd >= 0 && !attached) {
    close(fd);
    errval = errno;
    errno = 0;
  }
  else
    errval = 0;

  fd = -1;
  attached = false;
  
  return errval;
}

int T2_BFile::getflags()
{
  int val;
  val = fcntl(fd,F_GETFL,0);
  errval = errno;
  errno = 0;
  return val;
}

int T2_BFile::setflags(int flags)
{
  int val;
  val = fcntl(fd,F_SETFL,flags);
  errval = errno;
  errno = 0;
  return val;
}

int T2_BFile::berr() const
{
//  perror("Error in File.C");
  return errval;
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_IBFile
//
//////////////////////////////////////////////////////////////////////////////
ssize_t T2_IBFile::bread(void *buf, int count, size_t size)
{
  size_t totsize = count * size;
  if (cur + totsize <= len) {           // fbuf has what's needed
     bcopy(fbuf+cur, buf, totsize);
     cur += totsize;
     return totsize;
  }

  // put into buf whatever is left in fbuf
  size_t remains = len - cur;
  if (remains > 0)
     bcopy(fbuf+cur, buf, remains);

  char *bufptr = (char *) buf + remains;
  size_t readsize = totsize - remains;	// this is how much more is needed

  // hmm...., must read from file now
  if (readsize >= T2_IBFile_buflen) {
     // need to read more than what can be held in fbuf, so simply
     // read the data directly into buf with one read
     ssize_t val = read(fd, bufptr, readsize);
     errval = errno;
     errno = 0;

     len = cur = 0;
     if (val < 0) {			// read failed
        return val;
     }
     pos += val;
     if ((u_int) val < readsize)	// reach eof
        saweof = true;
     return (remains + val);
  }

  // fill fbuf with one read and then copy data into buf from fbuf
  ssize_t val = read(fd, fbuf, T2_IBFile_buflen);
  errval = errno;
  errno = 0;

  if (val < 0) {			// read failed
     len = cur = 0;
     return val;
  }

  len = val;				// val bytes were read
  pos += val;
  if ((u_int) val < T2_IBFile_buflen)
     saweof = true;

  if ((u_int) val < readsize) {		// not enough bytes were read
     bcopy(fbuf, bufptr, val);		// give buf all that were read
     len = cur = 0;
     return (val + remains);
  }

  bcopy(fbuf, bufptr, readsize);	// more than enough were read
  cur = readsize;
  return totsize;
     


//  cur = readsize;
//
//  // need to read another chunk into fbuf
//  size_t remains = len - cur;
//  if (remains > 0)
//     bcopy(fbuf+cur, fbuf, remains);
//  size_t readsize = T2_IBFile_buflen - remains;
//  ssize_t val = read(fd, fbuf+remains, readsize);
//  errval = errno;
//  errno = 0;
//
//  if (val < 0) {			// read failed
//     len = remains;
//  }
//  else {				// read succeeded
//     len = remains + val;
//     pos += val;
//     if (val < readsize)		// saw eof
//	saweof = true;
//  }
//
//  cur = 0;
//  if ((errval == 0) && len < totsize)
//     errval = -1;
//  if (val < 0)
//     return val;
//
//  size_t s = (len < totsize)? len : totsize;
//  bcopy(fbuf, buf, s);
//  cur += totsize;
//  return s;
}

off_t T2_IBFile::bseek(off_t offset, int whence)
{
  t2_Assert (whence == SEEK_SET || whence == SEEK_CUR || whence == SEEK_END);

  if (noFile())		// no file is open
     return -1;

  // For now, I'll just throw away whatever has been read into the
  // buffer, seek to the appropriate position and set len to zero, except
  // in the case where whence is SEEK_CUR and the seeked position is still
  // within the buffer. In this case, I'll just move cur accordingly.

  if (whence == SEEK_CUR) {
     off_t newpos = cur + offset;
     if (t2_between((off_t) 0, newpos, (off_t) len)) {
        // good, I don't need to move cur outside the current buffer
        cur = newpos;
        return curToOffset();
     }

     // the seeked position is outside the buffer, need to adjust offset
     if (newpos < 0) {
        // need to seek backward, so must decrement offset even more by len
        offset = newpos - len;
     }
     else {				// newpos must be > len
        // need to seek forward, so must decrement offset by len
        offset = newpos - len;
     }
  }

  // throw away the buffer and move the file pointer by offset
  len = cur = 0;
  saweof = false;
  pos = lseek(fd, offset, whence);
  errval = errno;
  errno = 0;
  return pos;
}


bool
T2_IBFile::eof() const
{
  if (saweof == true && cur == len)
     return true;
  return false;
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_OBFile
//
//////////////////////////////////////////////////////////////////////////////
ssize_t T2_OBFile::bwrite(const void *buf, int count, size_t size)
{
  if (size > 1 && T2_OBStream::convertEndian() == true) {
     char *bb = (char *) buf;
     t2_swapBytesArray(bb, count, size);
  }

  ssize_t val = write(fd, buf, count*size);
  errval = errno;
  errno = 0;

  if (size > 1 && T2_OBStream::convertEndian() == true) {
     char *bb = (char *) buf;
     t2_swapBytesArray(bb, count, size);
  }
  return val;
}

off_t T2_OBFile::bseek(off_t offset, int whence)
{
  t2_Assert (whence == SEEK_SET || whence == SEEK_CUR || whence == SEEK_END);

  if (noFile())		// no file is open
     return -1;

  if(fd >= 0) {
    off_t val = lseek(fd, offset, whence);
    errval = errno;
    errno = 0;
    return val;
  }
  else
    return -1;
}

off_t T2_OBFile::btell()
{
  return lseek(fd, (off_t)0, SEEK_CUR);
}
