
/* $Id: utility.c,v 1.19 2000/09/22 02:11:24 chialin Exp $ */

/* utility.c */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/22/97
 * Error handling routines.
 *
 * These functions were adapted from the examples used in "UNIX Network
 * Programming", written by R. Stevens
 *
 * The functions in this file are independent of any application
 * variables, and may be used with any C program.
 * The "CLIENT" version of the code (that is, the version that output to stderr)
 * are used.
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<strings.h>
#include	<ctype.h>
#include	<sys/time.h>
#include	<sys/resource.h>
#include	<errno.h>
#include	"t2_header.h"


static char *pname = NULL;
/*
static char *pname = "T2";
*/


/* global var to store local processor number, mostly for printing debugging 
 * information
 */
int _g_mynode = -1;


/* these error handlers all output to stderr */

/*
 * Fatal error.  Print a message and terminate.
 * Don't dump core and don't print the system's errno value.
 *
 *	err_quit(filename, linenumber, str, arg1, arg2, ...)
 *
 * The string "str" must specify the conversion specification for any args.
 */

/*VARARGS1*/
void
t2_err_quit(char *fname, int lineno, ...)
{
	va_list		args;
	char		*fmt;

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	if (fname != NULL)
	   fprintf(stderr, "quit at FILE %s, LINE %d\n", fname,lineno);

	va_start(args, lineno);
	if (pname != NULL)
		fprintf(stderr, "%s: ", pname);

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	fmt = va_arg(args, char *);
	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);
	va_end(args);

	fprintf(stderr, "Quit.\n");
	exit(1);
}

/*
 * Fatal error related to a system call.  Print a message and terminate.
 * Don't dump core, but do print the system's errno value and its
 * associated message.
 *
 *	err_sys(filename, linenumber, str, arg1, arg2, ...)
 *
 * The string "str" must specify the conversion specification for any args.
 */

/*VARARGS1*/
void
t2_err_sys(char *fname, int lineno, ...)
{
	va_list		args;
	char		*fmt;

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	if (fname != NULL)
	   fprintf(stderr, "error at FILE %s, LINE %d\n", fname,lineno);

	va_start(args, lineno);
	if (pname != NULL)
		fprintf(stderr, "%s: ", pname);

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	fmt = va_arg(args, char *);
	vfprintf(stderr, fmt, args);
	va_end(args);

	t2_perror();

	fprintf(stderr, "Quit.\n");
	exit(1);
}

/*
 * Recoverable error.  Print a message, and return to caller.
 *
 *	err_ret(filename, linenumber, str, arg1, arg2, ...)
 *
 * The string "str" must specify the conversion specification for any args.
 */

/*VARARGS1*/
void
t2_err_ret(char *fname, int lineno, ...)
{
	va_list		args;
	char		*fmt;

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	if (fname != NULL)
	   fprintf(stderr, "at FILE %s, LINE %d\n", fname,lineno);

	va_start(args, lineno);
	if (pname != NULL)
	   fprintf(stderr, "%s: ", pname);

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	fmt = va_arg(args, char *);
	vfprintf(stderr, fmt, args);
	va_end(args);

/*
	t2_perror();
*/
	fprintf(stderr, " \n");
	errno = 0;

	fflush(stdout);
	fflush(stderr);

	return;
}

/*
 * Fatal error.  Print a message, dump core (for debugging) and terminate.
 *
 *	err_dump(filename, linenumber, str, arg1, arg2, ...)
 *
 * The string "str" must specify the conversion specification for any args.
 */

/*VARARGS1*/
void
t2_err_dump(char *fname, int lineno, ...)
{
	va_list		args;
	char		*fmt;

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	if (fname != NULL)
	   fprintf(stderr, "error at FILE %s, LINE %d\n", fname,lineno);

	va_start(args, lineno);
	if (pname != NULL)
		fprintf(stderr, "%s: ", pname);

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
	fprintf(stderr, "%d: ", _g_mynode);
#endif

	fmt = va_arg(args, char *);
	vfprintf(stderr, fmt, args);
	va_end(args);

	t2_perror();

	fflush(stdout);		/* abort doesn't flush stdio buffers */
	fflush(stderr);

/*	fprintf(stderr, "Abort.\n"); */
	abort();		/* dump core and terminate */
	exit(1);		/* shouldn't get here */
}

/*
 * Print the UNIX errno value.
 */

void
t2_perror()
{
	char	*t2_sys_err_str();

	fprintf(stderr, " %s\n", t2_sys_err_str());
}


extern int	errno;		/* Unix error number */
extern int	sys_nerr;	/* # of error message strings in sys table */

#ifndef T2_SYS_DEFINE_SYS_ERRLIST
   extern char	*sys_errlist[];	/* the system error message table */
#endif

#ifdef	SYS5
   int	t_errno;	/* in case caller is using TLI, these are "tentative
			   definitions"; else they're "definitions" */
   int	t_nerr;
   char	*t_errlist[1];
#endif


/*
 * Return a string containing some additional operating-system
 * dependent information.
 * Note that different versions of UNIX assign different meanings
 * to the same value of "errno" (compare errno's starting with 35
 * between System V and BSD, for example).  This means that if an error
 * condition is being sent to another UNIX system, we must interpret
 * the errno value on the system that generated the error, and not
 * just send the decimal value of errno to the other system.
 */

char *
t2_sys_err_str()
{
	static char	msgstr[200];

	if (errno != 0) {
		if (errno > 0 && errno < sys_nerr)
			sprintf(msgstr, "(%s)", sys_errlist[errno]);
		else
			sprintf(msgstr, "(errno = %d)", errno);
	} else {
		msgstr[0] = '\0';
	}

#ifdef	SYS5
	if (t_errno != 0) {
		char	tmsgstr[100];

		if (t_errno > 0 && t_errno < sys_nerr)
			sprintf(tmsgstr, " (%s)", t_errlist[t_errno]);
		else
			sprintf(tmsgstr, ", (t_errno = %d)", t_errno);

		strcat(msgstr, tmsgstr);	/* catenate strings */
	}
#endif

	return(msgstr);
}


/*
 * Halt and report the program execution.
 */
void
t2_halt(msg, fname, lineno)
char* msg;
char* fname;
int lineno;
{
  FILE *fp;
  fp = fopen("HALT", "w");
  if (fp) {

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
     fprintf(fp, "%d: ", _g_mynode);
#endif

     fprintf(fp, "HALT: %s failed in %s at %d\n", msg, fname, lineno);
     fclose(fp);
  }

#if (defined T2_LABEL_IO && defined T2_PBACKEND)
  fprintf(stderr, "%d: HALT: %s failed in %s at %d\n",
		  _g_mynode, msg, fname, lineno);
#else
  fprintf(stderr, "HALT: %s failed in %s at %d\n", msg, fname, lineno);
#endif

  fflush(stdout);
  fflush(stderr);

  abort();		/* dump core and terminate */
  exit(1);
}


/*
 * Search for the value of the environment variable target
 */
char *
t2_search_env(char *envp[], const char *target)
{
  const int tlen = strlen(target);
  int i;

  for (i=0; envp[i] != NULL; i++) {
     if (strncmp(envp[i], target, tlen) == 0) {
        return &(envp[i][tlen+1]);		/* target found */
     }
  }

  return NULL;
}



/*
 * wclock function
 */

#ifndef t2_dclock
double
t2_dclock(void) {
  struct timeval tv;
  struct timezone tz;

  if (gettimeofday(&tv, &tz) != 0) {
    fprintf(stderr, "t2_dclock(): gettimeofday() failed.\n");
    return 0;
  }
  return (double)tv.tv_sec + (double)tv.tv_usec / (double)1000000;
}

#endif



/*
 * searchString function
 */
const char *
searchString(const char* line, u_int start, u_int* next)
{
  /* this function would return the pointer to the 1st string starting from
     line[start], skipping white spaces, and set argument *next to the 1st 
     char after the white space that immediately follows the last char of
     the returned string; returns NULL if no string is found
  */

  const char *p, *q;

  t2_Assert (line != NULL);
  p = line + start;
  while (*p != '\0' && isspace((int) *p))
    p++;

  if (*p == '\0')			/* end of line */
     return NULL;

  q = p+1;

  while (*q != '\0' && !isspace((int) *q))
    q++;

#if 0
  if (*q != '\0') 			/* *q is space */
     *q = '\0';
#endif

  *next = q - line + 1;
  return p;
}




/* function to test for endianess */
u_char t2_checkEndianess()
{ 
   short s=1;  
   char* p = (char *) &s; 
   if (*p == '\0') 
	return (u_char) T2_BigEndian;
   return (u_char) T2_LittleEndian; 
}


/* function to convert an Endian-char to a string */
static const char T2_EndianStrings[4][32] = {
	"unknown endian",
	"big endian",
	"little endian",
	"error"
};

const char* endian2string(u_char c)
{
  switch (c) {
    case T2_UnknownEndian:	return T2_EndianStrings[0]; break;
    case T2_BigEndian:		return T2_EndianStrings[1]; break;
    case T2_LittleEndian:	return T2_EndianStrings[2]; break;
  }
  return T2_EndianStrings[3];		/* error */
}



/* function to convert a string into enum T2_UDFType */
enum T2_UDFType
t2_shortstring2UDFType(const char* name)
{ 
  if (name == NULL || name[0] == '\0')
     return T2_UDF_Unknown;

  switch (name[0]) {
     case 'd':
     case 'D':
		switch (name[1]) {
		   case 'a':
		   case 'A':
				return T2_UDF_Dataset;
		}
		return T2_UDF_Unknown;
     case 'i':
     case 'I':
		switch (name[1]) {
		   case 'n':
		   case 'N':
				return T2_UDF_Index;
		}
		return T2_UDF_Unknown;
     case 'a':
     case 'A':
		switch (name[1]) {
		   case 'c':
		   case 'C':
				return T2_UDF_AccMeta;
		   case 'g':
		   case 'G':
				return T2_UDF_Aggregation;
		}
		return T2_UDF_Unknown;
     case 'p':
     case 'P':
		switch (name[1]) {
		   case 'r':
		   case 'R':
				return T2_UDF_Projection;
		}
		return T2_UDF_Unknown;
     case 'm':
     case 'M':
		switch (name[1]) {
		   case 'e':
		   case 'E':
		   case 'c':
		   case 'C':
				return T2_UDF_MetaChaos;
		}
		return T2_UDF_Unknown;
  }

  return T2_UDF_Unknown;
}


/* function to convert a enum T2_UDFType into a string */
static const char T2_UDFTypeStrings[7][32] = {
	"T2_UDF_Unknown",
	"T2_UDF_Dataset",
	"T2_UDF_Index",
	"T2_UDF_AccMeta",
	"T2_UDF_Aggregation",
	"T2_UDF_Projection",
	"T2_UDF_MetaChaos"
};
const char*
t2_UDFType2string(enum T2_UDFType t)
{
  switch (t) {
    case T2_UDF_Unknown:	return T2_UDFTypeStrings[0];
    case T2_UDF_Dataset:	return T2_UDFTypeStrings[1];
    case T2_UDF_Index:		return T2_UDFTypeStrings[2];
    case T2_UDF_AccMeta:	return T2_UDFTypeStrings[3];
    case T2_UDF_Aggregation:	return T2_UDFTypeStrings[4];
    case T2_UDF_Projection:	return T2_UDFTypeStrings[5];
    case T2_UDF_MetaChaos:	return T2_UDFTypeStrings[6];
  }
  return T2_UDFTypeStrings[0];
}


/* function to convert a enum T2_UDFType into a short string */
static const char T2_UDFTypeShortStrings[7][32] = {
	"unknown",
	"dataset",
	"index",
	"accumulator",
	"aggregation",
	"projection",
	"metachaos"
};
const char*
t2_UDFType2shortstring(enum T2_UDFType t)
{
  switch (t) {
    case T2_UDF_Unknown:	return T2_UDFTypeShortStrings[0];
    case T2_UDF_Dataset:	return T2_UDFTypeShortStrings[1];
    case T2_UDF_Index:		return T2_UDFTypeShortStrings[2];
    case T2_UDF_AccMeta:	return T2_UDFTypeShortStrings[3];
    case T2_UDF_Aggregation:	return T2_UDFTypeShortStrings[4];
    case T2_UDF_Projection:	return T2_UDFTypeShortStrings[5];
    case T2_UDF_MetaChaos:	return T2_UDFTypeShortStrings[6];
  }
  return T2_UDFTypeShortStrings[0];
}



/* functions to wrap around a string (ie char* ):
 * t2_stringNullWrap returns the string NULL when the given pointer is NULL,
 * t2_stringEmptyWrap returns an empty string when the given pointer is NULL
 */
const char T2_StringNULL[] = "NULL",
	   T2_StringEmpty[] = "";
const char*
t2_stringNullWrap(const char* p)
{
  if (p == NULL)
     return T2_StringNULL;
  return p;
}
const char*
t2_stringEmptyWrap(const char* p)
{
  if (p == NULL)
     return T2_StringEmpty;
  return p;
}




/* a wrapper for atoi, the translated value is returned by the function,
 * and the 2nd argument is set to 0 if atoi succeeded, EINVAL (or errno)
 * otherwise
 */
int
t2_atoi(const char* s, int *atoi_err)
{
  int i;
  const char *p;

  if (s == NULL) {
     *atoi_err = EINVAL;
     return 0;
  }

  for (p = s; *p != '\0'; p++) {
     if (*p != '-' && isdigit((int) *p) == 0) {	/* not a number */
	*atoi_err = EINVAL;
	return 0;
     }
  }

  errno = 0;
  i = atoi(s);
  *atoi_err = errno;
  return i;
}
