
/* $Id: svm_acc.C,v 1.2 2000/03/09 01:32:51 chialin Exp $ */

/* svm_acc.C */


#include	<stdio.h>
#include	<iostream.h>
#include	<stdlib.h>
#include	<errno.h>
#include	"svm_acc.h"


/////////////////////////////////////////////////////////////////////
//
// svm_ImageAcc methods
//
/////////////////////////////////////////////////////////////////////
svm_ImageAcc::svm_ImageAcc(u_int tw, u_int th,
			   u_int xp, u_int yp, u_int w, u_int h,
			   u_int s, u_int v)
			: svm_Image(), T2_Accumulator(), x_res(tw), y_res(th),
			  x_pos(xp), y_pos(yp)
{
  svm_Image::getWidth() = w;
  svm_Image::getHeight() = h;
  svm_Image::getPixelSize() = s;
  svm_Image::getMaxValue() = v;
  svm_Image::getBytes() = new u_char[svm_Image::getNumberPixels() *
				     svm_Image::getPixelSize()];
  if (svm_Image::getBytes() == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to allocate %d bytes for accumulator pixels.",
		svm_Image::getNumberPixels() * svm_Image::getPixelSize());
  }
  // I'll leave the pixels un-initialized, since they should be initialized
  // by the aggregation funtion later
}


T2_UDFRet
svm_ImageAcc::navigateAll(const T2_QueryInfo& qinfo, T2_AccIterator*& aitp)
{
  aitp = (T2_AccIterator *)
	new svm_ImageAccIterator(getWidth(), getHeight(),
				 getPixelSize(), getBytes(),
				 0, 0, getWidth()-1, getHeight()-1);
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageAcc::navigate(const T2_QueryInfo& qinfo, T2_IteratorID i,
		       const T2_Region& r, T2_AccIterator*& aitp)
{
  // ignore iterator id, since there is only one iterator

  if (r.getNumberBoxes() != 1) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a region with 1 box, but got one with %d boxes",
	r.getNumberBoxes());
     return T2_UDFRet_ERROR;
  }

  const u_int ndims = 2;
  const T2_Box& box = r[0];
  if (box.getNumberDimensions() != ndims) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a %d-dim region, but got one with %d dims",
	ndims, box.getNumberDimensions());
     return T2_UDFRet_ERROR;
  }
  u_int x_low = (u_int) box.getLow()[0],
	y_low = (u_int) box.getLow()[1],
	x_high = (u_int) box.getHigh()[0],
	y_high = (u_int) box.getHigh()[1];

  t2_Assert (x_high >= x_pos);
  t2_Assert (y_high >= y_pos);

  // need to translate the area between (x_low, y_low) and (x_high, y_high)
  // into local array index inside the chunk for fast iterating;
  // the case where x_end and y_end exceeds the width and height of the image
  // would be taken care of by the constructor of svm_ImageIterator
  u_int x_start = (x_low > x_pos)? (x_low - x_pos) : 0,
        y_start = (y_low > y_pos)? (y_low - y_pos) : 0,
        x_end = x_high - x_pos,
        y_end = y_high - y_pos;

  aitp = (T2_AccIterator *)
	new svm_ImageAccIterator(getWidth(), getHeight(),
				 getPixelSize(), getBytes(),
				 x_start, y_start, x_end, y_end);

  return T2_UDFRet_OK;
}



/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccIterator methods
//
/////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageAccIterator::getNextElement(const T2_QueryInfo& qinfo,
				     T2_Accumulator& acc,
				     bool& eod, void*& elem)
{
  u_int x_off, y_off;			// offsets for pixel to return

  if (svm_ImageIterator::getNextElement(x_off, y_off, pixel) == false) {
     // end-of-data is true
     eod = true;
     return T2_UDFRet_OK;
  }

  elem = (void *) &pixel;
  eod = false;
  return T2_UDFRet_OK;
}



/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccMeta methods
//
/////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageAccMeta::allocAcc(const T2_QueryInfo& qinfo,
			   T2_Iteration i, const T2_Region& tile_mbr,
			   T2_Accumulator*& accp)
{
  if (tile_mbr.getNumberBoxes() != 1) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a single box for the tile's mbr, but got %d boxes instead.",
	tile_mbr.getNumberBoxes());
     return T2_UDFRet_ERROR;
  }

  const u_int ndims = 2;		// can only handle 2-dim boxes
  const T2_Box& mbr = tile_mbr[0];
  if (mbr.getNumberDimensions() != ndims) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a %d-dim tile mbr, but got a %d-dim box instead.",
	ndims, mbr.getNumberDimensions());
     return T2_UDFRet_ERROR;
  }

  const T2_Point& low = mbr.getLow();
  const T2_Point& high = mbr.getHigh();
   
  u_int tile_width = (u_int) high[0] - (u_int) low[0] + 1,
        tile_height = (u_int) high[1] - (u_int) low[1] + 1;

  // for now, assume that we always use rgb (3 bytes) per pixel,
  // and the max value per color component is always 255
  const u_int maxval = 255;
  accp = (T2_Accumulator *)
		new svm_ImageAcc(width, height, (u_int) low[0], (u_int) low[1],
				 tile_width, tile_height, pixelsize, maxval);

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageAccMeta::stripMine(const T2_QueryInfo& qinfo, size_t mem,
			    const T2_ClusterInfoList& infolist,
			    T2_VArray<T2_Region>& tile_mbrs)
{
  if (mem < sizeof(svm_ImageAcc)) {	// this shouldn't happen
     t2_err_ret(__FILE__,__LINE__,
	"ERROR:not enough memory %d to even store an svm_ImageAcc of %d bytes",
	mem, sizeof(svm_ImageAcc));
     return T2_UDFRet_ERROR;
  }

  mem -= sizeof(svm_ImageAcc);	    // reserve space for an svm_ImageAcc obj

  // for simplicity, we are going to partition the image either by rows or
  // by columns, depending on whether the image is tall and skinny or short
  // and fat
  if (width >= height)
     return partitionByCols(mem, tile_mbrs);
  else
     return partitionByRows(mem, tile_mbrs);
}


T2_UDFRet
svm_ImageAccMeta::partitionByRows(size_t mem, T2_VArray<T2_Region>& tile_mbrs)
const
{
  size_t rowsize = width * pixelsize;
  if (mem < rowsize) {			// not enough memory to hold one row
     // give up for now
     t2_err_ret(__FILE__,__LINE__,
	"ERROR: cannot handle the case where there is not enough memory (%d bytes) to store a row of pixels (%d bytes).",
	mem, rowsize);
     return T2_UDFRet_ERROR;
  }

  u_int y_mem = mem / rowsize;		// # rows that fit inside memory
  u_int ntiles = height / y_mem,	// # tiles to generate
	y_excess = height % y_mem;
  if (y_excess > 0)
     ntiles++;

  const u_int ndims = 2;		// only deal with 2-dim tiles

  for (u_int i=0, y_start=0; i<ntiles; i++) {
     T2_Box mbr(ndims);
     T2_Point& low = mbr.getLow();
     T2_Point& high = mbr.getHigh();
     low[0] = 0;
     low[1] = (T2_ShapeCoord_t) y_start;
     high[0] = (T2_ShapeCoord_t) (width - 1);

     u_int y_end = y_start + y_mem - 1;
     if (y_end >= height)
	high[1] = (T2_ShapeCoord_t) (height - 1);
     else
	high[1] = (T2_ShapeCoord_t) y_end;

     tile_mbrs << mbr;

     y_start = y_end + 1;
  }

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageAccMeta::partitionByCols(size_t mem, T2_VArray<T2_Region>& tile_mbrs)
const
{
  size_t colsize = height * pixelsize;
  if (mem < colsize) {			// not enough memory to hold one col
     // give up for now
     t2_err_ret(__FILE__,__LINE__,
	"ERROR: cannot handle the case where there is not enough memory (%d bytes) to store a col of pixels (%d bytes).",
	mem, colsize);
     return T2_UDFRet_ERROR;
  }

  u_int x_mem = mem / colsize;		// # rows that fit inside memory
  u_int ntiles = width / x_mem,		// # tiles to generate
	x_excess = width % x_mem;
  if (x_excess > 0)
     ntiles++;

  const u_int ndims = 2;		// only deal with 2-dim tiles

  for (u_int i=0, x_start=0; i<ntiles; i++) {
     T2_Box mbr(ndims);
     T2_Point& low = mbr.getLow();
     T2_Point& high = mbr.getHigh();
     low[0] = (T2_ShapeCoord_t) x_start;
     low[1] = 0;
     high[1] = (T2_ShapeCoord_t) (height - 1);

     u_int x_end = x_start + x_mem - 1;
     if (x_end >= width)
	high[0] = (T2_ShapeCoord_t) (width - 1);
     else
	high[0] = (T2_ShapeCoord_t) x_end;

     tile_mbrs << mbr;

     x_start = x_end + 1;
  }

  return T2_UDFRet_OK;
}



/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccMeta constructor
//
// Expects the following values in binary format:
//	w, h		number of pixels along each dimension
//	psize		number of bytes per pixel
//
/////////////////////////////////////////////////////////////////////
T2_UDFRet
svmImageAccMetaConstructor(const T2_QueryInfo& qinfo, const T2_Box& output_box,
			   T2_UsrArgReader& arg, T2_AccMetaObj*& acmp)
{
  u_int w, h, psize;

  if (arg.getDataSize() < sizeof(w)+sizeof(h)+sizeof(psize)) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: not enough arguments for svmImageAccMetaConstructor.");
     return T2_UDFRet_ERROR;
  }

  arg >> w >> h >> psize;
  acmp = (T2_AccMetaObj *) new svm_ImageAccMeta(w, h, psize);

  t2_Assert (output_box.getNumberDimensions() == 2);
  t2_Assert ((u_int) output_box.getLow()[0] == 0);
  t2_Assert ((u_int) output_box.getLow()[1] == 0);
  t2_Assert (w-1 == (u_int) output_box.getHigh()[0]);
  t2_Assert (h-1 == (u_int) output_box.getHigh()[1]);

  return T2_UDFRet_OK;
}

