
/* $Id: svm_accavg.h,v 1.1 2000/02/11 04:47:42 chialin Exp $ */

/* svm_accavg.h */


#ifndef _SVM_ACCAVG_H_
#define _SVM_ACCAVG_H_

#include	"t2_acc.h"
#include	"svm_image.h"
#include	"svm_acc.h"


/////////////////////////////////////////////////////////////////////////
//
// svm_ImageAccAVG
//
/////////////////////////////////////////////////////////////////////////
class svm_ImageAccAVG: public T2_Accumulator {
public:
  svm_ImageAccAVG(u_int tw, u_int th,
	          u_int xp, u_int yp, u_int w, u_int h, u_int s, u_int v);
  ~svm_ImageAccAVG() { delete [] sum_count; };

  // virtual methods
  virtual T2_UDFRet navigateAll(const T2_QueryInfo& qinfo,
				T2_AccIterator*& aitp);
  virtual T2_UDFRet navigate(const T2_QueryInfo& qinfo, T2_IteratorID i,
			     const T2_Region& r, T2_AccIterator*& aitp);

  // query methods for info on total output image
  u_int getTotalOutputImageWidth() const { return x_res; };
  u_int getTotalOutputImageHeight() const { return y_res; };

  // query methods
  u_int getWidth() const { return width; };
  u_int getHeight() const { return height; };
  u_int getPixelSize() const { return pixelsize; };
  u_int getMaxValue() const { return maxval; };
  u_int getXPosition() const { return x_pos; };
  u_int getYPosition() const { return y_pos; };
  u_int* getSumCount() { return sum_count; };
  u_int getSumCountSize() const { return n_sumcounts; };

protected:
  // the following is the information regarding the entire output image
  u_int x_res, y_res;		// width and height of the entire output image

  // the following is the information regarding the tile image;
  // the information regarding the entire output image is kept with
  // svm_ImageAccMetaAVG (some are also stored in this class)
  u_int x_pos, y_pos,		// coordinate in the output space for the
				// starting (ie top-left) pixel
	width, height,		// number of pixels along each dim for this acc
	pixelsize,		// number of bytes per pixel
	maxval;			// max value for each color component

  // an array with one sum for every bytes of a pixel, and one count per pixel
  // (for each pixel, stored as sum_0, sum_1, ..., count)
  u_int *sum_count;
  u_int n_sumcounts;		// total number of u_ints in sum_count[]
				// (should equal to width*height*(pixelsize+1))

private:
};



/////////////////////////////////////////////////////////////////////////
//
// svm_ImageAccAVGIterator
//
/////////////////////////////////////////////////////////////////////////
class svm_ImageAccAVGIterator: public svm_ImageIterator, public T2_AccIterator {
public:
  svm_ImageAccAVGIterator(u_int w, u_int h, u_int psize, u_char* b,
		          u_int x_s, u_int y_s, u_int x_e, u_int y_e);
  ~svm_ImageAccAVGIterator() {};

  // virtual methods
  virtual T2_UDFRet getNextElement(const T2_QueryInfo& qinfo,
				   T2_Accumulator& acc, bool& eod, void*& elem);

protected:
  svm_ImagePixel pixel;		// an object to store the count/sum to return

private:
};


inline
svm_ImageAccAVGIterator::
svm_ImageAccAVGIterator(u_int w, u_int h, u_int psize, u_char* b,
		        u_int x_s, u_int y_s, u_int x_e, u_int y_e)
	: svm_ImageIterator(w, h, (psize+1)*sizeof(u_int), b, 
			    x_s, y_s, x_e, y_e),
	  T2_AccIterator(), pixel()
{
}



/////////////////////////////////////////////////////////////////////////
//
// app_ImageAccMetaAVG
//
/////////////////////////////////////////////////////////////////////////
class svm_ImageAccMetaAVG: public svm_ImageAccMeta {
public:
  svm_ImageAccMetaAVG(u_int w, u_int h, u_int ps): svm_ImageAccMeta(w,h,ps) {};
  ~svm_ImageAccMetaAVG() {};

  // virtual functions
  virtual T2_UDFRet allocAcc(const T2_QueryInfo& qinfo,
			     T2_Iteration i, const T2_Region& r,
			     T2_Accumulator*& accp);
  virtual T2_UDFRet stripMine(const T2_QueryInfo& qinfo, size_t mem,
			      const T2_ClusterInfoList& infolist,
			      T2_VArray<T2_Region>& oqw);

protected:
private:
};



#endif	/* _SVM_ACCAVG_H_ */
