
/* $Id: svm_aggr.h,v 1.3 2000/03/09 01:31:42 chialin Exp $ */

/* svm_aggr.h */


#ifndef _SVM_AGGR_H_
#define _SVM_AGGR_H_


#include	"t2_tarray.h"
#include	"t2_aggr.h"
#include	"svm_acc.h"
#include	"svm_dataset.h"
#include	"svm_utility.h"


////////////////////////////////////////////////////////////////////////
//
// svm_aggregation
//
////////////////////////////////////////////////////////////////////////
class svm_aggregation: public T2_AggregateFuncObj {
public:
  svm_aggregation();
  svm_aggregation(u_int pkno, u_int pktype);
  virtual ~svm_aggregation() {};

  // virtual methods for accumulator initialization
  virtual T2_UDFRet aifElem(const T2_QueryInfo& qinfo, void *accElem) = 0;

  // virtual methods for local reduction phase
  virtual T2_UDFRet dafElem(const T2_QueryInfo& qinfo, T2_DSID i,
			    const void* dataElem, void* accElem) = 0;

  // virtual methods for global reduction phase
  virtual bool needGlobalCombine(const T2_QueryInfo& qinfo,
				 T2_Accumulator& acc);
//  virtual T2_UDFRet getAccMsgSize(const T2_QueryInfo& qinfo,
//				    T2_Accumulator& acc,
//				  T2_Array<size_t>& msgsize);
  virtual T2_UDFRet fillAccMsgBuffer(const T2_QueryInfo& qinfo,
				     T2_Accumulator& acc,
                                     T2_Array<T2_AccMsgBufferWriter>& msgbuf);
  virtual T2_UDFRet processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
			          T2_AccMsgBufferReader& msgbuf,
                                  T2_Accumulator& acc) = 0;

  // virtual methods for output finalization phase
  virtual T2_UDFRet finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			     T2_Output*& outp);

protected:
  // the following are read from the user arg and will be included in the
  // header of each output sub-image that the back-end sends to the client
  u_int packet_no, packet_type;

  // the following describes the sub-image to be combined on the local 
  // processor during the global combine phase;
  // the information for the entire tile-image is kept with svm_ImageAcc
  u_char *my_bytes;		// the assigned bytes to combine on local proc
  size_t my_nbytes;		// the number of bytes assigned to local proc
  u_int my_x_pos, my_y_pos,	// the position of the 1st assigned pixel 
				// in the entire output image
        my_width,		// width of local assigned sub-image
	my_height;		// height of local assigned sub-image

private:
};


svm_aggregation::
svm_aggregation(): T2_AggregateFuncObj(), packet_no(0), packet_type(0),
		   my_bytes(NULL), my_nbytes(0),
		   my_x_pos(0), my_y_pos(0), my_width(0), my_height(0)
{
}

svm_aggregation::
svm_aggregation(u_int pkno, u_int pktype)
		: T2_AggregateFuncObj(), packet_no(pkno), packet_type(pktype),
		  my_bytes(NULL), my_nbytes(0),
		  my_x_pos(0), my_y_pos(0), my_width(0), my_height(0)
{
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMax
//
////////////////////////////////////////////////////////////////////////
class svm_aggrMax: public svm_aggregation {
public:
  svm_aggrMax(u_int pkno, u_int pktype): svm_aggregation(pkno, pktype) {};
  virtual ~svm_aggrMax() {};

  // virtual methods for accumulator initialization
  virtual T2_UDFRet aifElem(const T2_QueryInfo& qinfo, void *accElem);

  // virtual methods for local reduction phase
  virtual T2_UDFRet dafElem(const T2_QueryInfo& qinfo, T2_DSID i,
			    const void* dataElem, void* accElem);

  // virtual methods for global reduction phase
  virtual T2_UDFRet processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
				  T2_AccMsgBufferReader& msgbuf,
				  T2_Accumulator& acc);
protected:
private:
};



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMin
//
////////////////////////////////////////////////////////////////////////
class svm_aggrMin: public svm_aggregation {
public:
  svm_aggrMin(u_int pkno, u_int pktype): svm_aggregation(pkno, pktype) {};
  virtual ~svm_aggrMin() {};

  // virtual methods for accumulator initialization
  virtual T2_UDFRet aifElem(const T2_QueryInfo& qinfo, void *accElem);

  // virtual methods for local reduction phase
  virtual T2_UDFRet dafElem(const T2_QueryInfo& qinfo, T2_DSID i,
			    const void* dataElem, void* accElem);

  // virtual methods for global reduction phase
  virtual T2_UDFRet processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
				  T2_AccMsgBufferReader& msgbuf,
				  T2_Accumulator& acc);

protected:
private:
};



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMaxPPM
//
// A class that is the same as svm_aggrMax, except that it generates
// a set of ppm images as the final output.
// This is mostly used for debugging.
//
////////////////////////////////////////////////////////////////////////
class svm_aggrMaxPPM: public svm_aggrMax  {
public:
  svm_aggrMaxPPM(u_int pkno, u_int pktype): svm_aggrMax(pkno, pktype) {};
  virtual ~svm_aggrMaxPPM() {};

  // generate an output in ppm format
  T2_UDFRet finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
		     T2_Output*& outp);

protected:
private:
};



////////////////////////////////////////////////////////////////////////
//
// svm_aggrFast
//
// A class that is used to support svm_aggrXXXFast.
//
////////////////////////////////////////////////////////////////////////
class svm_aggrFast {
public:
  svm_aggrFast() {};
  ~svm_aggrFast() {};

protected:
  // compute the intersection between mbr, iqr, and tilereg[0]
  T2_UDFRet computeIntersection(const T2_QueryInfo& qinfo,
			const T2_Cluster& input, const T2_Box& mbr,
			T2_ClusterAuxInfoReader& meta,
			T2_Dataset& dataset, T2_DSID dsid,
			T2_IteratorID dsiterid,
			const T2_Box& iqr, const T2_Region& tilereg,
			T2_ProjectFuncObj& proj,
			u_int acciterid, T2_Accumulator& acc,
			const u_char*& in_ptr, u_char*& out_ptr,
			u_int& in_x_start, u_int& in_y_start,
			u_int& in_x_end, u_int& in_y_end,
			u_int& out_x_start, u_int& out_y_start,
			u_int& out_x_end, u_int& out_y_end,
			u_int& x_start_size, u_int& y_start_size,
			u_int& x_mid_size, u_int& y_mid_size,
			u_int& x_end_size, u_int& y_end_size,
			u_int& in_skip_bytes, u_int& out_skip_bytes,
			u_int& pixelsize);

private:
};


////////////////////////////////////////////////////////////////////////
//
// svm_aggrMaxFast
//
// A class that is the same as svm_aggregation, except that it uses
// the interface dafAcc() for local reductoin phase.
//
////////////////////////////////////////////////////////////////////////
class svm_aggrMaxFast: public svm_aggrFast, public svm_aggrMax  {
public:
  svm_aggrMaxFast(u_int pkno, u_int pktype)
				: svm_aggrFast(), svm_aggrMax(pkno, pktype) {};
  ~svm_aggrMaxFast() {};

  // virtual methods for accumulator initialization
//  virtual T2_UDFRet aifElem(const T2_QueryInfo& qinfo, void *accElem);

  // virtual methods for local reduction phase
  virtual
  T2_UDFRet dafAcc(const T2_QueryInfo& qinfo,
		   const T2_Cluster& input, const T2_Box& mbr,
		   T2_ClusterAuxInfoReader& meta,
		   T2_Dataset& dataset, T2_DSID dsid,
		   T2_IteratorID dsiterid,
		   const T2_Box& iqr, const T2_Region& tilereg,
		   T2_ProjectFuncObj& proj,
		   u_int acciterid, T2_Accumulator& acc);

protected:
private:
};



#endif	/* _SVM_AGGR_H_ */
