
/* $Id: svm_aggravg.h,v 1.2 2000/03/09 01:31:55 chialin Exp $ */

/* svm_aggravg.h */


#ifndef _SVM_AGGRAVG_H_
#define _SVM_AGGRAVG_H_


#include	"t2_tarray.h"
#include	"t2_aggr.h"
#include	"svm_accavg.h"


////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVG
//
////////////////////////////////////////////////////////////////////////
class svm_aggrAVG: public T2_AggregateFuncObj {
public:
  svm_aggrAVG();
  svm_aggrAVG(u_int pkno, u_int pktype);
  virtual ~svm_aggrAVG() { delete [] output_bytes; };

  // virtual methods for accumulator initialization
  virtual T2_UDFRet aifElem(const T2_QueryInfo& qinfo, void *accElem);

  // virtual methods for local reduction phase
  virtual T2_UDFRet dafElem(const T2_QueryInfo& qinfo, T2_DSID i,
			    const void* dataElem, void* accElem);

  // virtual methods for global reduction phase
  virtual bool needGlobalCombine(const T2_QueryInfo& qinfo,
				 T2_Accumulator& acc);
//  virtual T2_UDFRet getAccMsgSize(const T2_QueryInfo& qinfo,
//				    T2_Accumulator& acc,
//				  T2_Array<size_t>& msgsize);
  virtual T2_UDFRet fillAccMsgBuffer(const T2_QueryInfo& qinfo,
				     T2_Accumulator& acc,
                                     T2_Array<T2_AccMsgBufferWriter>& msgbuf);
  virtual T2_UDFRet processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
			          T2_AccMsgBufferReader& msgbuf,
                                  T2_Accumulator& acc);

  // virtual methods for output finalization phase
  virtual T2_UDFRet finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			     T2_Output*& outp);

protected:
  // the following are read from the user arg and will be included in the
  // header of each output sub-image that the back-end sends to the client
  u_int packet_no, packet_type;

  // the following describes the sub-image to be combined on the local 
  // processor during the global combine phase;
  // the information for the entire tile-image is kept with svm_ImageAccAVG
  u_int *my_sumcount;		// the assigned bytes to combine on local proc
  size_t my_num_sumcount_ints;	// the total number of sum/count ints 
				// assigned to local proc
  u_int my_x_pos, my_y_pos,	// the position of the 1st assigned pixel 
				// in the entire output image
        my_width,		// width of local assigned sub-image
	my_height;		// height of local assigned sub-image

  // since the output image assumes that it points to a byte array owned 
  // by other object, it does not delete the byte array when it is done; 
  // therefore the byte array has to be deleted by svm_aggrAVG
  u_char *output_bytes;

  // allocate memory for output_bytes[], compute the avg's out of 
  // my_sumcount[], and store the avg's into output_bytes[]
  void computeAVGBytes(svm_ImageAccAVG* acc_image);

private:
};


svm_aggrAVG::
svm_aggrAVG(): T2_AggregateFuncObj(), packet_no(0), packet_type(0),
	       my_sumcount(NULL), my_num_sumcount_ints(0),
	       my_x_pos(0), my_y_pos(0), my_width(0), my_height(0),
	       output_bytes(NULL)
{
}

svm_aggrAVG::
svm_aggrAVG(u_int pkno, u_int pktype)
	: T2_AggregateFuncObj(), packet_no(pkno), packet_type(pktype),
	  my_sumcount(NULL), my_num_sumcount_ints(0),
	  my_x_pos(0), my_y_pos(0), my_width(0), my_height(0),
	  output_bytes(NULL)
{
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVGPPM
//
// A class that is the same as svm_aggrAVG, except that it generates
// a set of ppm images as the final output.
// This is mostly used for debugging.
//
////////////////////////////////////////////////////////////////////////
class svm_aggrAVGPPM: public svm_aggrAVG {
public:
  svm_aggrAVGPPM(u_int pkno, u_int pktype): svm_aggrAVG(pkno, pktype) {};
  virtual ~svm_aggrAVGPPM() {};

  // generate an output in ppm format
  T2_UDFRet finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
		     T2_Output*& outp);

protected:
private:
};



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVGFast
//
////////////////////////////////////////////////////////////////////////
class svm_aggrAVGFast: public svm_aggrAVG {
public:
  svm_aggrAVGFast(u_int pkno, u_int pktype): svm_aggrAVG(pkno, pktype) {};
  virtual ~svm_aggrAVGFast() {};

  // virtual methods for accumulator initialization
  // initializing entire acc -- but since svm_accAVG has initialized all
  // the sums and counts, so do nothing here
  virtual T2_UDFRet aifAcc(const T2_QueryInfo& qinfo, T2_Accumulator& acc)
						{ return T2_UDFRet_OK; };

  // virtual methods for local reduction phase
  // dafElem is never called, so do nothing
  virtual T2_UDFRet dafElem(const T2_QueryInfo& qinfo, T2_DSID i,
			    const void* dataElem, void* accElem)
						{ return T2_UDFRet_OK; };
  // aggregate the entire cluster with the accumulator
  virtual T2_UDFRet dafAcc(const T2_QueryInfo& qinfo,
			   const T2_Cluster& input, const T2_Box& mbr,
			   T2_ClusterAuxInfoReader& meta,
			   T2_Dataset& dataset, T2_DSID dsid,
			   T2_IteratorID dsiterid,
			   const T2_Box& iqr, const T2_Region& tilereg,
			   T2_ProjectFuncObj& proj,
			   u_int acciterid, T2_Accumulator& acc);

protected:
private:
};



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVGPPMFast
//
// A clas that is the same as svm_aggrAVGFast, except that it generates
// ppm files as output images.
//
////////////////////////////////////////////////////////////////////////
class svm_aggrAVGPPMFast: public svm_aggrAVGFast {
public:
  svm_aggrAVGPPMFast(u_int pkno, u_int pktype): svm_aggrAVGFast(pkno,pktype) {};
  virtual ~svm_aggrAVGPPMFast() {};

  // generate an output in ppm format
  T2_UDFRet finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
		     T2_Output*& outp);

protected:
private:
};



#endif	/* _SVM_AGGRAVG_H_ */
