
/* $Id: svm_image.C,v 1.2 2000/03/09 01:34:45 chialin Exp $ */

/* svm_image.C */

#include	"svm_image.h"


//////////////////////////////////////////////////////////////////////////
//
// Image methods
//
//////////////////////////////////////////////////////////////////////////
int
svm_Image::copySubImage(svm_Image& sub_image, int x_pos, int y_pos, int w,int h)
{
  sub_image.clear();

  int x_end = x_pos + w,
      y_end = y_pos + h;

  // check ranges
  if (x_pos < 0 || x_pos >= width || y_pos < 0 || y_pos >= height ||
      x_end < 0 || x_end > width  || y_end < 0 || y_end > height)
  {
     cerr << "ERROR: invalid range for Image::copySubImage" << endl;
     cerr << "width=" << width << " height=" << height
	  << " x_pos=" << x_pos << " y_pos=" << y_pos
	  << " w=" << w << " h=" << h
	  << endl;
     return 1;
  }

  sub_image.getWidth() = w;
  sub_image.getHeight() = h;
  sub_image.getPixelSize() = getPixelSize();
  sub_image.getMaxValue() = getMaxValue();

  u_char*& b = sub_image.getBytes();
  b = new u_char[sub_image.getNumberPixels() * sub_image.getPixelSize()];
  if (b == NULL) {
     cerr << "ERROR: cannot allocate bytes for sub-image in Image::copySubImage"
	  << endl;
     sub_image.clear();
     return 1;
  }

  for (int y=y_pos; y<y_end; y++)
     for (int x=x_pos; x<x_end; x++)
	for (u_int z=0; z<getPixelSize(); z++) {
	   int r = 
	      sub_image.setPixelByte(x-x_pos, y-y_pos, z, getPixelByte(x,y,z));
	   assert (r == 0);
	}

  return 0;
}


u_char
svm_Image::getPixelByte(int x, int y, u_int z) const
{
  assert (x >= 0 && x < width);
  assert (y >= 0 && y < height);
  assert (z >= 0 && z < pixelsize);

  int i = z + pixelsize * (x + width * y);
  return bytes[i];
}


int
svm_Image::setPixelByte(int x, int y, u_int z, u_char v)
{
  if (x >= 0 && x < width && y >= 0 && y < height &&
      z >= 0 && z < pixelsize)
  {
     int i = z + pixelsize * (x + width * y);
     bytes[i] = v;
     return 0;
  }
  return 1;
}



////////////////////////////////////////////////////////////////////////////
//
// svm_ImageIterator methods
//
////////////////////////////////////////////////////////////////////////////
bool
svm_ImageIterator::getNextElement(u_int& x_off, u_int& y_off,
				  svm_ImagePixel& pixel)
{
  // check for end-of-data
  if (y_next > y_end) {			// end of data
     return false;
  }

  // set values for x_off, y_off and pixel
  x_off = x_next;
  y_off = y_next;
  pixel.setBytes(pixelsize, pixel_next);

  // advance pointers to the next pixel 
  if (++x_next > x_end) {
     ++y_next;				// next row
     x_next = x_start;
     pixel_next += skip_bytes;		// move to x_start pos in next row
  }
  else
     pixel_next += pixelsize;

  return true;
}

