
/* $Id: svm_prj.C,v 1.1 2000/02/04 00:21:21 chialin Exp $ */

/* svm_prj.C */

#include	"svm_prj.h"


///////////////////////////////////////////////////////////////////////
//
// svm_ImageProjection methods
//
///////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageProjection::project(const T2_System& system,
			     const T2_Box& iqr, const T2_Region& tilereg,
                             const T2_Point& input_pt,
			     bool& succeed, T2_Region& output_rg)
{
  if (iqr.getNumberDimensions() != 2 ||
      input_pt.getNumberDimensions() != 2)
  {  // only handles 2-dim attribute spaces
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: can only handle 2-dim attribute space, but ndims(iqr)= %d ndims(input_pt)= %d",
		iqr.getNumberDimensions(), input_pt.getNumberDimensions());
     return T2_UDFRet_ERROR;
  }

  succeed = false;

  if (!iqr.contains(input_pt))		// point not inside query window
     return T2_UDFRet_OK;

  const u_int ndims = 2;		// only handles 2-dim attribute spaces

  T2_Point output_pt(ndims);		// a 2-dim point
  projectPoint(input_pt, output_pt);

//  if (out_bound.contains(output_pt) == true	// inside the output bounds
  if (tilereg.contains(output_pt) == true)	// inside the tile region
  {
     output_rg.setNumberDimensions(ndims);  // set # dims for the output region
     output_rg.growMaxSetSize(1);	// make space for one point
     output_rg << output_pt;
     succeed = true;
  }

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageProjection::projectBox(const T2_System& system, const T2_Box& inbox,
				bool& succeed, T2_Box& outbox)
{
  if (inbox.getNumberDimensions() != 2) {
     // only handles 2-dim attribute spaces
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: can only handle 2-dim attribute space, but ndims(inbox)= %d",
		inbox.getNumberDimensions());
     return T2_UDFRet_ERROR;
  }

  const u_int ndims = 2;		// only handles 2-dim attribute spaces
  outbox.setNumberDimensions(ndims);
  T2_Point& out_low = outbox.getLow();
  T2_Point& out_high = outbox.getHigh();

  projectPoint(inbox.getLow(), out_low);
  projectPoint(inbox.getHigh(), out_high);

//  // does the output box fall inside the bound?
//  if (! (out_bound ^ outbox)) {		// no
//      succeed = false;
//      return T2_UDFRet_OK;
//  }
//
//  // adjust the output box, if it is not entirely inside the bounds
//  u_int i;
//  T2_Point& outbound_low = out_bound.getLow();
//  T2_Point& outbound_high = out_bound.getHigh();
//  for (i=0; i<ndims; i++) {		// check the lower bounds
//     if (out_low[i] < out_bound_low[i])
//	out_low[i] = out_bound_low[i];
//  }
//  for (i=0; i<ndims; i++) {		// check the higher bounds
//     if (out_high[i] < outbound_high[i])
//	out_high[i] = outbound_high[i];
//  }

  succeed = true;
  return T2_UDFRet_OK;
}



///////////////////////////////////////////////////////////////////////
//
// svm_ImageProjection constructor
//
// The arguments for the constructor consists of the following
// fields, in binary format.
//
//	int xo, yo	the coordinates in the input space for the
//			origin of the output space
//	u_int xs, ys	the number of input units that are contained
//			by an output unit along each dimension
//
// All numbers are of type u_int, and they are separated by blanks.
//
///////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImgPrjConstructor(const T2_System& system, T2_UsrArgReader& arg,
		      T2_ProjectFuncObj*& prjfunc)
{
  int xo, yo;
  u_int xs, ys;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(xo)+sizeof(yo)+sizeof(xs)+sizeof(ys)) {
     // not enough arguments !!
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(xo)+sizeof(yo)+sizeof(xs)+sizeof(ys));
     return T2_UDFRet_ERROR;
  }

  arg >> xo >> yo >> xs >> ys;

  // create the projection function object
  prjfunc = (T2_ProjectFuncObj *) new svm_ImageProjection(xo, yo, xs, ys);

  return T2_UDFRet_OK;
}

