
/* $Id: chunkify-ppm.C,v 1.1 2000/02/01 00:52:18 chialin Exp $ */

/* chunkify-ppm.C 
 *
 * This utility program breaks a ppm file into multiple chunks and assign
 * chunks to different files.
 */


#include	<iostream.h>
#include	<fstream.h>
#include	<stdlib.h>
#include 	<string.h>
#include	<stdio.h>
#include	<assert.h>
#include	<strings.h>
#include	<time.h>
#include	<sys/types.h>

#ifndef NULL
#  define NULL		0
#endif


//////////////////////////////////////////////////////////////////////////
//
// a class for an image
//
//////////////////////////////////////////////////////////////////////////
class Image {
public:
  Image(): width(0), height(0), pixelsize(0), maxval(0), bytes(NULL) {};
  ~Image() { delete [] bytes; };

  // access methods
  int getNumberPixels() const { return width*height; };
  int& getWidth() { return width; };
  int& getHeight() { return height; };
  int& getPixelSize() { return pixelsize; };
  int& getMaxValue() { return maxval; };
  char*& getBytes() { return bytes; };

  // access the z-th component of a pixel with position x and y
  char getPixelByte(int x, int y, int z) const;

  // set the z-th component of a pixel with position x and y;
  // return 0 if succeeds, 1 otherwise
  int setPixelByte(int x, int y, int z, char v);

  // wipe out the image and remove all pixels;
  void clear() { width = height = pixelsize = 0; maxval = 0;
		 delete [] bytes; bytes=NULL; };

  // create a sub-image starting from position x_pos, y_pos with width=w
  // and height=h; returns 0 if succeeds, 0 otherwise
  int copySubImage(Image& sub_image, int x_pos, int y_pos, int w, int h);
  
protected:
  int width, height,		// number of pixels for the image
      pixelsize,		// number of bytes per pixel
      maxval;			// maximum color-component value
  char* bytes;			// an array of size width*height*bytes

private:
};


//////////////////////////////////////////////////////////////////////////
//
// prototypes
//
//////////////////////////////////////////////////////////////////////////
void
printusage(const char* prog);
int
read_ppm_file(ifstream& file, Image& image);
int
get_next_uncomment_string(ifstream& file, char* buf, int buf_len, char comment);
int
read_raw_bytes(ifstream& file, int nbytes, char* bytes);
void
write_raw_ppm_header(Image& image, ofstream& os, const char *descrition);
int
chunkify_image(Image& image, int x_part, int y_part, char output_type,
	       int ndfiles, ofstream* out_dfs, ofstream* out_ifs);

//////////////////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  if (argc < 9) {
     printusage(argv[0]);
     return -1;
  }

  // variables
  const char *inp_imagefile;
  const char *out_datafileprefix, *out_idxfileprefix, *out_auxfile;
  char  output_type;
  int   x_part, y_part,
	ndfiles,			// number of dataset files
	i = 1;

  // get arguments
  inp_imagefile = argv[i++];
  x_part = atoi(argv[i++]);
  y_part = atoi(argv[i++]);
  output_type = argv[i++][0];
  ndfiles = atoi(argv[i++]);
  out_datafileprefix = argv[i++];
  out_idxfileprefix = argv[i++];
  out_auxfile = argv[i++];
  if (x_part <=0 || y_part <=0) {
     cerr << "ERROR: invalid x/y partitioning '" << argv[2] << " " << argv[3]
	  << "'" << endl;
     printusage(argv[0]);
     return -1;
  }
  if (output_type != 'p' && output_type != 'b') {
     cerr << "ERROR: invalid output type '" << argv[4] << "'" << endl;
     printusage(argv[0]);
     return -1;
  }

  ifstream inp(inp_imagefile, ios::in);
  if (!inp) {
     cerr << "ERROR: cannot open input image file '" << inp_imagefile
	  << "'" << endl;
     return -1;
  }

  Image inp_image;
  if (read_ppm_file(inp, inp_image) != 0) {	// some thing went wrong
     inp.close();
     return -1;
  }
  inp.close();

  // get current time
  time_t t = time(NULL);
  char *curtime = ctime(&t);
  if (curtime[24] == '\n')
     curtime[24] = '\0';

  // get ready to open output files
  ofstream *out_dfs = new ofstream[ndfiles],
	   *out_ifs = new ofstream[ndfiles],
	   out_aux;

  // open output aux file
  out_aux.open(out_auxfile, ios::out);
  if (!out_aux) {
     cerr << "ERROR: cannot open output aux file '" << out_auxfile << "'"
	  << endl;
     return -1;
  }
  out_aux << inp_image.getWidth() << " " << inp_image.getHeight()
	  << " " << inp_image.getPixelSize() << " " << inp_image.getMaxValue()
	  << endl
	  << "# generated by " << argv[0]
		<< " " << argv[1] << " " << argv[2]
		<< " " << argv[3] << " " << argv[4]
		<< " " << argv[5] << " " << argv[6]
		<< " " << argv[7] << " " << argv[8]
		<< endl
	  << "# generated at " << curtime << " " << tzname[0] << endl
	  << "# format: <image width> <image height> <number of bytes per pixel> <max byte value>"
	  << endl;
  out_aux.close();

  // open remaining output files
  char outfilename[256];
  for (i = 0; i<ndfiles; i++) {
     // open output dataset file
     sprintf(outfilename, "%s.%d", out_datafileprefix, i);
     out_dfs[i].open(outfilename, ios::out);
     if (!out_dfs[i]) {			// open fails
	cerr << "ERROR: cannot open output file '" << outfilename
	     << "'" << endl;
	for (int k=0; k<i; k++) {
	    out_dfs[k].close();
	    out_ifs[k].close();
	}
	delete [] out_dfs;
	delete [] out_ifs;
	return -1;
     }

     // open output index file
     sprintf(outfilename, "%s.%d", out_idxfileprefix, i);
     out_ifs[i].open(outfilename, ios::out);
     if (!out_ifs[i]) {			// open fails
	cerr << "ERROR: cannot open output file '" << outfilename
	     << "'" << endl;
	for (int k=0; k<i; k++) {
	    out_dfs[k].close();
	    out_ifs[k].close();
	}
	out_dfs[i].close();	
	delete [] out_dfs;
	delete [] out_ifs;
	return -1;
     }

     // write header information into index file
     out_ifs[i] << "# generated by " << argv[0]
		<< " " << argv[1] << " " << argv[2]
		<< " " << argv[3] << " " << argv[4]
		<< " " << argv[5] << " " << argv[6]
		<< endl
		<< "# generated at " << curtime << " " << tzname[0]
		<< endl
		<< "# format:" << endl
		<< "# <x_pos> <y_pos> <width> <height> <# bytes per pixel> <offset> <size> <file id>"
		<< endl
		<< "#"
		<< endl;
  }

  int r = chunkify_image(inp_image, x_part, y_part, output_type,
			 ndfiles, out_dfs, out_ifs);

  for (i=0; i<ndfiles; i++) {
     out_dfs[i].close();
     out_ifs[i].close();
  }

  delete [] out_dfs;
  delete [] out_ifs;

  return r;
}


//////////////////////////////////////////////////////////////////////////
//
// printusage
//
//////////////////////////////////////////////////////////////////////////
void
printusage(const char* prog)
{
  cerr << "Usage: " << prog << "\\" << endl
       << "       <ppm/pgm file> \\"
       << endl
       << "       <number of x partitions> <number of y partitions> <output type> \\"
       << endl
       << "       <number of dataset files> <dataset file prefix> <index file prefix> \\"
       << endl
       << "       <aux file prefix>"
       << endl
       << "where <number of x partitions> and <number of y partitions> must be"
       << endl
       << "positive numbers,"
       << endl
       << "<output type> is either p, in which case each chunk will have a ppm format,"
       << endl
       << "or b, in which case only the bytes of the pixles are written into data files."
       << endl
       << "The ppm output type only makes sense when the number of dataset files"
       << endl
       << "is equal to <<number of x partitions> times <number of y partitions>."
       << endl
       << endl
       << "This program splits a ppm/pgm file into "
       << endl
       << "<number of x partitions> X <number of y partitions> sub-images, and"
       << endl
       << "assigns each sub-image to one of the <number of dataset files> output files"
       << endl
       << "in a round-robin fashion. An index file is also generated for each output"
       << endl
       << "dataset file, which describes the chunks assign to that dataset file."
       << endl
       << "The output dataset file would be named as <dataset file prefix>.n, and"
       << endl
       << "its corresponding index file is named as <index file prefix>.n, where"
       << endl
       << "n is number between 0 and <number of dataset files>."
       << endl;
}



//////////////////////////////////////////////////////////////////////////
//
// read_ppm_file
//
// Read a ppm/pgm file. Return 0 if succeeds, 1 otherwise
//
//////////////////////////////////////////////////////////////////////////
int
read_ppm_file(ifstream& file, Image& image)
{
  image.clear();

  const int max_line_length = 71;
  char line[max_line_length];
  const char comment = '#';

  // get magic number for file type
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read magic number for file type." << endl;
     image.clear();
     return 1;
  }
  if (line[0] != 'P' || (line[1] != '3' && line[1] != '6')) {
     cerr << "ERROR: invalid magic number '" << line << "'" << endl;
     image.clear();
     return 1;
  }
  char ppm_type = line[1];

  // get width
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read width for image." << endl;
     image.clear();
     return 1;
  }
  if ((image.getWidth() = atoi(line)) <= 0) {
     cerr << "ERROR: invalid width = " << image.getWidth() << endl;
     image.clear();
     return 1;
  }

  // get height
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read height for image." << endl;
     image.clear();
     return 1;
  }
  if ((image.getHeight() = atoi(line)) <= 0) {
     cerr << "ERROR: invalid width = " << image.getHeight() << endl;
     image.clear();
     return 1;
  }

  // get maxval
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read max value for image." << endl;
     image.clear();
     return 1;
  }
  if ((image.getMaxValue() = atoi(line)) <= 0) {
     cerr << "ERROR: invalid width = " << image.getMaxValue() << endl;
     image.clear();
     return 1;
  }

  // skip the white space immediately after maxval
  char ch;
  file.get(ch);

  switch (ppm_type) {
    case '3':	// ascii format, not implelmented yet ...
		image.getPixelSize() = 3;	// 3 bytes per pixel
		cerr << "ERROR: sorry, P3 (ascii) format is not implemented yet"		     << endl;
		image.clear();
		return 1;

    case '6':	// raw format
		image.getPixelSize() = 3;	// 3 bytes per pixel
		{
		   const int image_size_bytes =
			image.getNumberPixels() * image.getPixelSize();
		   image.getBytes() = new char[image_size_bytes];
		   if (image.getBytes() == NULL) {		// new fails
			cerr << "ERROR: fail to new " << image_size_bytes
			     << " bytes for image." << endl;
			image.clear();
			return 1;
		   }
		   if (read_raw_bytes(file, image_size_bytes,
				      image.getBytes()) != 0)
		   {	// error
			image.clear();
			return 1;
		   }
		}
		break;
  }
  return 0;
}


//////////////////////////////////////////////////////////////////////////
//
// some I/O help functions
//
// return 0 if succeeds, 1 otherwise
//
//////////////////////////////////////////////////////////////////////////
int
get_next_uncomment_string(ifstream& file, char* buf, int buf_len, char comment)
{
  while (!file.eof()) {
    file >> buf;
    if (file.eof())			// eof
       return 1;

    // search for comments
    int string_len = strlen(buf);
    for (int i=0; i<string_len; i++) {
	if (buf[i] == comment) {
	   buf[i] = '\0';

	   // skip the rest of the line
	   char ch;
	   while (file.get(ch) && ch != '\n');

	   break;
	}
    }
    if (buf[0] != '\0')			// found a string
	return 0;
  }
  return 1;				// eof
}


int
read_raw_bytes(ifstream& file, int nbytes, char* bytes)
{
  char *p = bytes,
       *endp = bytes + nbytes;
  while (p != endp && file.get(*p))
	p++;

  if (p == endp)			// all bytes are read
     return 0;
  return 1;
}


int
chunkify_image(Image& image, int x_part, int y_part, char output_type,
	       int ndfiles, ofstream* out_dfs, ofstream* out_ifs)
{
  assert (out_dfs != NULL);
  assert (out_ifs != NULL);
  assert (output_type == 'p' || output_type == 'b');

  // compute how many pixels per sub-image in each dimension;
  // when cannot divide, the excess ones always to the last sub-image along
  // each dimension
  // (eg partitioning 5 pixels into 3 sub-images, first two sub-images get
  // 1 pixel each and last one gets 3 pixels)
  // partitioning fails when image is too small to be partitioned

  int sub_width = image.getWidth() / x_part,	// width of sub-image
      sub_height = image.getHeight() / y_part;	// height of sub-image

  if (sub_width == 0 || sub_height == 0) {
     cerr << "ERROR: image is too small to be partitioned"
	  << endl;
     return 1;
  }
  int excess_width = image.getWidth() - sub_width * x_part,
      excess_height = image.getHeight() - sub_height * y_part;

  // just so that I don't have to keep all the sub-images in the memory
  // at any time, I am simply going to create each sub-image at a time
  // and write them to disk as soon as they are created
  int x_pos = 0, y_pos = 0,	// top-left position of current sub-image
      cur_dfile = 0;		// current data file to assign sub-image to
  for (int y=0; y<y_part; y++) {
      int h = sub_height;
      if (y == y_part - 1)
	  h += excess_height;

      x_pos = 0;
      for (int x=0; x<x_part; x++) {
	  int w = sub_width;
	  if (x == x_part - 1)
	      w += excess_width;

	  Image sub_image;
	  if (image.copySubImage(sub_image, x_pos, y_pos, w, h) != 0) {
	     // copy fails
	     return 1;
	  }

	  // write header if necessary
	  switch (output_type) {
	    case 'p':	// output ppm file 
			{
			   char pos_info[512];
			   sprintf(pos_info, "# position x=%d y=%d",
				   x_pos, y_pos);
			   write_raw_ppm_header(sub_image, out_dfs[cur_dfile],
						pos_info);

			}
			break;

	    case 'b':	// output binary file, no header is needed
			break;
	  }


	  // remember offset before writing bytes into dataset file
	  off_t offset = out_dfs[cur_dfile].tellp();

	  // output bytes
	  char *pixel = sub_image.getBytes();
	  for (int k=0; k<sub_image.getNumberPixels(); k++) {
	     for (int n=0; n<sub_image.getPixelSize(); n++) {
		out_dfs[cur_dfile].put(*pixel);
		pixel++;
	     }
	  }

	  // output index
	  out_ifs[cur_dfile]
		<< x_pos << " " << y_pos << " " << w << " " << h
		<< " " << sub_image.getPixelSize()
		<< " " << offset
		<< " " << sub_image.getNumberPixels()*sub_image.getPixelSize()
		<< " " << cur_dfile
		<< "		# sub-image." << x << "-" << y
		<< endl;

	  x_pos += w;				// advance x position

	  if (++cur_dfile >= ndfiles) {		// get to next data file
	     cur_dfile = 0;
	  }
      }

      y_pos += h;			// advance y position
  }

  return 0;
}


void
write_raw_ppm_header(Image& image, ofstream& os, const char *description)
{
  assert (image.getPixelSize() == 3);
  os << "P6" << endl;
//  os << "# generated by split-image" << endl;
  if (description != NULL)
     os << description << endl;
  os << image.getWidth() << " " << image.getHeight()
     << endl << image.getMaxValue() << endl;
}


//////////////////////////////////////////////////////////////////////////
//
// Image methods
//
//////////////////////////////////////////////////////////////////////////
int
Image::copySubImage(Image& sub_image, int x_pos, int y_pos, int w, int h)
{
  sub_image.clear();

  int x_end = x_pos + w,
      y_end = y_pos + h;

  // check ranges
  if (x_pos < 0 || x_pos >= width || y_pos < 0 || y_pos >= height ||
      x_end < 0 || x_end > width  || y_end < 0 || y_end > height)
  {
     cerr << "ERROR: invalid range for Image::copySubImage" << endl;
     cerr << "width=" << width << " height=" << height
	  << " x_pos=" << x_pos << " y_pos=" << y_pos
	  << " w=" << w << " h=" << h
	  << endl;
     return 1;
  }

  sub_image.getWidth() = w;
  sub_image.getHeight() = h;
  sub_image.getPixelSize() = getPixelSize();
  sub_image.getMaxValue() = getMaxValue();

  char*& b = sub_image.getBytes();
  b = new char[sub_image.getNumberPixels() * sub_image.getPixelSize()];
  if (b == NULL) {
     cerr << "ERROR: cannot allocate bytes for sub-image in Image::copySubImage"
	  << endl;
     sub_image.clear();
     return 1;
  }

  for (int y=y_pos; y<y_end; y++)
     for (int x=x_pos; x<x_end; x++)
	for (int z=0; z<getPixelSize(); z++) {
	   int r = 
	      sub_image.setPixelByte(x-x_pos, y-y_pos, z, getPixelByte(x,y,z));
	   assert (r == 0);
	}

  return 0;
}


char
Image::getPixelByte(int x, int y, int z) const
{
  assert (x >= 0 && x < width);
  assert (y >= 0 && y < height);
  assert (z >= 0 && z < pixelsize);

  int i = z + pixelsize * (x + width * y);
  return bytes[i];
}


int
Image::setPixelByte(int x, int y, int z, char v)
{
  if (x >= 0 && x < width && y >= 0 && y < height &&
      z >= 0 && z < pixelsize)
  {
     int i = z + pixelsize * (x + width * y);
     bytes[i] = v;
     return 0;
  }
  return 1;
}
