
/* $Id: gen-svm-qbatch.C,v 1.3 2000/03/01 06:00:14 chialin Exp $ */

/* gen-svm-qbatch.C */

#include	<iostream.h>
#include 	<stdlib.h>
#include	"t2_bfile.h"
#include	"t2_qspec.h"

u_char my_endian = T2_BigEndian;

///////////////////////////////////////////////////////////////////////////
//
// some help function
//
///////////////////////////////////////////////////////////////////////////
void
printusage(const char *prog)
{
  cerr << "Usage: " << prog
       << " <r/w> <query batch file> [<e>]"
       << endl
       << "where" << endl
       << "<r/w> specifies a read or write operation of a query batch file,"
       << endl
       << "<query batch file> specifies the query batch file,"
       << endl
       << "<e> , when given, will have the program prompt for endian information."
       << endl << endl
       << "The optional argument <e> is only effective in write operation."
       << endl
       << "Note that a query specification object could specify the endian used for"
       << endl
       << "the user-args for various constructor functions, and the endian used for"
       << endl
       << "the final query output. When <e> is not specified, by default the program"
       << endl
       << "uses the same endian of the machine that this program runs on."
       << endl;

}

void get_datasetspec(T2_QSpecDataset& spec, u_int& w, u_int& h);
void input_qspec(T2_QSpec& qspec, const bool use_default_endian);
void read_svm_query(const char* fname);
void input_svm_query(const char* fname, const bool use_default_endian);
void output_qbatch(T2_QBatch& qbatch, T2_OBFile& obf);
char enable_protocol_or_not();


///////////////////////////////////////////////////////////////////////////
//
// main function
//
///////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  if (argc < 3) {
     printusage(argv[0]);
     return -1;
  }

  bool use_default_endian = true;
  if (argc > 3) {
     if (argv[3][0] == 'e')
        use_default_endian = false;
     else {
	cerr << "ERROR: invalid argument 3" << endl;
        printusage(argv[0]);
        return -1;
     }
  }

  // decide the endian of the local machine
  my_endian = t2_checkEndianess();

  switch (argv[1][0]) {
     case 'R':
     case 'r':	// read query
		read_svm_query(argv[2]);
		break;

     case 'W':
     case 'w':	// write query
		input_svm_query(argv[2], use_default_endian);
		break;
     default:
		cerr << "ERROR: invalid operation '" << argv[1] << "'" << endl;
		printusage(argv[0]);
		return -1;
  }

  return 0;
}



///////////////////////////////////////////////////////////////////////////
//
// read_svm_query
//
///////////////////////////////////////////////////////////////////////////
void
read_svm_query(const char* fname)
{
  T2_IBFile ibf(fname);
  if (!ibf) {
     t2_err_quit(__FILE__, __LINE__, 
		"cannot open input qbatch file %s", fname);
     return;
  }

  T2_QBatch qbatch;
  ibf >> qbatch;
  if (qbatch.isValid() == false)
     cout << "read failed." << endl;
  else
     cout << qbatch << endl;

  // read some of the constructor usr args
  T2_UsrArgReader arg;

  for (u_int i=0; i<qbatch.getNumberQueries(); i++) {
     cout << "for query " << i << endl;
     T2_QSpec& qspec = qbatch.getQuerySpec(i);

     // print out accumulator constructor usr arg
     u_int width, height, pixelsize;
     arg.open(qspec.getAccConstructorArg());
     if (arg.getDataSize() < sizeof(u_int)*3)
	cerr << "ERROR: expecting " << sizeof(u_int)*3
	     << " bytes for accumulator constructor arg, but only see "
	     << arg.getDataSize() << " bytes." << endl;
     else {
	arg >> width >> height >> pixelsize;
	cout << "accumulator constructor arg: width=" << width
	     << " height=" << height << " pixelsize=" << pixelsize << endl;
     }

     // print out aggregation constructor usr arg
     u_int packet_no, packet_type;
     arg.open(qspec.getAggrConstructorArg());
     if (arg.getDataSize() < sizeof(u_int)*2)
	cerr << "ERROR: expecting " << sizeof(u_int)*2
	     << " bytes for aggregation constructor arg, but only see "
	     << arg.getDataSize() << " bytes." << endl;
     else {
	arg >> packet_no >> packet_type;
	cout << "aggregation constructor arg: packet_no=" << packet_no
	     << " packet_type=" << packet_type << endl;
     }

     for (u_int j=0; j<qspec.getNumberDatasets(); j++) {
	T2_QSpecDataset& dspec = qspec.getDatasetSpec(j);
	int x_pos, y_pos, x_scale, y_scale;
	arg.open(dspec.getProjConstructorArg());
	if (arg.getDataSize() < sizeof(u_int)*4)
	   cerr << "ERROR: expecting " << sizeof(u_int)*4
	        << " bytes for projection constructor arg, but only see "
	        << arg.getDataSize() << " bytes." << endl;
	else {
	   arg >> x_pos >> y_pos >> x_scale >> y_scale;
	   cout << "projection constructor arg: x_pos=" << x_pos
	        << " y_pos=" << y_pos << " x_zoom=" << x_scale
		<< " y_zoom=" << y_scale << endl;
	}
     }

  }
}


///////////////////////////////////////////////////////////////////////////
//
// input_svm_query
//
///////////////////////////////////////////////////////////////////////////
void
input_svm_query(const char* fname, const bool use_default_endian)
{
  if (fname == NULL)
     return;

  cout << "Creating a query batch." << endl;
  T2_QBatch qbatch;

  int nqueries;
  cout << "How many query specifications in the query batch?" << endl;
  cout.flush();
  cin >> nqueries;
  if (nqueries == 0)
     return;

  T2_OBFile obf(fname);
  if (!obf) {
     t2_err_quit(__FILE__, __LINE__, 
		"cannot open output qbatch file %s", fname);
     return;
  }

  qbatch.setNumberQueries(nqueries);
  for (int i=0; i<nqueries; i++) {
    cout << endl << "Inputting query specification for query " << i << endl;

    T2_QSpec& qspec = qbatch.getQuerySpec(i);
    input_qspec(qspec, use_default_endian);
  }
  cout << endl << "query specification setting completes" << endl;
  output_qbatch(qbatch, obf);

  cout << endl;
  obf.close();
}


void input_qspec(T2_QSpec& qspec, const bool use_default_endian)
{
  char buf[129];
  buf[128] = '\0';

  T2_UsrArgWriter argwr;
  cout << endl;

  // ask about endian-ness of the usr arg first
  if (use_default_endian == true) {
     switch (my_endian) {
	case T2_BigEndian:	qspec.useBigEndianUserArg(); break;
	case T2_LittleEndian:	qspec.useLittleEndianUserArg(); break;
	default:		t2_err_quit(__FILE__, __LINE__,
				    "ERROR: invalid endian %d",(int) my_endian);
				exit (-1);
     }
     cout << "using " << endian2string(my_endian) << " for query specification"
	  << endl;
  }
  else {
     u_int choice;
     do {
       cout << "endian-ness for the query specification : 1 - big endian, 2 - little endian: ";
       cout.flush();
       cin >> choice;
       switch (choice) {
	 case 1:	qspec.useBigEndianUserArg(); break;
	 case 2:	qspec.useLittleEndianUserArg(); break;
	 default:	cout << "invalid choice, try again ....." << endl;
       }
     }
     while (choice < 1 || choice > 2);
  }

  // now with aggregation function stuff
  cout << "Now, let's work on the aggregatoin function part." << endl;
  cout.flush();
  cout << "aggregation function id: "; cout.flush();
  cin >> qspec.getAggrID();

  // aggregation constructor user arg
  u_int packet_no, packet_type;
  cout << "aggregation constructor arg: packet_no = "; cout.flush();
  cin >> packet_no;
  cout << "                             packet_type (should be 6) = ";
  cout.flush();
  cin >> packet_type;
  qspec.getAggrConstructorArg().allocBuffer(sizeof(u_int)*2);
  argwr.open(qspec.getAggrConstructorArg());
  argwr << packet_no << packet_type;
  argwr.close();
  cout << endl;

  // now with dataset stuff
  u_int width, height, pixelsize;
  qspec.setNumberDatasets(1);
  get_datasetspec(qspec.getDatasetSpec(0), width, height);
  cout << endl;

  // start with accumulator stuff
  cout << "accumulator id: "; cout.flush();
  cin >> qspec.getAccID();
  qspec.getAccNavigatorID() = 0;		// navigate id = 0

  // accumulator constructor user arg
  cout << "acc meta constructor arg: width = " << width << endl;
  cout << "                          height = " << height << endl;
  cout << "                          pixelsize = "; cout.flush();
  cin >> pixelsize;
  qspec.getAccConstructorArg().allocBuffer(sizeof(u_int)*3);
  argwr.open(qspec.getAccConstructorArg());
  argwr << width << height << pixelsize;
  argwr.close();
  cout << endl;

  // set the output query box
  T2_Box& output_box = qspec.getOutputBox();
  output_box.setNumberDimensions(2);
  output_box.getLow()[0] = 0;
  output_box.getLow()[1] = 0;
  output_box.getHigh()[0] = width - 1;
  output_box.getHigh()[1] = height - 1;
  

  // now with output handling stuff
  int port;
  T2_QSpecOutput& outspec = qspec.getOutputSpec();
  cout << "Now, let's specify what to do with the final output" << endl;

  // deal with endian-ness
  if (use_default_endian == true) {
     switch (my_endian) {
	case T2_BigEndian:	outspec.useBigEndianOutput(); break;
	case T2_LittleEndian:	outspec.useLittleEndianOutput(); break;
	default:		t2_err_quit(__FILE__, __LINE__,
				    "ERROR: invalid endian %d",(int) my_endian);
				exit (-1);
     }
     cout << "using " << endian2string(my_endian) << " for query output"
	  << endl;
  }
  else {
     u_int choice;
     do {
      cout << "endian-ness for the output: 1 - big endian, 2 - little endian: ";
      cout.flush();
      cin >> choice;
      switch (choice) {
	 case 1:	outspec.useBigEndianOutput(); break;
	 case 2:	outspec.useLittleEndianOutput(); break;
	 default:	cout << "invalid choice, try again ....." << endl;
      }
     } while (choice < 1 || choice > 2);
  }

  u_int choice;
  do {
      cout << "your choice: 1 - write into a file, 2 - write to socket: ";
      cout.flush();
      cin >> choice;
      if (choice >=1 && choice <=2)
	 break;
      cout << "invalid choice, try again ....." << endl;
  }
  while (1);
  switch (choice) {
    case 1:				// output to a file
	outspec.setOutputHandleType(t2_oBFile);
	cout << "output file name (less than 128 bytes):" << endl;
	cin >> buf;
	outspec.setFilePrefix(buf);
	break;

    case 2:				// output to a socket
	outspec.setOutputHandleType(t2_oSocket);
	cout << "client host name (less than 64 bytes):"; cout.flush();
	cin >> buf;
	outspec.setHostName(buf);
	cout << "port number:"; cout.flush();
	cin >> port;
	outspec.setPortNumber(port);

	{
	  char c;
          switch(c = enable_protocol_or_not()) {
	    case 'y': outspec.enableT2Protocol(); break;
	    case 'n': outspec.disableT2Protocol(); break;
	    default : t2_err_quit(__FILE__, __LINE__,
		     	  "unexpected char returned : %d", (int) c);
		      exit(1);
	  }
	}
	break;

    default:
	t2_err_quit(__FILE__, __LINE__, "unexpected choice %d", choice);
	exit(1);
  }
  cout << endl;
}


void
get_datasetspec(T2_QSpecDataset& spec, u_int& output_w, u_int& output_h)
{
  cout << "	dataset id: "; cout.flush();
  cin >> spec.getDatasetID();
  spec.getIteratorID() = 0;			// iterator = 0
  cout << "	index id: "; cout.flush();
  cin >> spec.getIndexID();
  // no usr arg for dataset constructor

  cout << "	projection function id: "; cout.flush();
  cin >> spec.getProjID();

  // get projection arguments
  u_int x_pos, y_pos, x_scale, y_scale;
  cout << "	projection arg: x_pos= "; cout.flush();
  cin >> x_pos;
  cout << "	                y_pos= "; cout.flush();
  cin >> y_pos;
  cout << "	                output image width= "; cout.flush();
  cin >> output_w;
  cout << "	                output image height= "; cout.flush();
  cin >> output_h;
  cout << "	                x_zoomout_factor= "; cout.flush();
  cin >> x_scale;
  cout << "	                y_zoomout_factor= "; cout.flush();
  cin >> y_scale;

  if (output_w == 0 || output_h == 0)
  {
     cerr << "output width and output height must be non-zero" << endl;
     exit (1);
  }

  // construct projection constructor user arg
  spec.getProjConstructorArg().allocBuffer(sizeof(u_int)*4);
  T2_UsrArgWriter argwr(spec.getProjConstructorArg());
  argwr << x_pos << y_pos << x_scale << y_scale;
  argwr.close();

  // construct input query box
  T2_Box& inbox = spec.getQueryBox();
  inbox.setNumberDimensions(2);
  inbox.getLow()[0] = x_pos;
  inbox.getLow()[1] = y_pos;
  inbox.getHigh()[0] = x_pos + output_w * x_scale - 1;
  inbox.getHigh()[1] = y_pos + output_h * y_scale - 1;

  cout << endl;
}


char enable_protocol_or_not()
{
  char c;
  do {
    cout <<"do you want to enable or disable T2 protocol (y-enable/n-disable) ?"
	 << endl;
    cin >> c;
    switch (c) {
      case 'y':
      case 'Y':
		  return 'y';
      case 'n':
      case 'N':
		  return 'n';
    }
    cout << "invalid choice, please try again ...." << endl;
  } while (1);
  return 'y';
}


void
output_qbatch(T2_QBatch& qbatch, T2_OBFile& obf)
{
  cout << "writing into the output file now ...." << endl;
  obf << qbatch;
}

