
/* $Id: app-client-inquiry.C,v 1.1 2000/01/28 23:27:31 chialin Exp $ */

/* app-client-inquiry.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/24/2000
 * An simple client that reads queries from files, and uses T2_FrontEnd
 * interface to talk to a T2 Front-End process, and send inquiries.
 */


#include	<iostream.h>
#include	<unistd.h>
#include	"t2_bfile.h"
#include	"t2_frontend.h"
#include	"t2_feinquiry.h"


/////////////////////////////////////////////////////////////////////////////
//
// some help function prototypes and variables
//
/////////////////////////////////////////////////////////////////////////////
static void printusage(const char*);

/////////////////////////////////////////////////////////////////////////////
//
// main
//
/////////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  int t2fe_port;
  const char *hostname;

  // deal with the command-line arguments
  if (argc < 6) {
     printusage(argv[0]);
     return -1;
  }

  int i = 1;
  hostname = argv[i++];
  t2fe_port = atoi(argv[i++]);
  if (t2fe_port < 0) {
     cout << "invalid port number " << argv[i-1] << endl;
     printusage(argv[0]);
     return -1;
  }

  cout << "T2 front-end host = " << hostname
       << ", port = " << t2fe_port << endl;

  T2_FrontEnd fe;
  if (fe.connectT2FrontEndByHostname(hostname, t2fe_port) == false) {
     // cannot connect to T2 front-end
     if (fe.getErrorVal() == t2_feT2Busy)
	cout << "T2 front-end is busy" << endl;
     else
	t2_err_ret(__FILE__, __LINE__, "fe.connectT2FrontEndByHostname error");
     return -1;
  }

  cout << "Connected to T2 front-end" << endl;
  cout << "T2 current has " << fe.getNumberBackEndNodes()
       << " back-end nodes" << endl;


  // vars to hold inquiry results
  T2_FEDatasetInquiryResults dataset_results;
  T2_FEFunctionInquiryResults func_results;

  // go through the inquiries
  for (; i<argc;) {
     cout << endl;

     // check if enough arguments are given
     if (argc < i+3) {
	cout << "ERROR: not enough arguments for the last inquiry, inquiry ignored."
	     << endl;
	break;
     }

     cout << "--------- processing inquiry " << argv[i] << " " << argv[i+1]
	  << " " << argv[i+2] << endl;

     // go through the arguments
     char it, match, *pattern;
     it = argv[i++][0];
     match = argv[i++][0];
     pattern = argv[i++];

     bool do_inquiry = true,
	  exact_match,		// exact match (true) or regular exp (false)
	  inq_dataset;		// for dataset (true) or for functions (false)

     T2_UDFType ftype = T2_UDF_Unknown;

     switch (it) {
       case 'D':
       case 'd':
		inq_dataset = true; break;
       case 'C':
       case 'c':
		inq_dataset = false; ftype = T2_UDF_AccMeta; break;
       case 'A':
       case 'a':
		inq_dataset = false; ftype = T2_UDF_Aggregation; break;
       case 'P':
       case 'p':
		inq_dataset = false; ftype = T2_UDF_Projection; break;
       default:
		cout << "ERROR: invalid inquiry type '"
		     << argv[i-3] << "' in argument " << i-3
		     << ", inquiry ignored.";
		do_inquiry = false;
     }

     switch (match) {
       case 'E':
       case 'e':
		exact_match = true; break;
       case 'R':
       case 'r':
		exact_match = false; break;
       default:
		cout << "ERROR: invalid matching option '"
		     << argv[i-2] << "' in argument " << i-2
		     << ", inquiry ignored.";
		do_inquiry = false;
     }
  
     if (do_inquiry == true) {		// ready to do inquiry
	if (inq_dataset == true) {	// a dataset inquiry
	   dataset_results.clear();	// remove old entries

	   // choose to return all fields
	   int fields = T2_FEInquiry::dataset_datasetid_field |
			T2_FEInquiry::dataset_datasetname_field |
			T2_FEInquiry::dataset_datasetdescription_field |
			T2_FEInquiry::dataset_blobsize_field |
			T2_FEInquiry::dataset_blobobj_field |
			T2_FEInquiry::dataset_iteratorname_field |
			T2_FEInquiry::dataset_indexid_field |
			T2_FEInquiry::dataset_indexname_field |
			T2_FEInquiry::dataset_indexdescription_field;


	   // want to put any limit on how many bytes can be allocated for
	   // all the blob objects ?
	   size_t max_total_blobsize_limit = 0;		// no limit

	   bool r;
	   if (exact_match == true)
	      r = fe.inquireDatasetExactMatch(pattern, fields,
					      max_total_blobsize_limit,
					      dataset_results);
	   else
	      r = fe.inquireDatasetRegExp(pattern, fields,
					  max_total_blobsize_limit,
					  dataset_results);

	   if (r == false ||
	       dataset_results.getStatus() != T2_FEInquiry::results_ok)
	      cout << "WARNING: inquire dataset returns error" << endl;

	   cout << "Number of dataset entries returned = "
		<< dataset_results.getNumberEntries() << endl;
	   for (u_int k=0; k<dataset_results.getNumberEntries(); k++) {
	      T2_FEDatasetEntry& entry = dataset_results[k];
	      cout << "record " << k << endl << entry;
	   }
	}
	else {				// if (inq_dataset == true)
	   func_results.clear();	// remove old entries

	   // choose to return all fields
	   int fields = T2_FEInquiry::function_id_field |
			T2_FEInquiry::function_name_field |
			T2_FEInquiry::function_description_field;

	   bool r;
	   if (exact_match == true)
	      r = fe.inquireFunctionExactMatch(pattern, ftype, fields,
					       func_results);
	   else
	      r = fe.inquireFunctionRegExp(pattern, ftype, fields,
					   func_results);

	   if (r == false ||
	       func_results.getStatus() != T2_FEInquiry::results_ok)
	      cout << "WARNING: inquire function returns error" << endl;

	   cout << "Number of constructor function entries returned = "
		<< func_results.getNumberEntries() << endl;
	   for (u_int k=0; k<func_results.getNumberEntries(); k++) {
	      T2_FEFunctionEntry& entry = func_results[k];
	      cout << "record " << k << endl << entry;
	   }
	}				// if (inq_dataset == true) else

     }
  }


  cout << endl << "now disconnect from T2 front-end" << endl;
  fe.disconnectT2FrontEnd();

  cout << "program " << argv[0] << " exits gracefully" << endl;
  return 0;
}



/////////////////////////////////////////////////////////////////////////////
//
// help functions
//
/////////////////////////////////////////////////////////////////////////////
void
printusage(const char* pname)
{
  cout << "Usage: " << pname
    << " <T2 front-end hostname> <port number> <inquiry 1> [<inquiry 2> ..]"
    << endl;

  cout << "Where:" << endl;
  cout << "<T2 front-end hostname> <port number> specifies the hostname and the"
       << endl
       << "	port number to connect to the T2 front-end process,"
       << endl
       << "and each <inquiry> has the following format:"
       << endl
       << "	<inquiry type> <matching option> <pattern>"
       << endl
       << "where <inquiry type> = d-dataset, c-accmeta, a-aggregation, p-projection"
       << endl
       << "      <matching option> = e-exact match, r-regular expression"
       << endl
       << "      <pattern> = name or regular expression to match"
       << endl;
}

