
/* $Id: conscat.C,v 1.2 2000/02/23 22:09:06 chialin Exp $ */

/* conscat.C */

#include	<stdio.h>
#include	<stdlib.h>
#include	<time.h>
#include	<strings.h>
#include	<string.h>
#include	"t2_conscat.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_ConsCatalogEntry methods
//
/////////////////////////////////////////////////////////////////////////
T2_TxtFile&
T2_ConsCatalogEntry::read(T2_TxtFile& file)
{
  clear();
  id = 0;
  if (!file) {
     type = T2_UDF_Unknown;
     return file;
  }

  t2_Assert (file.getFileName() != NULL);
  const char *string;

  // get type
  type = t2_shortstring2UDFType((string = file.getNextUncommentString()));
  if (type == T2_UDF_Unknown) {
     t2_err_ret(NULL, 0,
		"ERROR: invalid constructor type '%s' at line %d in file '%s'",
		string, file.getCurrentLineNumber(), file.getFileName());
     return file;
  }
  const u_int lineno = file.getCurrentLineNumber();
  
  // get id
  if ((string = file.getNextUncommentString()) == NULL) {
     type = T2_UDF_Unknown;
     t2_err_ret(NULL, 0,
		"ERROR: missing constructor id for constructor entry at line %d in file '%s'",
		lineno, file.getFileName());
     return file;
  }
  int atoi_err;
  int ii = t2_atoi(string, &atoi_err);
  if (atoi_err != 0) {
     type = T2_UDF_Unknown;
     t2_err_ret(NULL, 0,
		"ERROR: invalid constructor id '%s' for constructor entry at line %d in file '%s'.",
		string, lineno, file.getFileName());
     return file;
  }
  if (ii < 0) {
     type = T2_UDF_Unknown;
     t2_err_ret(NULL, 0,
		"ERROR: negative id for constructor entry at line %d in file '%s' is not allowed.",
		lineno, file.getFileName());
     return file;
  }
  id = (u_int) ii;

  // get name
  if ((string = file.getNextUncommentString()) == NULL) {
     type = T2_UDF_Unknown;
     t2_err_ret(NULL, 0,
		"ERROR: no constructor name is given for the entry defined at line %d in file '%s'",
		lineno, file.getFileName());
     return file;
  }
  size_t len = strlen(string) + 1;
  t2_Alloc(name, len, char);
  memcpy(name, string, len);

  // get description, if any
  file.getNextLine();
  if (file.eof() != true && file.getLine()[0] != file.getCommentMark()) {
     // description is not null
     char* p = (char *) file.getLine();
     while (*p != '\0' && *p != file.getCommentMark())
	p++;
     *p = '\0';
     string = file.getLine();
     len = strlen(string) + 1;
     t2_Alloc(description, len, char);
     memcpy(description, string, len);
  }
  file.getNextLine();

  if (type == T2_UDF_Dataset) {		// read iterators
     if ((string = file.getNextUncommentString()) == NULL) {
	type = T2_UDF_Unknown;
	t2_err_ret(NULL, 0,
		"ERROR: expecting number of iterators for the dataset constructor function defined at line %d in file '%s'.",
		lineno, file.getFileName());
	return file;
     }
     int num_iters = t2_atoi(string, &atoi_err);
     if (atoi_err != 0 || num_iters < 0) {
	type = T2_UDF_Unknown;
	t2_err_ret(NULL, 0,
		"ERROR: invalid number of iterators '%s' at line %d for the dataset constructor function defined at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), 
		lineno, file.getFileName());
	return file;
     }
     iterators.clear(num_iters);
     for (u_int i=0; i<(u_int) num_iters; i++) {
	if ((string = file.getNextUncommentString()) == NULL) {
	   type = T2_UDF_Unknown;
	   t2_err_ret(NULL, 0,
		"ERROR: only %d of %d iterator are defined for the dataset constructor function defined at line %d in file '%s'.",
		i, num_iters, lineno, file.getFileName());
	   return file;
	}
	char*& iter_p = iterators[i];
	str_alloc_cpy(iter_p, string);
     }
  }			// if (type == T2_UDF_Dataset)

  return file;
}


ostream&
T2_ConsCatalogEntry::write(ostream& os) const
{
  t2_Assert (name != NULL);
  os << t2_UDFType2shortstring(type) << "		" << id << endl
     << name << endl;
  if (description == NULL)
     os << "    # no description" << endl;
  else
     os << description << endl;
  if (type == T2_UDF_Dataset) {
     if (iterators.getNumberElements() == 0)
	os << "0	# no iterators" << endl;
     else {
	os << iterators.getNumberElements();
	for (u_int i=0; i<iterators.getNumberElements(); i++) {
	   const char* it_name = iterators[i];
	   os << " " << it_name;
	}
	os << "  # iterators" << endl;
     }
  }
  return os;
}



/////////////////////////////////////////////////////////////////////////
//
// T2_ConsCatalogWriter methods
//
/////////////////////////////////////////////////////////////////////////
T2_ConsCatalogWriter::T2_ConsCatalogWriter(const char* fname)
				: file(), valid(false)
{
  if (fname != NULL)
     open(fname);
}


void
T2_ConsCatalogWriter::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  ifstream ifs(fname, ios::in);
  if (ifs) {				// open succeeds
     ifs.close();
     char bakfname[256];
     assert (strlen(fname) < 256);
     sprintf(bakfname, "%s.bak", fname);

     if (rename(fname, bakfname) != 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename to '%s'",
		fname, bakfname);
	return;
     }
     t2_say(cout, "WARNING: Rename " << fname << " to " << bakfname << endl);
  }

  file.open(fname, ios::out);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__, 
		"ERROR: cannot open constructor catalog file '%s'", fname);
     return;
  }

  time_t t = time(NULL);
  char *curtime = ctime(&t);
  if (curtime[24] == '\n')
     curtime[24] = '\0';
  file << "# constructor catalog" << endl
       << "# generated by class T2_ConsCatalogWriter at " << curtime 
       << " " << tzname[0]
       << endl;
  valid = true;
}


void
T2_ConsCatalogWriter::close()
{
  file.close();
  valid = false;
}


bool
T2_ConsCatalogWriter::addEntry(T2_UDFType t, u_int n, const char* name,
			       const char* descr,
			       const T2_VArray<T2_CharPtr>& iterators)
{
  T2_ConsCatalogEntry e;
  e.type = t;
  e.name = (char *) name;
  e.description = (char *) descr;
  e.id = n;

  if (t == T2_UDF_Dataset) {
    e.iterators.clear(iterators.getNumberElements());
    for (u_int i=0;i<iterators.getNumberElements(); i++) {
       char*& e_iter = e.iterators[i];
       const char* iter = iterators[i];
       e_iter = (char *) iter;
    }
  }
  else
    t2_Assert (e.iterators.getNumberElements() == 0);

  file << endl << e;

  // to avoid deleting memory
  e.name = NULL;
  e.description = NULL;
  if (t == T2_UDF_Dataset) {
    for (u_int i=0;i<iterators.getNumberElements(); i++) {
       char*& e_iter = e.iterators[i];
       e_iter = NULL;
    }
  }

  return true;
}



/////////////////////////////////////////////////////////////////////////
//
// T2_ConsCatalogReader methods
//
/////////////////////////////////////////////////////////////////////////
T2_ConsCatalogReader::T2_ConsCatalogReader(const char* fname)
			: file('#', 2048), 
			  search_type(T2_UDF_Unknown), re_init(false)
{
  if (fname != NULL)
     open(fname);
}


void
T2_ConsCatalogReader::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  file.open(fname, true);	// true makes file to remember its filename
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open constructor catalog file '%s'", fname);
     return;
  }
}


bool
T2_ConsCatalogReader::fetchRegExpInit(const char* pattern,
				      T2_UDFType t, int flags)
{
  file.rewind();
  if (re_init == true) {
     regfree(&re);
  }
  else
     re_init = true;
  if (regcomp(&re, pattern, flags) != 0)
       return false;   			// report error
  search_type = t;
  return true;
}


bool
T2_ConsCatalogReader::fetchRegExpName(T2_ConsCatalogEntry& e)
{
  if (file.eof() == true)
     return false;

  while (file.eof() == false) {
     e.read(file);
     if (e.type == T2_UDF_Unknown) 	// some thing went wrong
	return false;

     if (search_type == T2_UDF_Unknown || search_type == e.type) {
        if (regexec(&re, e.name, (size_t) 0, NULL, 0) == 0)
	   return true;			// successful match
     }
  }
  e.clear();
  return false;
}


bool
T2_ConsCatalogReader::fetchExactMatch(T2_ConsCatalogEntry& e,
				      const char* name, const T2_UDFType t)
{
  file.rewind();
  while (file.eof() == false) {
     e.read(file);
     if (e.type == T2_UDF_Unknown) 	// some thing went wrong
	return false;

     if (t == T2_UDF_Unknown || t == e.type) {
        if (strcmp(name, e.name) == 0)
	   return true;			// successful match
     }
  }
  e.clear();
  return false;
}

