
/* $Id: t2dld_index.h,v 1.4 1999/03/27 19:44:14 chialin Exp $ */

/* t2dld_index.h */

#ifndef T2DLD_INDEX_H_
#define T2DLD_INDEX_H_


#include	<iostream.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_bfile.h"
#include	"t2_usrarg.h"
#include	"t2_shape.h"
#include	"t2_txtfile.h"
#include	"t2dld_base.h"


////////////////////////////////////////////////////////////////////////////
//
// T2DLD_STDIndexFile
//
// A class used to access a T2 index file, which is coupled with a name
// file. The index file provides information about data clusters stored
// in a data subset, such as bounding boxes (u-mbr), offsets, sizes,
// etc.
//
////////////////////////////////////////////////////////////////////////////
class T2DLD_STDIndexFile {
public:
  T2DLD_STDIndexFile();
  ~T2DLD_STDIndexFile() { reset(); };

  // valid?
  operator const void*() const { return (index_format == 0)? NULL:this; };

  // open files
  bool open(const char* index_file);
  
  // query methods
  char getIndexFormat() const { return index_format; };
  u_int getMBRNumberDimensions() const { return ndims; };
  u_int getNumberEntries() const { return nentries; };

  // read a data cluster from index file
  bool read(T2DLD_DataCluster& leaf);

  // read all data clusters from index file into cluster array
  bool read(T2DLD_TapeClusterArray& c);

  // output to ostream
  ostream& write(ostream& os) const;

  // write header information into an index file using the binary format
  void bwrite(T2_OBFile& obf) const;

protected:
  T2_TxtFile index_a;		// for index file in ascii format
  T2_IBFile index_b;		// for index file in binary format

  char index_format;		// index file format
  u_int ndims,			// # dims for all bounding boxes
        nentries;		// # entries listed in the index file

  // reset data members, including closing files and discarding data file names
  void reset();

private:
};


//inline
//T2DLD_STDIndexFile::operator const void*() const
//{
//  if (namefile) return this;
//  return NULL;
//}


inline ostream&
operator << (ostream& os, const T2DLD_STDIndexFile& indf)
{
  return indf.write(os);
}



#endif	/* ! T2DLD_INDEX_H_ */
