
/* $Id: dld-base.C,v 1.3 2000/02/21 07:21:13 chialin Exp $ */

/* dld-base.C */

#include	"t2dld_base.h"


////////////////////////////////////////////////////////////////////////////
//
// T2DLD_DataBlock methods
//
////////////////////////////////////////////////////////////////////////////
ostream&
T2DLD_DataBlock::write(ostream& os) const
{
  if (data == NULL) {
     os << "(" << getDataFileID() << "," << getOffset()
	<< "," << getSize() << ")";
  }
  else
     os << "<data of " << getSize() << " bytes>";

  return os;
}



////////////////////////////////////////////////////////////////////////////
//
// T2DLD_DataCluster methods
//
////////////////////////////////////////////////////////////////////////////
T2DLD_DataCluster::T2DLD_DataCluster() : mbr(), tblocks(0), user_data()
{
}


ostream&
T2DLD_DataCluster::write(ostream& os) const
{
  os << "low=" << mbr.getLow() << ", high=" << mbr.getHigh() << endl;
  os << "\t(fid,offset,size)=";
  for (u_int i=0; i<getNumberDataBlocks(); i++) {
     const T2DLD_DataBlock& block = getDataBlock(i);
     os << " " << block;
  }
  os << endl;

  os << "user data:[" << user_data.getStartOfData() << "]" << endl;
  return os;
}


void
T2DLD_DataCluster::bwrite(T2_OBFile& obf) const
{
  if (!obf)
     return;

  u_int d;
  const T2_Point& low = mbr.getLow();
  const T2_Point& high = mbr.getHigh();

  // write coordinates of low and high points for bounding box
  for (d=0; d<mbr.getNumberDimensions(); d++)
     obf << low[d];
  for (d=0; d<mbr.getNumberDimensions(); d++)
     obf << high[d];

  // write # of physical blocks and each block
  obf << getNumberDataBlocks();
  for (u_int i=0; i<getNumberDataBlocks(); i++) {
     const T2DLD_DataBlock& block = getDataBlock(i);
     obf << block.getDataFileID()
	 << block.getOffset()
	 << block.getSize();
  }

  // write user data
  obf << user_data.getDataSize();
  if (user_data.getDataSize() > 0)
     obf.bwrite(user_data.getStartOfData(), user_data.getDataSize(),
		sizeof(char));
}






////////////////////////////////////////////////////////////////////////////
//
// T2DLD_DataClusterArray methods
//
////////////////////////////////////////////////////////////////////////////
T2DLD_DataClusterArray::T2DLD_DataClusterArray(u_int l)
		: tape_filenames(0), clusters(l)
{
}


T2DLD_DataClusterArray::~T2DLD_DataClusterArray()
{
  for (u_int f=0; f<getNumberDataFiles(); f++)
     delete [] tape_filenames[f];

  for (u_int i=0; i<getNumberDataClusters(); i++)
     delete clusters[i];
}


