/*
  File name: t2mc_output.H

  Class for output pointer list. 
*/
#ifndef T2MC_OUTPUT_H
#define T2MC_OUTPUT_H

#include "t2_header.h"
#include "t2_error.h"
#include "t2_tarray.h"
#include "t2mc_types.H"


class T2_MCOutputDataPtrList {
private:
  struct MC_Buf {
    const char*     data;
    size_t          nelems;
    T2_DataType     data_type;
    T2_OutputType   comm_type;

    MC_Buf() { 
      data      = NULL; 
      nelems    = 0; 
      data_type = DATA_UNKNOWN;
      comm_type = T2_OUT_UNKNOWN;
    };

    MC_Buf(const char* buf, size_t n, T2_DataType t, T2_OutputType ct) {
      data      = buf;
      nelems    = n;
      data_type = t;
      comm_type = ct;
    };

    ~MC_Buf() { };
  };

  T2_VArray<MC_Buf> MC_ptrs;

public:
  T2_MCOutputDataPtrList(u_int l=100): MC_ptrs(l) {};
  ~T2_MCOutputDataPtrList() { };
  
  u_int getNumberPointers() const { return MC_ptrs.getNumberElements(); };
  const char*   getDataPointer(u_int i) const;
  size_t        getDataSize(u_int i) const;
  T2_DataType   getDataType(u_int i) const;
  T2_OutputType getCommType(u_int i) const;

  const char* getDataPointer(u_int i);
  void        insertDataPointer(const char* p, size_t n, T2_DataType t, 
				T2_OutputType ct) 
  { MC_ptrs << MC_Buf(p,n,t,ct); };

  void removeAllPointers() { MC_ptrs.clear(); };
};

inline const char*
T2_MCOutputDataPtrList::getDataPointer(u_int i) const
{
  if (i < getNumberPointers())
    return MC_ptrs[i].data;
  return NULL;
}

inline size_t
T2_MCOutputDataPtrList::getDataSize(u_int i) const
{
  if (i < getNumberPointers())
    return MC_ptrs[i].nelems;
  return 0;
}

inline T2_DataType
T2_MCOutputDataPtrList::getDataType(u_int i) const
{
  if (i < getNumberPointers())
    return MC_ptrs[i].data_type;
  return DATA_UNKNOWN;
}

inline T2_OutputType
T2_MCOutputDataPtrList::getCommType(u_int i) const
{
  if (i < getNumberPointers())
    return MC_ptrs[i].comm_type;
  return T2_OUT_UNKNOWN;
}


inline const char*
T2_MCOutputDataPtrList::getDataPointer(u_int i)
{
  if (i < getNumberPointers())
    return MC_ptrs[i].data;
  return NULL;
}

#endif /* T2MC_OUTPUT_H */
