/*
  File name: t2mc_user.H

  Contains the class definition for user customization.
*/
#ifndef T2MC_USER_H
#define T2MC_USER_H

#include "t2_header.h"
#include "t2_error.h"
#include "t2_output.h"
#include "responsable.h"
#include "mc_regions.H"
#include "mc_options.H"
#include "t2mc_output.H"
#include "mc_data.H"

/*
 * This class is used to customize Meta-Chaos implementation for 
 * T2.
 *
 *
 * We need the following function in T2_AggerationFunc obj. 
 * 
 * T2_MetaChaosUser *finalizeMC(T2_Accumulator  &acc, 
 *                              T2_System       &system,
 *                              MC_Server       &srvr, 
 *                              MC_Client       &clnt, 
 *                              T2_UsrArgReader &arg);
 *
 */
class T2_MetaChaosUser : public MC_DataSender {
private:
  MC_Options       opts;
public:
  T2_MetaChaosUser() {
    opts.setWhereTo(ComputeOnReceiver);
    opts.setCommOpt(DoNotOptComm);
  };
  virtual ~T2_MetaChaosUser() { };

  virtual MC_SetOfRegions&  getSetOfRegions() = 0;
  virtual MC_Liste&         derefLocal(const MC_Region& rgn, 
				       int glbOffset) = 0;
  virtual MC_OtherProgData& otherProgData() = 0;

  // A default implementation is available
  virtual MC_Options& getOptions() { return opts; };

  virtual size_t requiredMCBufferSpace() = 0;
  virtual void   flushMCBuffer(T2_OutputBuffer& buf,
			       T2_MCOutputDataPtrList& ptr_list) = 0;
};

#endif /* T2MC_USER_H */
