/*
  File name: MC_liste.H

  Contains utility classes.
*/
#ifndef MC_LISTE_H
#define MC_LISTE_H

#include "zm.h"
#include "liste.h"
#include "mc_assert.H"

extern "C" void MC_FreeListeZM(liste_s *l, void (*f)(ZM_s *zm));

class MC_Zm {
  friend class MC_Liste;
private:
  ZM_s *zm;
public:
  MC_Zm() { zm = NULL; };

  MC_Zm(int glbOffset, int procid, int lclOffset, int nelems) {
    init(glbOffset,procid,lclOffset,nelems);
  };

  ~MC_Zm() { };

  void freeZm() {
    if (zm!=NULL) { free(zm); zm=NULL; }
  };
  
  inline int getGlobalOffset() {
    return globalOffset_ZM(zm);
  };
  
  inline int getSize() {
    return size_ZM(zm);
  };

  inline int getProcId() {
    return proc_ZM(zm);
  };
  
  inline int getLocalOffset() {
    return offset_ZM(zm);
  };
  
  inline void init(int glbOffset, int procid, int lclOffset, int nelems) {
    zm = init_ZM(glbOffset,procid,lclOffset,nelems);
    MC_ASSERT((zm!=NULL));
  };

  inline void set(int glbOffset, int procid, int lclOffset, int nelems) {
    MC_ASSERT(zm!=NULL);
    zm->globalOffset = glbOffset;
    zm->proc         = procid;
    zm->offset       = lclOffset;
    zm->size         = nelems;
  };
};

/*
 * Class to handle list of ZM's 
 */
class MC_Liste {
  friend class MC_Schedule;
private:
  MListe_s *mlist;
  liste_s  *liste;
  ZM_s     *zm;
  MC_Zm    mc_zm;

  liste_s *getListePtr() {
    return liste;
  };

  MC_Liste(liste_s *l) {
    this->liste = l;
  };
public:
  MC_Liste() {
    liste = NULL;
    mlist = NULL;
    zm    = NULL;
  };

  MC_Liste(MC_Liste& l) {
    this->liste = l.liste;
    this->zm    = l.zm;
    mlist = NULL;
  };

  ~MC_Liste() { };

  void init() {
    if (liste==NULL) {
      liste = AllocListe();
      MC_ASSERT(liste!=NULL);
    }
  };    

  void addZM(int glbOffset, int procid, int lclOffset, int nelems) {
    zm = init_ZM(glbOffset,procid,lclOffset,nelems);
    MC_ASSERT((zm!=NULL));
    AddElemListe(liste,zm);
  };

  void addZM(MC_Zm &zm) {
    this->zm = init_ZM(zm.getGlobalOffset(),zm.getProcId(),
		       zm.getLocalOffset(),zm.getSize());
    AddElemListe(liste,this->zm);
  };

  void free() {
    if (liste!=NULL) {
      MC_FreeListeZM(liste,free_ZM);
      liste = NULL;
    }
  };

  void reset() {
    MC_ASSERT(liste!=NULL);
    mlist = liste->premier;
  };

  void advance() {
    if (mlist!=NULL)
      mlist = mlist->suiv;
  };

  MC_Zm& getElem() {
    if (mlist!=NULL)
      mc_zm.zm = (ZM_s*)(mlist->elem);
    else 
      mc_zm.zm = NULL;
    return mc_zm;
  };

  int end() {
    if (mlist==NULL) 
      return 1;
    return 0;
  };
};

#endif /* MC_LISTE_H */

