/*
  File name: mc_offbandcomm.C

  Contains code for offband communication in meta-chaos.

  $Id: mc_offbandcomm.C,v 1.2 1998/09/03 14:56:37 kurc Exp $
*/
#ifndef MC_OFFBANDCOMM_C
#define MC_OFFBANDCOMM_C

#include "mc_offbandcomm.H"

int 
MC_OffBandComm::bcast(MC_Client& to, void *data, int size, DATA_Type type, 
		      int tag)
{
  int rslt;

  switch (type) {
  case DATA_SHORT:
    size = size/sizeof(short);
    rslt = shortBcastOffBand(to.program(),(short*)data,size,tag);
    break;
  case DATA_INT:
    size = size/sizeof(int);
    rslt = intBcastOffBand(to.program(),(int*)data,size,tag);
    break;
  case DATA_FLOAT:
    size = size/sizeof(float);
    rslt = floatBcastOffBand(to.program(),(float*)data,size,tag);
    break;
  case DATA_DOUBLE:
    size = size/sizeof(double);
    rslt = doubleBcastOffBand(to.program(),(double*)data,size,tag);
    break;
  case DATA_STRING:
    rslt = stringBcastOffBand(to.program(),(char*)data,size,tag);
    break;
  case DATA_CHAR:
    rslt = byteBcastOffBand(to.program(),(char*)data,size,tag);
    break;
  default:
    fprintf(stderr,"Unknown bcast data type in %s --> %d\n",__FILE__,__LINE__);
    fflush(stderr);
    rslt = 0;
    break;
  }
  return rslt;
}

int 
MC_OffBandComm::recv(MC_Client& from, void *data, int size, DATA_Type type, 
		      int tag)
{
  int rslt;

  switch (type) {
  case DATA_SHORT:
    size = size/sizeof(short);
    rslt = shortRecvOffBand(from.program(),(short*)data,size,tag);
    break;
  case DATA_INT:
    size = size/sizeof(int);
    rslt = intRecvOffBand(from.program(),(int*)data,size,tag);
    break;
  case DATA_FLOAT:
    size = size/sizeof(float);
    rslt = floatRecvOffBand(from.program(),(float*)data,size,tag);
    break;
  case DATA_DOUBLE:
    size = size/sizeof(double);
    rslt = doubleRecvOffBand(from.program(),(double*)data,size,tag);
    break;
  case DATA_STRING:
    rslt = stringRecvOffBand(from.program(),(char*)data,size,tag);
    break;
  case DATA_CHAR:
    rslt = byteRecvOffBand(from.program(),(char*)data,size,tag);
    break;
  default:
    fprintf(stderr,"Unknown recv. data type in %s --> %d\n",__FILE__,__LINE__);
    fflush(stderr);
    rslt = 0;
    break;
  }
  return rslt;
}

int 
MC_OffBandComm::recv(void *data, int size, DATA_Type type, int tag)
{
  int rslt;

  switch (type) {
  case DATA_SHORT:
    size = size/sizeof(short);
    rslt = shortRecvOffBand(NULL,(short*)data,size,tag);
    break;
  case DATA_INT:
    size = size/sizeof(int);
    rslt = intRecvOffBand(NULL,(int*)data,size,tag);
    break;
  case DATA_FLOAT:
    size = size/sizeof(float);
    rslt = floatRecvOffBand(NULL,(float*)data,size,tag);
    break;
  case DATA_DOUBLE:
    size = size/sizeof(double);
    rslt = doubleRecvOffBand(NULL,(double*)data,size,tag);
    break;
  case DATA_STRING:
    rslt = stringRecvOffBand(NULL,(char*)data,size,tag);
    break;
  case DATA_CHAR:
    rslt = byteRecvOffBand(NULL,(char*)data,size,tag);
    break;
  default:
    fprintf(stderr,"Unknown recv. data type in %s --> %d\n",__FILE__,__LINE__);
    fflush(stderr);
    rslt = 0;
    break;
  }
  return rslt;
}

#endif /* MC_OFFBANDCOMM_C */






