#ifndef __liste__
#define __liste__

#include "util.h"

/*
 * Liste : mode d emploi
 * ---------------------
 * liste_s decrit toute la liste
 * tete_liste_s decrit la tete de liste
 * membre_liste_s decrit un membre de la liste (pas un element)
 * elem decrit un element de la liste
 * 
 * ecriture plus rapide:
 * TListe tete_liste_s 
 * MListe membre_liste_s
 * EListe elem 
 *
 */

/* ----------------------------------------------------------------
   Alias
---------------------------------------------------------------- */

#define TListe_s liste_s
#define MListe_s membre_liste_s
#define EListe_s void

/* ----------------------------------------------------------------
   Structures
---------------------------------------------------------------- */
typedef struct Smembre_liste_s{
  struct Smembre_liste_s *suiv;
  void            *elem;
  } membre_liste_s;

typedef struct{
  membre_liste_s *premier;
  membre_liste_s *dernier;  
  int             nbElem ;
} liste_s;



#ifdef __cplusplus
  extern "C" {
#endif


/* ----------------------------------------------------------------
   Functions
---------------------------------------------------------------- */

liste_s *AllocListe            /* alloue la tete de la liste     */
_AnsiC_((
	 void
	 ));

MListe_s *AllocMListe          /* alloue un membre de la liste   */ 
_AnsiC_((
	 void
	 ));

TListe_s *AllocTListe          /* alloue la tete de liste        */
_AnsiC_((
	 void
	 ));

void FreeListe                 /* desalloue la liste liste, en   */
_AnsiC_((                      /* utilisant la fct FreeTListe    */
	 liste_s *liste,       /* pour la tete de la liste       */
	 void (*FreeElem)()    /* FreeMListe pour les membre     */
	 ));                   /* de la liste et FreeElem pour   */
                               /* les elements de la liste       */

void FreeTListe                /* desalloue la tete de liste     */
_AnsiC_((
	 TListe_s *liste
	 ));

void FreeMListe                /* desalloue les membres d une    */
_AnsiC_((                      /* liste                          */
	 MListe_s *membre
	 ));

void     AddElemListe          /* Alloue un membre utilisant la  */
_AnsiC_((                      /* function AllocMListe, puis     */
	 liste_s *liste,       /* ajoute elem a la liste en      */
	 void    *elem         /* utilisant ce membre            */
	 ));

void
AddElemListeUsingMListe        /* ajoute elem a la liste en      */  
_AnsiC_((                      /* utilisant le membre fourni     */
	 liste_s  *liste,
	 MListe_s *membre,
	 void     *elem 
	 ));

int NbElemListe                /* retourne le nombre d element   */
_AnsiC_((                      /* actuellement dans la liste     */
	 liste_s  *liste
	 ));

MListe_s *PremierListe         /* retourne le premier membre     */
_AnsiC_((                      /* de la liste                    */
	 liste_s  *liste
	 ));

MListe_s *DernierListe         /* retourne le dernier membre     */
_AnsiC_((                      /* de la liste                    */
	 liste_s  *liste
	 ));

EListe_s *MSuivListe           /* retourne le membre suivant     */
_AnsiC_((                      /* de la liste                    */
	 MListe_s  *membre
	 ));

EListe_s *ElemListe            /* retourne l element             */
_AnsiC_((                      /* correspondant au membre        */
	 MListe_s  *membre
	 ));

liste_s *MoveListe             /* deplace les elements de la     */
_AnsiC_((                      /* liste phagociter dans la liste */
	 liste_s *res,         /* res (resultat).                */
	 liste_s *phagociter   /* Attention la tete de la liste  */
	 ));                   /* phagociter est desallouer      */

void ApplyFctListe             /* applique une fonction fct a    */ 
_AnsiC_((                      /* tous les elements de la liste  */
	 liste_s *liste,
	 void (*fct)()
	 ));

void PrintListe
_AnsiC_((        
	 liste_s *liste,
	 void (*fct)()
	 ));


#ifdef __cplusplus
  }
#endif


#endif




