#include <stdio.h>
#include <string.h>

#include "pvm3.h"

#include "util.h"
#include "pgme.h"
#include "schedule.h"

int _stat_recv_123=0;
int _stat_count_123=0;

/*
 * Begin: Added by Tahsin Kurc
 *
 * The following functions are used to perform off-band communication 
 * between two programs.
 *
 */
int shortBcastOffBand(pgme_s *to, short *data, int nelems, int tag)
{
  if (MyPosPgme(my_pgme__)==0) { /* do the broadcast to other program */
    pvm_initsend(PvmDataDefault);
    pvm_pkshort(data,nelems,1);
    pvm_bcast(to->name,tag);
  }

  /* syncronize all processors after send. */
  pvm_barrier(my_pgme__->name,NumNodePgme(my_pgme__));
  return 1;
}

int intBcastOffBand(pgme_s *to, int *data, int nelems, int tag)
{
  if (MyPosPgme(my_pgme__)==0) { /* do the broadcast to other program */
    pvm_initsend(PvmDataDefault);
    pvm_pkint(data,nelems,1);
    pvm_bcast(to->name,tag);
  }

  /* syncronize all processors after send. */
  pvm_barrier(my_pgme__->name,NumNodePgme(my_pgme__));
  return 1;
}

int floatBcastOffBand(pgme_s *to, float *data, int nelems, int tag)
{
  if (MyPosPgme(my_pgme__)==0) { /* do the broadcast to other program */
    pvm_initsend(PvmDataDefault);
    pvm_pkfloat(data,nelems,1);
    pvm_bcast(to->name,tag);
  }

  /* syncronize all processors after send. */
  pvm_barrier(my_pgme__->name,NumNodePgme(my_pgme__));
  return 1;
}

int doubleBcastOffBand(pgme_s *to, double *data, int nelems, int tag)
{
  if (MyPosPgme(my_pgme__)==0) { /* do the broadcast to other program */
    pvm_initsend(PvmDataDefault);
    pvm_pkdouble(data,nelems,1);
    pvm_bcast(to->name,tag);
  }

  /* syncronize all processors after send. */
  pvm_barrier(my_pgme__->name,NumNodePgme(my_pgme__));
  return 1;
}

int stringBcastOffBand(pgme_s *to, char *data, int nelems, int tag)
{
   if (MyPosPgme(my_pgme__)==0) { /* do the broadcast to other program */
    pvm_initsend(PvmDataDefault);
    pvm_pkstr(data);
    pvm_bcast(to->name,tag);
  }

  /* syncronize all processors after send. */
  pvm_barrier(my_pgme__->name,NumNodePgme(my_pgme__));
  return 1;
} 

int byteBcastOffBand(pgme_s *to, char *data, int nelems, int tag)
{
   if (MyPosPgme(my_pgme__)==0) { /* do the broadcast to other program */
    pvm_initsend(PvmDataDefault);
    pvm_pkbyte(data,nelems,1);
    pvm_bcast(to->name,tag);
  }

  /* syncronize all processors after send. */
  pvm_barrier(my_pgme__->name,NumNodePgme(my_pgme__));
  return 1;
}

/*
 * Implementation of recv operations is closely related to implementation 
 * of broadcast operations above.
 */
int shortRecvOffBand(pgme_s *from, short *data, int nelems, int tag)
{
  int bufid;
  int recvtid;

  recvtid = -1;
  if (from!=NULL)
    recvtid = pvm_gettid(from->name,0);
  bufid = pvm_recv(recvtid,tag);
  pvm_upkshort(data,nelems,1);
  return 1;
}  

int intRecvOffBand(pgme_s *from, int *data, int nelems, int tag)
{
  int bufid;
  int recvtid;

  recvtid = -1;
  if (from!=NULL)
    recvtid = pvm_gettid(from->name,0);
  bufid = pvm_recv(recvtid,tag);
  pvm_upkint(data,nelems,1);
  return 1;
}  

int floatRecvOffBand(pgme_s *from, float *data, int nelems, int tag)
{
  int bufid;
  int recvtid;

  recvtid = -1;
  if (from!=NULL)
    recvtid = pvm_gettid(from->name,0);
  bufid = pvm_recv(recvtid,tag);
  pvm_upkfloat(data,nelems,1);
  return 1;
}  

int doubleRecvOffBand(pgme_s *from, double *data, int nelems, int tag)
{
  int bufid;
  int recvtid;

  recvtid = -1;
  if (from!=NULL)
    recvtid = pvm_gettid(from->name,0);
  bufid = pvm_recv(recvtid,tag);
  pvm_upkdouble(data,nelems,1);
  return 1;
}  

int stringRecvOffBand(pgme_s *from, char *data, int nelems, int tag)
{
  int bufid;
  int recvtid;

  recvtid = -1;
  if (from!=NULL)
    recvtid = pvm_gettid(from->name,0);
  bufid = pvm_recv(recvtid,tag);
  pvm_upkstr(data);
  return 1;
}  

int byteRecvOffBand(pgme_s *from, char *data, int nelems, int tag)
{
  int bufid;
  int recvtid;

  recvtid = -1;
  if (from!=NULL)
    recvtid = pvm_gettid(from->name,0);
  bufid = pvm_recv(recvtid,tag);
  pvm_upkbyte(data,nelems,1);
  return 1;
}  

/*
 * End: Added by Tahsin Kurc
 */

/* short */

int sDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     short *tab;
     int tag;
{
  int proc;
  OneSched_s *Osched; 
  int sendFlag;

  if (sched==NULL) return(0);
  _affirmation_(("to->nbNode==sched->nproc",to->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    { sendFlag=0;
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  sendFlag=1;
	  pvm_initsend(GlueEncoding);
	}
      
      while (Osched!=NULL)
	{
#ifdef _verbose_flag_on_
	  {int ii;
	   for (ii=0;ii<Osched->size;ii++)
	     printf("<%i>",tab[Osched->offset+ii]);
	   for (ii=0;ii<Osched->size;ii++)
	     printf("[%i]",Osched->offset+ii);
	   printf("\n");
  	  }
#endif
	  pvm_pkshort(tab+Osched->offset,Osched->size,1);
	  Osched=Osched->next;
	}     
      if (sendFlag==1)
	{
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
  return(0);
}

int sDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     short *tab;
     int tag;
{  
  int proc;
  OneSched_s *Osched;

  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    {       
      Osched=sched->sched[proc];
      if  (Osched!=NULL)
	{pvm_recv(TidPgme(from,proc),tag);
	 while (Osched!=NULL)
	   {
#ifdef _verbose_flag_on_
	     printf("\n Received: ");
	     printf(" %i elem put %i ",Osched->size,Osched->offset);
#endif
	     pvm_upkshort(tab+Osched->offset,Osched->size,1);
	     Osched=Osched->next;
	   }
       }
    }
  return(0);
}

/* integer */

int iDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     int *tab;
     int tag;
{
  int proc;
  OneSched_s *Osched; 
  int sendFlag;

  if (sched==NULL) return(0);
  _affirmation_(("to->nbNode==sched->nproc",to->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    { sendFlag=0;
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  sendFlag=1;
	  pvm_initsend(GlueEncoding);
	}
      
      while (Osched!=NULL)
	{
#ifdef _verbose_flag_on_
	  {int ii;
	   for (ii=0;ii<Osched->size;ii++)
	     printf("<%i>",tab[Osched->offset+ii]);
	   for (ii=0;ii<Osched->size;ii++)
	     printf("[%i]",Osched->offset+ii);
	   printf("\n");
  	  }
#endif
	  pvm_pkint(tab+Osched->offset,Osched->size,1);
	  Osched=Osched->next;
	}     
      if (sendFlag==1)
	{
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
  return(0);
}

int iDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     int *tab;
     int tag;
{  
  int proc;
  OneSched_s *Osched;

  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    {       
      Osched=sched->sched[proc];
      if  (Osched!=NULL)
	{pvm_recv(TidPgme(from,proc),tag);
	 while (Osched!=NULL)
	   {
#ifdef _verbose_flag_on_
	     printf("\n Received: ");
	     printf(" %i elem put %i ",Osched->size,Osched->offset);
#endif
	     pvm_upkint(tab+Osched->offset,Osched->size,1);
	     Osched=Osched->next;
	   }
       }
    }
  return(0);
}

  

/* float */

int fDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     float *tab;
     int tag;
{
  int proc;
  OneSched_s *Osched; 
  int sendFlag;

  if (sched==NULL) return(0);
  _affirmation_(("to->nbNode==sched->nproc",to->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    { sendFlag=0;
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  sendFlag=1;
	  pvm_initsend(GlueEncoding);
	}
      
      while (Osched!=NULL)
	{
	  pvm_pkfloat(tab+Osched->offset,Osched->size,1);
	  Osched=Osched->next;
	}     
      if (sendFlag==1)
	{
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
  return(0);
}

int fDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     float *tab;
     int tag;
{  
  int proc;
  OneSched_s *Osched;


  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    {       
      Osched=sched->sched[proc];
      if  (Osched!=NULL)
	{pvm_recv(TidPgme(from,proc),tag);
	 while (Osched!=NULL)
	   {
	     pvm_upkfloat(tab+Osched->offset,Osched->size,1);
	     Osched=Osched->next;
	   }
       }
    }
  return(0);
}

/* double */
int dDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     char *tab;
     int tag;
{
  int proc;
  OneSched_s *Osched; 
  int sendFlag;

  if (sched==NULL) return(0);
  _affirmation_(("to->nbNode==sched->nproc",to->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    { sendFlag=0;
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  sendFlag=1;
	  pvm_initsend(GlueEncoding);
	}
      
      while (Osched!=NULL)
	{
/*	  pvm_pkdouble(tab+Osched->offset,Osched->size,1); */	 
/*	  for (ii=0;ii<Osched->size;ii++)
	    pvm_pkbyte(tab+Osched->offset*sizeof(double)+ii*sizeof(double),
		       sizeof(double),1);
*/
	  pvm_pkbyte(tab+Osched->offset*sizeof(double),Osched->size*sizeof(double),1);
	  Osched=Osched->next;
	}     
      if (sendFlag==1)
	{
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
  return(0);
}

int dDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     char *tab;
     int tag;
{  
  int proc;
  OneSched_s *Osched;

  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));

  for (proc=0;proc<sched->nproc;proc++)
    {       
      Osched=sched->sched[proc];
      if  (Osched!=NULL)
	{pvm_recv(TidPgme(from,proc),tag);
	 while (Osched!=NULL)
	   {
	     /*	     pvm_upkdouble(tab+Osched->offset,Osched->size,1); */
/*	     for (ii=0;ii<Osched->size;ii++)
	       pvm_upkbyte(tab+Osched->offset*sizeof(double)+ii*sizeof(double),
			  sizeof(double),1);
*/	     
	     pvm_upkbyte(tab+Osched->offset*sizeof(double),Osched->size*sizeof(double),1);
	     Osched=Osched->next;	    
	   }
       }
    }
  return(0);
}

  


/* double */
/*
int OptdDataMoveSendRecv(pgme,sched_send,tab_send,tag,
			 sched_recv,tab_recv)
     pgme_s  *pgme;
     sched_s *sched_send;
     char    *tab_send;
     int      tag;
     char    *tab_recv;
     sched_s *sched_recv;
{
  int me=MyPosPgme(pgme);
  OneSched_s *Osched_send; 
  OneSched_s *Osched_recv; 
  int proc, proc1;
  int sendFlag;
  int buffer_id,tmp,tid;
  int message_a_recevoir;
  
  int _stat_send_local=0;
  
  int  isize;
  int  tmptmpSize;
  char *tmptmpPt ;
  
  if (sched_send==NULL) return(0);
  if (tmptmpBuffer==NULL) tmptmpBuffer=(char *)malloc(1000000);

  for (proc1=0;proc1<sched_send->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched_send->nproc);
      if (proc!=me)
	{
	  sendFlag=0;
	  Osched_send=sched_send->sched[proc];
	  if (Osched_send!=NULL)
	    {
	      sendFlag=1;
	      pvm_initsend(PvmDataDefault);
	      tmptmpSize=0;
	      tmptmpPt=tmptmpBuffer;	      
	    }

	  while (Osched_send!=NULL)
	    {
	      isize=Osched_send->size * sizeof(double);
	      memcpy(tmptmpPt,tab_send+(Osched_send->offset * sizeof(double)),
		     isize);
	      tmptmpPt   += isize;
	      tmptmpSize += isize;
	      Osched_send=Osched_send->next;
	      _stat_send_local+=Osched_send->size*sizeof(double);
	    }     
	  if (sendFlag==1) 
	    {
	      pvm_pkbyte(tmptmpBuffer,
			 tmptmpSize,
			 1);
	      pvm_send(TidPgme(pgme,proc),tag);
	      _stat_send_local=0;
	    }
	}
    }


  if (sched_send->sched[me]!=NULL)
    {
      Osched_send=sched_send->sched[me];
      Osched_recv=sched_recv->sched[me];
      while (Osched_send!=NULL)
       {
	 memcpy(tab_send+(Osched_recv->offset * sizeof(double)),
		tab_recv+(Osched_send->offset * sizeof(double)),
		Osched_send->size * sizeof(double)
		);
	 Osched_send=Osched_send->next;
	 Osched_recv=Osched_recv->next;
       }
    }

   message_a_recevoir=0;
   for (proc1=0;proc1<sched_recv->nproc;proc1++)
     {
       if (sched_recv->sched[proc1]!=NULL && proc1!=me) 
	 message_a_recevoir++;
     }
   _stat_count_123++;
  
  for (proc1=0;proc1<message_a_recevoir;proc1++)
    { 
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &isize,&tmp,&tid);
      pvm_upkbyte(tmptmpBuffer,isize,1);
      proc=Tid2PosPgme(pgme,tid);
      Osched_recv=sched_recv->sched[proc];
      tmptmpPt=tmptmpBuffer;
      if  (Osched_recv!=NULL)
	{
	 while (Osched_recv!=NULL)
	   {
	     isize=Osched_recv->size * sizeof(double);
	     _stat_recv_123 += Osched_recv->size;
	     memcpy(tab_recv+Osched_recv->offset * sizeof(double),
		    tmptmpPt,
		    isize);
	     tmptmpPt+=isize;
	     Osched_recv=Osched_recv->next;	    
	   }
       }
    }
}
*/
/*
int OptdDataMoveSendRecv(pgme,sched_send,tab_send,tag,
			 sched_recv,tab_recv)
     pgme_s  *pgme;
     sched_s *sched_send;
     char    *tab_send;
     int      tag;
     char    *tab_recv;
     sched_s *sched_recv;
{
  int me=MyPosPgme(pgme);
  OneSched_s *Osched_send; 
  OneSched_s *Osched_recv; 
  int proc, proc1;
  int sendFlag;
  int buffer_id,tmp,tid;
  int message_a_recevoir;

 int _stat_send_local=0;

  if (sched_send==NULL) return(0);

  for (proc1=0;proc1<sched_send->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched_send->nproc);
      if (proc!=me)
	{
	  sendFlag=0;
	  Osched_send=sched_send->sched[proc];
	  if (Osched_send!=NULL)
	    {
	      sendFlag=1;
	      pvm_initsend(PvmDataDefault);
	    }

	  while (Osched_send!=NULL)
	    {
	      pvm_pkbyte(tab_send+(Osched_send->offset * sizeof(double)),
			 Osched_send->size * sizeof(double),1);
	      Osched_send=Osched_send->next;
	      _stat_send_local+=Osched_send->size*sizeof(double);
	    }     
	  if (sendFlag==1) 
	    {
	      pvm_send(TidPgme(pgme,proc),tag);
	      _stat_send_local=0;
	    }
	}
    }


  if (sched_send->sched[me]!=NULL)
    {
      Osched_send=sched_send->sched[me];
      Osched_recv=sched_recv->sched[me];
      while (Osched_send!=NULL)
       {
	 memcpy(tab_send+(Osched_recv->offset * sizeof(double)),
		tab_recv+(Osched_send->offset * sizeof(double)),
		Osched_send->size * sizeof(double)
		);
	
	 Osched_send=Osched_send->next;
	 Osched_recv=Osched_recv->next;
       }
    }

   message_a_recevoir=0;
   for (proc1=0;proc1<sched_recv->nproc;proc1++)
     {
       if (sched_recv->sched[proc1]!=NULL && proc1!=me) 
	 message_a_recevoir++;
       _stat_recv_123 += sched_recv->size[proc1];
     }
   _stat_count_123++;
  
  for (proc1=0;proc1<message_a_recevoir;proc1++)
    {    
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &tmp,&tmp,&tid);
      proc=Tid2PosPgme(pgme,tid);
      Osched_recv=sched_recv->sched[proc];
      if  (Osched_recv!=NULL)
	{
	 while (Osched_recv!=NULL)
	   {
	     pvm_upkbyte(tab_recv+Osched_recv->offset * sizeof(double),
			 Osched_recv->size * sizeof(double) ,1);
	     Osched_recv=Osched_recv->next;	    
	   }
       }
    }
}


*/

/***
int OptdDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     char *tab;
     int tag;
{
  int proc1,proc;
  OneSched_s *Osched; 
  int sendFlag;
  int me=MyPosPgme(my_pgme__);

  if (sched==NULL) return(0);
  _affirmation_(("to->nbNode==sched->nproc",to->nbNode==sched->nproc));

  for (proc1=0;proc1<sched->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched->nproc);
      sendFlag=0;
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  sendFlag=1;
	  pvm_initsend(PvmDataDefault);
	}
      
      while (Osched!=NULL)
	{
	  pvm_pkbyte(tab+(Osched->offset* sizeof(double)),
		     Osched->size * sizeof(double),1);
	  Osched=Osched->next;
	}     
      if (sendFlag==1)
	{
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
  return(0);
}

int OptdDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     char *tab;
     int tag;
{  
  int proc;
  int proc1;

  OneSched_s *Osched;
  int message_a_recevoir=0;
  int tid,buffer_id;
  int tmp;

  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));

  for (proc1=0;proc1<sched->nproc;proc1++)
    {
      if (sched->sched[proc1]!=NULL) 
	message_a_recevoir++;
    }

  for (proc1=0;proc1<message_a_recevoir;proc1++)
    {    
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &tmp,&tmp,&tid);
      proc=Tid2PosPgme(from,tid);
      Osched=sched->sched[proc];
      if  (Osched!=NULL)
	{
	 while (Osched!=NULL)
	   {
	     pvm_upkbyte(tab+Osched->offset * sizeof(double),
			 Osched->size * sizeof(double) ,1);
	     Osched=Osched->next;	    
	   }
       }
    }
  return(0);
}
***/
/* ------------------------------------------------------------------ */
/* double */

char *tmptmpBuffer=NULL;

int OptdDataMoveSendRecv(pgme,sched_send,tab_send,tag,
			 sched_recv,tab_recv)
     pgme_s  *pgme;
     sched_s *sched_send;
     char    *tab_send;
     int      tag;
     char    *tab_recv;
     sched_s *sched_recv;
{
  int me=MyPosPgme(pgme);
  OneSched_s *Osched_send; 
  OneSched_s *Osched_recv; 
  int proc, proc1;
  int sendFlag;
  int buffer_id,tmp,tid;
  int message_a_recevoir;
    
  int  isize;
  int  tmptmpSize;
  char *tmptmpPt ;
  
  if (sched_send==NULL) return(0);
  if (tmptmpBuffer==NULL) tmptmpBuffer=(char *)malloc(1000000);

  for (proc1=0;proc1<sched_send->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched_send->nproc);
      if (proc!=me)
	{
	  Osched_send=sched_send->sched[proc];
	  if (Osched_send!=NULL)
	    {
	      pvm_initsend(PvmDataDefault);
	      tmptmpSize=0;
	      tmptmpPt=tmptmpBuffer;	      
	    
	      while (Osched_send!=NULL)
		{
		  isize=Osched_send->size * sizeof(double);
		  memcpy(tmptmpPt,tab_send+(Osched_send->offset * sizeof(double)),
			 isize);
		  tmptmpPt   += isize;
		  tmptmpSize += isize;
		  Osched_send=Osched_send->next;
		}     
	      pvm_pkbyte(tmptmpBuffer,
			 tmptmpSize,
			 1);
	      pvm_send(TidPgme(pgme,proc),tag);
	    }
	}
    }

  if (sched_send->sched[me]!=NULL)
    {
      Osched_send=sched_send->sched[me];
      Osched_recv=sched_recv->sched[me];
      while (Osched_send!=NULL)
       {
/*
	 memcpy(tab_send+(Osched_recv->offset * sizeof(double)),
		tab_recv+(Osched_send->offset * sizeof(double)),
		Osched_send->size * sizeof(double)
		);
*/
	 memcpy(tab_recv+(Osched_recv->offset * sizeof(double)),
		tab_send+(Osched_send->offset * sizeof(double)),
		Osched_send->size * sizeof(double)
		);

	 Osched_send=Osched_send->next;
	 Osched_recv=Osched_recv->next;
       }
    }

   message_a_recevoir=0;
   for (proc1=0;proc1<sched_recv->nproc;proc1++)
     {
       if (sched_recv->sched[proc1]!=NULL && proc1!=me) 
	 message_a_recevoir++;
     }
  
  for (proc1=0;proc1<message_a_recevoir;proc1++)
    { 
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &isize,&tmp,&tid);
      pvm_upkbyte(tmptmpBuffer,isize,1);
      proc=Tid2PosPgme(pgme,tid);
      Osched_recv=sched_recv->sched[proc];
      tmptmpPt=tmptmpBuffer;
      while (Osched_recv!=NULL)
	{
	  isize=Osched_recv->size * sizeof(double);
	  memcpy(tab_recv+Osched_recv->offset * sizeof(double),
		    tmptmpPt,
		 isize);
	  tmptmpPt+=isize;
	  Osched_recv=Osched_recv->next;	    
	}
    }
}


/* ------------------------------------------------------------------ */

int message_a_recevoir_CRADING;
int boucle1[100];
int boucle2;
int boucle3[100];

CRADING1(pgme,sched_send,tab_send,tag,
			 sched_recv,tab_recv)
     pgme_s  *pgme;
     sched_s *sched_send;
     char    *tab_send;
     int      tag;
     char    *tab_recv;
     sched_s *sched_recv;
{
  int me=MyPosPgme(pgme);
  OneSched_s *Osched_send; 
  OneSched_s *Osched_recv; 
  int proc, proc1;
  int sendFlag;
  int buffer_id,tmp,tid;
  int message_a_recevoir;
  
  int _stat_send_local=0;
  
  int  isize;
  int  tmptmpSize;
  char *tmptmpPt ;

  int i;

  for (i=0;i<100;i++)
    {
      boucle1[i]=0;
      boucle2=0;
      boucle3[i]=0;
    }

  message_a_recevoir_CRADING=0;
  for (proc1=0;proc1<sched_recv->nproc;proc1++)
    {
      if (sched_recv->sched[proc1]!=NULL && proc1!=me) 
	message_a_recevoir_CRADING++;
    }

  for (proc1=0;proc1<sched_send->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched_send->nproc);
      if (proc!=me)
	{
	  Osched_send=sched_send->sched[proc];
	  if (Osched_send!=NULL)
	    {
	      while (Osched_send!=NULL)
		{ boucle1[proc]++;
		  Osched_send=Osched_send->next;
		}
	    }
	}
    }
  
  if (sched_send->sched[me]!=NULL)
    {
      Osched_send=sched_send->sched[me];
      Osched_recv=sched_recv->sched[me];
      while (Osched_send!=NULL)
	{
	  boucle2++;
	  Osched_send=Osched_send->next;
	}
    }
  
  for (proc1=0;proc1<sched_recv->nproc;proc1++)
    {       
      Osched_recv=sched_recv->sched[proc1];
      if  (Osched_recv!=NULL)
	{
	 while (Osched_recv!=NULL)
	   {
	     boucle3[proc1]++;
	     Osched_recv=Osched_recv->next;	    
	   }
       }
    }
}



int OptdDataMoveSendRecvCRADING(pgme,sched_send,tab_send,tag,
			 sched_recv,tab_recv)
     pgme_s  *pgme;
     sched_s *sched_send;
     char    *tab_send;
     int      tag;
     char    *tab_recv;
     sched_s *sched_recv;
{
  int me=MyPosPgme(pgme);
  OneSched_s *Osched_send; 
  OneSched_s *Osched_recv; 
  int proc, proc1;
  int sendFlag;
  int buffer_id,tmp,tid;
  int message_a_recevoir;
  
  int _stat_send_local=0;
  
  int  isize;
  int  tmptmpSize;
  char *tmptmpPt ;

  int i;
  
  if (sched_send==NULL) return(0);
  if (tmptmpBuffer==NULL) tmptmpBuffer=(char *)malloc(1000000);

  for (proc1=0;proc1<sched_send->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched_send->nproc);
      if (proc!=me)
	{
	  Osched_send=sched_send->sched[proc];
	  if (Osched_send!=NULL)
	    {
	      pvm_initsend(PvmDataDefault);
	      tmptmpSize=0;
	      tmptmpPt=tmptmpBuffer;	      
	
	      for (i=0;i<boucle1[proc];i++)
		{
		  isize=Osched_send->size * sizeof(double);
		  memcpy(tmptmpPt,tab_send+(Osched_send->offset * sizeof(double)),
			 isize);
		  tmptmpPt   += isize;
		  tmptmpSize += isize;
		  Osched_send=Osched_send->next;
		  _stat_send_local+=Osched_send->size*sizeof(double);
		}     
	      pvm_pkbyte(tmptmpBuffer,
			 tmptmpSize,
			 1);
	      pvm_send(TidPgme(pgme,proc),tag);
	    }
	}
    }

  if (sched_send->sched[me]!=NULL)
    {
      Osched_send=sched_send->sched[me];
      Osched_recv=sched_recv->sched[me];
      for (i=0;i<boucle2;i++)
       {
	 memcpy(tab_send+(Osched_recv->offset * sizeof(double)),
		tab_recv+(Osched_send->offset * sizeof(double)),
		Osched_send->size * sizeof(double)
		);
	 Osched_send=Osched_send->next;
	 Osched_recv=Osched_recv->next;
       }
    }

  
  for (proc1=0;proc1<message_a_recevoir_CRADING;proc1++)
    { 
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &isize,&tmp,&tid);
      pvm_upkbyte(tmptmpBuffer,isize,1);
      proc=Tid2PosPgme(pgme,tid);
      Osched_recv=sched_recv->sched[proc];
      tmptmpPt=tmptmpBuffer;
      if  (Osched_recv!=NULL)
	{
	  for (i=0;i<boucle3[proc];i++)
	   {
	     isize=Osched_recv->size * sizeof(double);
	     memcpy(tab_recv+Osched_recv->offset * sizeof(double),
		    tmptmpPt,
		    isize);
	     tmptmpPt+=isize;
	     Osched_recv=Osched_recv->next;	    
	   }
       }
    }
}

int OptdDataMoveSendRecvCRADING1(pgme,sched_send,tab_send,tag,
			 sched_recv,tab_recv)
     pgme_s  *pgme;
     sched_s *sched_send;
     char    *tab_send;
     int      tag;
     char    *tab_recv;
     sched_s *sched_recv;
{
  int me=MyPosPgme(pgme);
  OneSched_s *Osched_send; 
  OneSched_s *Osched_recv; 
  int proc, proc1;
  int sendFlag;
  int buffer_id,tmp,tid;
  int message_a_recevoir;
  
  int _stat_send_local=0;
  
  int  isize;
  int  tmptmpSize;
  char *tmptmpPt ;

  int i;
  
  if (sched_send==NULL) return(0);
  if (tmptmpBuffer==NULL) tmptmpBuffer=(char *)malloc(1000000);

  for (proc1=0;proc1<sched_send->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched_send->nproc);
      if (proc!=me)
	{
	  Osched_send=sched_send->sched[proc];
	  if (Osched_send!=NULL)
	    {
	      pvm_initsend(PvmDataDefault);
	      tmptmpSize=0;
	      tmptmpPt=tmptmpBuffer;	      
	
	      for (i=0;i<boucle1[proc];i++)
		{
		  isize=sizeof(double);
		  memcpy(tmptmpPt,tab_send+(Osched_send->offset*sizeof(double)),
			 isize);
		  tmptmpPt   += isize;
		  tmptmpSize += isize;
		  Osched_send=Osched_send->next;
		}     
	      pvm_pkbyte(tmptmpBuffer,
			 tmptmpSize,
			 1);
	      pvm_send(TidPgme(pgme,proc),tag);
	    }
	}
    }

  if (sched_send->sched[me]!=NULL)
    {
      Osched_send=sched_send->sched[me];
      Osched_recv=sched_recv->sched[me];
      for (i=0;i<boucle2;i++)
       {
	 memcpy(tab_send+(Osched_recv->offset * sizeof(double)),
		tab_recv+(Osched_send->offset * sizeof(double)),
		sizeof(double)
		);
	 Osched_send=Osched_send->next;
	 Osched_recv=Osched_recv->next;
       }
    }

  
  for (proc1=0;proc1<message_a_recevoir_CRADING;proc1++)
    { 
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &isize,&tmp,&tid);
      pvm_upkbyte(tmptmpBuffer,isize,1);
      proc=Tid2PosPgme(pgme,tid);
      Osched_recv=sched_recv->sched[proc];
      tmptmpPt=tmptmpBuffer;
      if  (Osched_recv!=NULL)
	{
	  for (i=0;i<boucle3[proc];i++)
	   {
	     isize=sizeof(double);
	     memcpy(tab_recv+Osched_recv->offset * sizeof(double),
		    tmptmpPt,
		    isize);
	     tmptmpPt+=isize;
	     Osched_recv=Osched_recv->next;	    
	   }
       }
    }
}


/*******************************************/
/* double */
int OptdDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     char *tab;
     int tag;
{
  int proc,proc1;
  OneSched_s *Osched; 
  int sendFlag;
  int  isize;
  int  tmptmpSize;
  char *tmptmpPt ;
  int  me=MyPosMyPgme();

  int i;
  
  if (sched==NULL) return(0);
  if (tmptmpBuffer==NULL) tmptmpBuffer=(char *)malloc(2100000);

  for (proc1=0;proc1<sched->nproc;proc1++)
    { 
      proc = (me+proc1+1) % (sched->nproc);
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  pvm_initsend(PvmDataDefault);
	  tmptmpSize=0;
	  tmptmpPt=tmptmpBuffer;	      
	  
	  while (Osched!=NULL)
	    {
	      isize=Osched->size * sizeof(double);
	      memcpy(tmptmpPt,tab+(Osched->offset * sizeof(double)),
		     isize);
	      tmptmpPt   += isize;
	      tmptmpSize += isize;
	      Osched=Osched->next;
	    }     
	  pvm_pkbyte(tmptmpBuffer,
		     tmptmpSize,
		     1);
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
}

int OptdDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     char *tab;
     int tag;
{  
  OneSched_s *Osched;
  
  int message_a_recevoir;
  int proc,proc1;
  int isize,tid;
  int buffer_id;
  char *tmptmpPt;
  int tmp;

  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));
  if (tmptmpBuffer==NULL) tmptmpBuffer=(char *)malloc(2100000);

  message_a_recevoir=0;
  for (proc1=0;proc1<sched->nproc;proc1++)
    {
      if (sched->sched[proc1]!=NULL)
	message_a_recevoir++;
    }
  
  for (proc1=0;proc1<message_a_recevoir;proc1++)
    { 
      buffer_id=pvm_recv(-1,tag);         
      pvm_bufinfo(buffer_id,
		  &isize,&tmp,&tid);
      pvm_upkbyte(tmptmpBuffer,isize,1);
      proc=Tid2PosPgme(from,tid);
      Osched=sched->sched[proc];
      tmptmpPt=tmptmpBuffer;
      while (Osched!=NULL)
	{
	  isize=Osched->size * sizeof(double);
	  memcpy(tab+Osched->offset * sizeof(double),
		 tmptmpPt,
		 isize);
	  tmptmpPt+=isize;
	  Osched=Osched->next;	    
	}
    }    
}

/*
------------------------------------------------------------------------
PVM optimization 2
------------------------------------------------------------------------
*/

#define OPT2_MaxBuffDoubleSize 16384

/* double */
int OPT2_dDataMoveSend(to,sched,tab,tag)
     pgme_s *to;
     sched_s *sched;
     char *tab;
     int tag;
{
  int proc;
  OneSched_s *Osched; 
  int sendFlag;
  double *bufferDoubleTmp;
  int  used;

  return ( dDataMoveSend(to,sched,tab,tag) ); 

#if 0 /* not reached */
  if (sched==NULL) return(0);
  _affirmation_(("to->nbNode==sched->nproc",to->nbNode==sched->nproc));

  bufferDoubleTmp=(double *)malloc(sizeof(double)*OPT2_MaxBuffDoubleSize);

  for (proc=0;proc<sched->nproc;proc++)
    { 
      sendFlag=0;
      used=0;
      Osched=sched->sched[proc];
      if (Osched!=NULL)
	{
	  sendFlag=1;
	  pvm_initsend(GlueEncoding);
	}
      
      while (Osched!=NULL)
	{
	  if (Osched->size+used> OPT2_MaxBuffDoubleSize)
	    {
	      if (Osched->size>OPT2_MaxBuffDoubleSize)
		{
		  fprintf(stderr,"\nWarning: augmenter OPT2_MaxBuffDoubleSize\n");
		  exit(0);
		}
	      pvm_pkbyte((char*)bufferDoubleTmp,
			 used*sizeof(double),1);
	      used=0;	      
	    }
	  memcpy(bufferDoubleTmp+used,tab+Osched->offset*sizeof(double),
		 Osched->size*sizeof(double));
	  used+=Osched->size;
	  Osched=Osched->next;
	}
      
      if (used!=0)
	{
	  pvm_pkbyte((char*)bufferDoubleTmp,
		     used*sizeof(double),
		     1);
	  used=0;
	}
      if (sendFlag==1)
	{
	  pvm_send(TidPgme(to,proc),tag);
	}
    }
  free(bufferDoubleTmp);
  return(0);
#endif 
}

int OPT2_dDataMoveRecv(from,sched,tab,tag)
     pgme_s  *from;
     sched_s *sched;
     char *tab;
     int tag;
{  
  int proc;
  OneSched_s *Osched;
  OneSched_s *tmp_Osched;
  double *bufferDoubleTmp;
  double *pt_bufferDoubleTmp;
  int     allouer=0;
  int used,incr,i;
  int bufid,taille,tmp;
  
  return ( dDataMoveRecv(from,sched,tab,tag) );

#if 0 /* Not reached */
  if (sched==NULL) return(0);
  _affirmation_(("from->nbNode==sched->nproc",from->nbNode==sched->nproc));
  bufferDoubleTmp=NULL;

  for (proc=0;proc<sched->nproc;proc++)
    {
      Osched=sched->sched[proc];
      if  (Osched!=NULL)
	{
	  bufid=pvm_recv(TidPgme(from,proc),tag);
	  pvm_bufinfo(bufid,&taille,&tmp,&tmp);
	  if (allouer<taille)
	    {allouer=taille;
	     if (bufferDoubleTmp!=NULL)
	       free(bufferDoubleTmp);	     
	     bufferDoubleTmp=(double *)malloc(allouer);
	   }	  
	  pt_bufferDoubleTmp=bufferDoubleTmp;
	  while (Osched!=NULL)
	    {
	      memcpy(tab+Osched->offset*sizeof(double),
		     pt_bufferDoubleTmp,
		     Osched->size*sizeof(double));
	      Osched=Osched->next;
	      pt_bufferDoubleTmp+=Osched->size;
	    }	   
       }
    }
  if (bufferDoubleTmp!=NULL)
    free(bufferDoubleTmp);
  return(0);
#endif 
}

/* ---- */

#define DEBUG_BUFFER_01      200
char debugBuffer[DEBUG_BUFFER_01];

sendRecvTiming(recv,buffer_size)
pgme_s  *recv;
int buffer_size;
{

  fprintf(stderr,"\nNe doit pas etre utiliser (sendRecvTiming)\nExit\n");
  exit(0);
  pvm_initsend(GlueEncoding);  
  pvm_pkbyte(debugBuffer,
	     buffer_size,
	     1);
  pvm_send(TidPgme(recv,0),2000);
  pvm_recv(TidPgme(recv,0),2001);
}



RecvSendTiming(send,buffer_size)
pgme_s  *send;
int buffer_size;
{
  fprintf(stderr,"\nNe doit pas etre utiliser (RecvSendTiming)\nExit\n");
  exit(0);

  fprintf(stderr,"\nbufferSize=%i\n",buffer_size);
  _affirmation_((send!=NULL,"send!=NULL"));

  fprintf(stderr,"F1 %i \n",TidPgme(send,0));

  fprintf(stderr,"F2\n");
  PrintPgme(send);
  fprintf(stderr,"F2a\n");

  pvm_recv(TidPgme(send,0),2000);
  fprintf(stderr,"F3\n");

  pvm_upkbyte(debugBuffer,buffer_size,1);
  fprintf(stderr,"F4\n");

  pvm_initsend(GlueEncoding);  
  fprintf(stderr,"F5\n");

  pvm_send(TidPgme(send,0),2001);
  fprintf(stderr,"F6\n");

}

