#include <malloc.h>
#include <memory.h>

#include "regionHpf.h"
#include "regionX.h"
#include "util.h"
#include "global.h"


region_s *
Alloc_R_HPF(nbDim,left,right,stride,distribution,mem_type)
     int nbDim;
     int *left;
     int *right;
     int *stride;
     T_RegionDistribution distribution;
     T_MemOp               mem_type;
{
  region_hpf_s *r=Malloc_s(region_hpf_s);
  int taille[MAX_lbB_DIM];
  int i;

  r->type=R_HPF;
  r->distribution=distribution;
  r->nbDim=nbDim;

  r->mem_type=mem_type;
  switch (r->mem_type){
  case Mem_Copy:
    r->left  =(int *)malloc(sizeof(int) * nbDim);
    r->right =(int *)malloc(sizeof(int) * nbDim);
    r->stride=(int *)malloc(sizeof(int) * nbDim);
    memcpy(r->left  ,left  ,sizeof(int) * nbDim);
    memcpy(r->right ,right ,sizeof(int) * nbDim);
    memcpy(r->stride,stride,sizeof(int) * nbDim);
    break;
  case Mem_Pointer:
    r->left= left;
    r->right=right; 
    r->stride=stride;
    break;
  default:
    warning("Unkown Memory Operation");
    break;
  }

  for (i=0;i<nbDim;i++)
    taille[i]=(right[i]-left[i])/stride[i]+1;
  r->size=1;
  for (i=0;i<nbDim;i++)
    r->size=taille[i]*r->size;
  return((region_s *)r);
}

void 
Free_R_HPF(region)
region_hpf_s *region;
{
  if (region!=NULL)
     switch (region->mem_type){
    case Mem_Copy:
      Free(region->left);
      Free(region->right);
      Free(region->stride);
      break;
    case Mem_Pointer:	
      break;
    default:
      warning("Unkown Memory Operation");
      break;
    }  
  Free(region);  
}


void 
Print_R_HPF(r)
region_hpf_s *r;
{ 
  int i;
#if _verbose_flag_on_
  if (r!=NULL)
    {
      printf("\n");  
      printf("\nRegion type: HPF");
      printf("\n  distribution=%i",r->distribution);
      printf("\n  size=%i ",r->size);
      printf("\n  nbDim=%i",r->nbDim);
      printf("\n  mem_type=%i ",r->mem_type);
      printf("\n  left  =[");
      for (i=0;i<r->nbDim;i++)
	printf(" %3i",r->left[i]);
      printf(" ]");
      printf("\n  right =[");
      for (i=0;i<r->nbDim;i++)
	printf(" %3i",r->right[i]);
      printf(" ]");
      printf("\n  stride=[");
      for (i=0;i<r->nbDim;i++)
	printf(" %3i",r->stride[i]);
      printf(" ]\n");
    }
  else
    printf("\n Region : NULL \n");
#endif 
}

