#include <malloc.h>
#include <memory.h>

#include "util.h"
#include "regionX.h"
#include "regionHpf.h"
#include "regionEnum.h"

void 
Free_Region(region)
     region_s *region;
{
  switch (region->type) {
  case R_Enum : 
    Free_R_Enum(region);
    break;
  case R_HPF:
    Free_R_HPF((region_hpf_s *)region);
    break;
  case R_Unkown:
    NYI();
    break;
  default:
    warning("default unexpected");
    break;
  }
}  

region_s *
Alloc_R_Unkown()
{
  NYI();
}


void 
Print_Region(region)
     region_s *region;
{
  if (region!=NULL)
    {
      switch (region->type) {
      case R_Enum : 
	Print_R_Enum((region_enum_s *)region);
	break;
      case R_HPF:
	Print_R_HPF((region_hpf_s *)region);
	break;
      case R_Unkown:
	NYI();
	break;
      default:
	warning("default unexpected");
	break;
      }
    }
  else
    printf("\n Region : NULL ");
}  

void 
getRegionName(region,name)
     region_s *region;
     char *name;
{
  if (region!=NULL)
    {
      switch (region->type) {
      case R_Enum : 
	strcpy(name,"R_Enum");
	break;
      case R_HPF:
	strcpy(name,"R_HPF");
	break;
      case R_Unkown:
	strcpy(name,"R_Unkown");
	NYI();
	break;
      default:
	strcpy(name," [Error :: Unexpected Region Name] ");
	break;
      }
    }
  else
    strcpy(name," [region == NULL] ");
}  


/* ============================================ */

setOfRegion_s *
Alloc_setOfRegion()
{setOfRegion_s *setR;
 setR=Malloc_s(setOfRegion_s);
 setR->NumRegion=0;
 setR->regions=NULL;
 setR->BeginGlobalOffset=(int *)malloc(sizeof(int) * 1);
 setR->BeginGlobalOffset[0]=0;
 return(setR);
}

setOfRegion_s *
Add_Region_setOfRegion(region,setOfRegion)
     region_s      *region;
     setOfRegion_s *setOfRegion;
{
  if (region==NULL)
    return(setOfRegion);
  if (setOfRegion==NULL) 
    setOfRegion=Alloc_setOfRegion();
  setOfRegion->regions
    =
      (region_s **)Realloc(setOfRegion->regions,
			  (setOfRegion->NumRegion + 1 ) * sizeof(region_s *)
			  );

  _affirmation_(( "setOfRegion->regions!=NULL", setOfRegion->regions!=NULL ));

  setOfRegion->BeginGlobalOffset
    =
      (int *)realloc(setOfRegion->BeginGlobalOffset,
		     ((setOfRegion->NumRegion) +2) * sizeof(int)
		     );
  setOfRegion->regions[setOfRegion->NumRegion]=region;
  setOfRegion->BeginGlobalOffset[setOfRegion->NumRegion+1]
    =
      setOfRegion->BeginGlobalOffset[setOfRegion->NumRegion]+region->size;
  (setOfRegion->NumRegion)++;
  return(setOfRegion);
}




setOfRegion_s *
Add_LocalRegion_setOfRegion(region,setOfRegion,Goffset)
     region_s      *region;
     setOfRegion_s *setOfRegion;
     int Goffset ;
{
  if (region==NULL)
    return(setOfRegion);
  if (setOfRegion==NULL) 
    setOfRegion=Alloc_setOfRegion();
  setOfRegion->regions
    =
      (region_s **)Realloc(setOfRegion->regions,
			  (setOfRegion->NumRegion + 1 ) * sizeof(region_s *)
			  );

  _affirmation_(( "setOfRegion->regions!=NULL", setOfRegion->regions!=NULL ));

  setOfRegion->BeginGlobalOffset
    =
      (int *)realloc(setOfRegion->BeginGlobalOffset,
		     ((setOfRegion->NumRegion) +2) * sizeof(int)
		     );
  setOfRegion->regions[setOfRegion->NumRegion]=region;
  setOfRegion->BeginGlobalOffset[setOfRegion->NumRegion]
    = Goffset;
  setOfRegion->BeginGlobalOffset[setOfRegion->NumRegion+1]
    = Goffset+region->size;
  (setOfRegion->NumRegion)++;
  return(setOfRegion);
}


void 
Free_setOfRegion(setOfRegion)
setOfRegion_s *setOfRegion;
{
 if (setOfRegion!=NULL)
   {
     Free(setOfRegion->regions);
     Free(setOfRegion->BeginGlobalOffset);
     Free(setOfRegion);
   }
}

void 
Print_setOfRegion(s)
setOfRegion_s *s;
{
  int i;

#if _verbose_flag_on_
  printf("\n");
  printf("\nSet Of Region ");
  if (s!=NULL)
    {printf("\n   NumRegion=%i ",s->NumRegion);
     printf("\n   BeginGlobalOffset=[");
     for (i=0;i<s->NumRegion+1;i++)
       printf(" %i",s->BeginGlobalOffset[i]);
     printf(" ]");
     for (i=0;i<s->NumRegion;i++)
       Print_Region(s->regions[i]);     
    }
  else
    printf("Set Of Region : NULL \n");
  printf("\n");
#endif 
}


int SizeSetOfRegion(setOfRegion)
setOfRegion_s *setOfRegion;
{
  _affirmation_(("setOfRegion!=NULL",setOfRegion!=NULL));
  if (setOfRegion->BeginGlobalOffset==NULL)
    return(0);
  return(setOfRegion->BeginGlobalOffset[NbRegionSetOfRegion(setOfRegion)]);
}

int NbRegionSetOfRegion(setOfRegion)
setOfRegion_s *setOfRegion;
{return(setOfRegion->NumRegion);
}

