//	gist_unorderedn.cc
// Copyright (c) 1997, Regents of the University of California
// $Header: /fs/formosa/chialin/work/t2/loader/libgist/src/RCS/gist_unorderedn.C,v 1.1 1999/03/20 18:32:28 chialin Exp $

#ifdef __GNUG__
#pragma implementation "gist_unorderedn.h"
#endif

#include <stdlib.h>
#ifndef LIBGIST
#define SM_SOURCE
#define GIST_C
#include <e_error.h> // this shouldn't be necessary...
#include <sm_int_2.h>
#endif
#include "gist_unorderedn.h"
#include "gist_unordered.h"

rc_t
gist_unorderedn_ext_t::insert(
    gist_p& page,
    const vec_t& key,
    const vec_t& data,
    shpid_t child)
{
    return RC_AUGMENT(page.insert(key, data, page.nrecs(), child));
}

static int
intCmp(
    const void*		a,
    const void*		b)
{
    return *((int *) a) - *((int *) b);
}


rc_t
gist_unorderedn_ext_t::remove(
    gist_p& page,
    int slots[],
    int numSlots)
{
    qsort((void *) slots, numSlots, sizeof(int), intCmp);
    for (int i = numSlots-1; i >= 0; i--) {
        W_DO(page.remove(slots[i]));
    }
    return RCOK;
}

rc_t
gist_unorderedn_ext_t::updateKey(
    gist_p& page,
    int& slot,
    const vec_t& newKey)
{
    const keyrec_t& tup = page.rec(slot);

    // make sure we have enough space
    unsigned int need = align(newKey.len(0) + sizeof(keyrec_t)) - 
        align(page.rec_size(slot));
    if (need > page.usable_space()) return RC(eRECWONTFIT);

    // update item
    shpid_t child = tup.child();
    W_DO(page.remove(slot));
    rc_t status = page.insert(newKey, cvec_t(), slot, child);
    assert(_err_num(status) != eRECWONTFIT);
    return RC_AUGMENT(status);
}

void
gist_unorderedn_ext_t::findMinPen(
    const gist_p& page,
    const vec_t& newKey,
    const vec_t& data,
    int& slot)
{
    int count = page.nrecs();
    gist_penalty_t min_penalty(gist_penalty_t::max_penalty);
    gist_penalty_t penalty;

    // loop through all the entries
    for (int i = 0; i < count; i++) {
	const keyrec_t &tup = page.rec(i);
        ext.penalty((void *) tup.key(), tup.klen(), page.level(), newKey.ptr(0),
	    newKey.len(0), penalty);
	if (penalty < min_penalty) {
	    slot = i;
	    min_penalty = penalty;
	}
    }
}

void
gist_unorderedn_ext_t::search(
    gist_p& page,
    const void* query,
    int matches[],
    int& numMatches)
{
    numMatches = 0;
    int count = page.nrecs();

    for (int i = 0; i < count; i++) {
        const keyrec_t &tup = page.rec(i);
	if (ext.consistent(query, tup.key(), tup.klen(), page.level())) {
	    // we have a matching entry; push it on the stack
	    matches[numMatches] = i;
	    numMatches++;
	}
    }
}

void
gist_unorderedn_ext_t::getKey(
    gist_p& page,
    int slot,
    vec_t& key)
{
    const keyrec_t& tup = page.rec(slot);
    key.set(tup.key(), tup.klen());
}

rc_t
gist_unorderedn_ext_t::pickSplit(
    gist_p& page,
    int rightEntries[],
    int& numRight,
    const vec_t& oldBp,
    vec_t& leftBp,
    vec_t& rightBp,
    const vec_t& entry1,
    bool& oneGoesRight,
    const vec_t& entry2,
    bool& twoGoesRight)
{
    gist_predcursor_t pcursor;
    pcursor.add(page);
    // also add new entry/entries
    pcursor.add(entry1.ptr(0), entry1.len(0));
    if (entry2.size() != 0) {
	pcursor.add(entry2.ptr(0), entry2.len(0));
    }

    int entry1Pos = page.nrecs();
    int entry2Pos = entry1Pos + 1;

    int leftLen, rightLen;
    const void *bp = NULL;
    int bplen = 0;
    ext.pickSplit(pcursor, page.level(), rightEntries, numRight,
        oldBp.ptr(0), oldBp.len(0), leftBp.ptr(0), leftLen, rightBp.ptr(0), rightLen);

    // fix up left/rightBp
    const void* keyptr = leftBp.ptr(0);
    leftBp.set(keyptr, leftLen);
    keyptr = rightBp.ptr(0);
    rightBp.set(keyptr, rightLen);

    // find out where the new items (entry1 and -2) go
    // and clean up the slot array (we must not return the "slot indices" 
    // of entry1/-2)
    oneGoesRight = twoGoesRight = false;
    for (int i = 0; i < numRight; i++) {
        if (rightEntries[i] == entry1Pos || 
		rightEntries[i] == entry2Pos) {
	    if (rightEntries[i] == entry1Pos) {
		oneGoesRight = true;
	    } else {
		twoGoesRight = true;
	    }
	    // compress rightEntries
	    for (int j = i; j < numRight-1; j++) {
	        rightEntries[j] = rightEntries[j+1];
	    }
	    numRight--;
	}
    }
    return RCOK;
}
    
void
gist_unorderedn_ext_t::unionPage(
    gist_p& page,
    vec_t& bp,
    bool& bpChanged)
{
    gist_predcursor_t pcursor;
    pcursor.add(page);
    int bplen = bp.len(0);
    ext.union_page(bp.ptr(0), bplen, page.level(), pcursor, bpChanged);
    const void *ptr = bp.ptr(0);
    bp.set(ptr, bplen);
}

void
gist_unorderedn_ext_t::unionKey(
    vec_t& bp,
    int level, 
    const vec_t& pred,
    bool &bpChanged)
{
    int bplen = bp.len(0);
    ext.union_key(bp.ptr(0), bplen, level, pred.ptr(0), pred.len(0), bpChanged);
    const void *ptr = bp.ptr(0);
    bp.set(ptr, bplen);
}

#ifndef	LIBGIST
void *
gist_unorderedn_ext_t::eqQuery(
    const vec_t& keyv)
{
    vec_t& k = (vec_t&) keyv;
    return ext.eqQuery(k.ptr(0), k.len(0));
}
#endif

bool
gist_unorderedn_ext_t::check(
    const vec_t& bp,
    const vec_t& pred,
    int level)
{
    return ext.check(bp.ptr(0), bp.len(0), pred.ptr(0), pred.len(0), level);
}

void
gist_unorderedn_ext_t::printPred(
    const vec_t& pred,
    int level)
{
    ext.printPred(pred.ptr(0), pred.len(0), level);
}

void
gist_unorderedn_ext_t::printData(
    const vec_t& data)
{
    ext.printData(data.ptr(0), data.len(0));
}
        
gist_unorderedn_ext_t::gist_unorderedn_ext_t(
    gist_ext_t::gist_ext_ids id,
    gist_unordered_ext_t& ext) : ext(ext)
{
    myId = id;
    gist_ext_t::gist_ext_list[myId] = this;
}
