
/* $Id: t2_acc.h,v 1.8 2000/02/04 00:19:36 chialin Exp $ */

/* t2_acc.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/4/98
 * T2 accumulator classes
 */


#ifndef _T2_ACC_H_
#define _T2_ACC_H_

#include	"t2_header.h"
#include	"t2_usrarg.h"
#include	"t2_shape.h"
#include	"t2_arrayindex.h"
#include	"t2_queryinfo.h"


//forward declaration
class T2_AccIterator;
class T2_ClusterInfoList;

//////////////////////////////////////////////////////////////////
//
// T2_Accumulator
//
//////////////////////////////////////////////////////////////////
class T2_Accumulator {
public:
  T2_Accumulator() {};
  virtual ~T2_Accumulator() {};

  // virtual methods
  virtual T2_UDFRet navigateAll(const T2_QueryInfo& qinfo,
				T2_AccIterator*& aitp) = 0;
  virtual T2_UDFRet navigate(const T2_QueryInfo& qinfo, T2_IteratorID i,
			     const T2_Region& r, T2_AccIterator*& aitp) = 0;

protected:
private:
};



//////////////////////////////////////////////////////////////////
//
// T2_AccumulatorDenseArray
//
//////////////////////////////////////////////////////////////////
class T2_AccumulatorDenseArray: public T2_Accumulator {
public:
  T2_AccumulatorDenseArray();
  virtual ~T2_AccumulatorDenseArray();

  // virtual methods (inherited from T2_Accumulator) implemented by T2
  virtual T2_UDFRet navigateAll(const T2_QueryInfo& qinfo,
				T2_AccIterator*& aitp);
  virtual T2_UDFRet navigate(const T2_QueryInfo& qinfo, T2_IteratorID i,
			     const T2_Region& r, T2_AccIterator*& aitp);

  // virtual methods to be implemented by users (ie derived classes)
  virtual void* access(const T2_ArrayIndex&) = 0;
  virtual T2_ArrayIndexSet find(T2_IteratorID i, const T2_Region& r);

protected:
private:
};



//////////////////////////////////////////////////////////////////
//
// T2_AccIterator
//
//////////////////////////////////////////////////////////////////
class T2_AccIterator {
public:
  T2_AccIterator() {};
  virtual ~T2_AccIterator() {};

  // virtual methods
  virtual T2_UDFRet getNextElement(const T2_QueryInfo& qinfo,
				   T2_Accumulator& acc,
				   bool& eod, void*& elem) = 0;

protected:
private:
};



//////////////////////////////////////////////////////////////////
//
// T2_AccMetaObj
//
//////////////////////////////////////////////////////////////////
class T2_AccMetaObj {
public:
  T2_AccMetaObj() {};
  virtual ~T2_AccMetaObj() {};

  // virtual functions
  virtual T2_UDFRet allocAcc(const T2_QueryInfo& qinfo,
			     T2_Iteration i, const T2_Region& r,
			     T2_Accumulator*& accp) = 0;
  virtual T2_UDFRet stripMine(const T2_QueryInfo& qinfo, size_t mem,
			      const T2_ClusterInfoList& infolist,
			      T2_VArray<T2_Region>& oqw) = 0;

protected:
private:
};



//////////////////////////////////////////////////////////////////
//
// T2_AccMetaObjDenseArray
//
//////////////////////////////////////////////////////////////////
class T2_AccMetaObjDenseArray: public T2_AccMetaObj {
public:
  T2_AccMetaObjDenseArray();
  virtual ~T2_AccMetaObjDenseArray();

  // virtual functions (inherited from T2_AccMetaObj) implemented by T2
//  virtual T2_Box getMBR();
  virtual T2_UDFRet allocAcc(const T2_QueryInfo& qinfo,
			     T2_Iteration i, const T2_Region& r,
			     T2_Accumulator*& accp);
  virtual T2_UDFRet stripMine(const T2_QueryInfo& qinfo, size_t mem,
			      const T2_ClusterInfoList& infolist,
			      T2_VArray<T2_Region>& oqw);

  // virtual functions implemented by users (ie in derived classes)
  virtual u_int getNumberDimensions() = 0;
  virtual u_int getDimensionSize(T2_Array<u_long>& nelem) = 0;
  virtual size_t getElementSize() = 0;
  virtual T2_UDFRet partition(const T2_QueryInfo& qinfo, u_int nparts,
			      u_int parts[], T2_Array<T2_Region>& acw) = 0;
};



//////////////////////////////////////////////////////////////////
//
// T2_AccMetaObjUsrDef
//
//////////////////////////////////////////////////////////////////
class T2_AccMetaObjUsrDef: public T2_AccMetaObj {
public:
  T2_AccMetaObjUsrDef(): T2_AccMetaObj() {};
  virtual ~T2_AccMetaObjUsrDef() {};

  // virtual functions inherited from T2_AccMetaObj
  // virtual T2_Box getMBR() = 0;
  virtual T2_UDFRet allocAcc(const T2_QueryInfo& qinfo,
			     T2_Iteration i, const T2_Region& r,
			     T2_Accumulator*& accp) = 0;
  virtual T2_UDFRet stripMine(const T2_QueryInfo& qinfo, size_t mem,
			      const T2_ClusterInfoList& infolist,
			      T2_VArray<T2_Region>& oqw) = 0;
};



//////////////////////////////////////////////////////////////////
//
// T2_AccMetaConstructor
//
//////////////////////////////////////////////////////////////////
typedef T2_UDFRet (T2_AccMetaConstructor)
			(const T2_QueryInfo& qinfo, const T2_Box& output_box,
			 T2_UsrArgReader& arg, T2_AccMetaObj*& amp);
typedef T2_UDFRet (*T2_AccMetaConstructorPtr)
			(const T2_QueryInfo& qinfo, const T2_Box& output_box,
			 T2_UsrArgReader& arg, T2_AccMetaObj*& amp);



#endif	/* _T2_ACC_H_ */
