
/* $Id: t2_aggr.h,v 1.8 2000/02/04 00:19:36 chialin Exp $ */

/* t2_aggr.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/3/98
 * aggregation function classes
 */


#ifndef _T2_AGGR_H_
#define _T2_AGGR_H_


#include	"t2_header.h"
#include	"t2_usrarg.h"
#include	"t2_shape.h"
#include	"t2_accmsg.h"
#include	"t2_cluster.h"
#include	"t2_dataset.h"
#include	"t2_prj.h"
#include	"t2_acc.h"
#include	"t2_output.h"
#include	"t2_queryinfo.h"


////////////////////////////////////////////////////////////////////////
//
// T2_AggregateFuncObj
//
////////////////////////////////////////////////////////////////////////
class T2_AggregateFuncObj {
public:
  T2_AggregateFuncObj() {};
  virtual ~T2_AggregateFuncObj() {};

  // virtual methods for accumulator initialization
  // initializing acc element
  virtual T2_UDFRet aifElem(const T2_QueryInfo& qinfo, void *accElem);
  // initializing entire acc
  virtual T2_UDFRet aifAcc(const T2_QueryInfo& qinfo, T2_Accumulator& acc);

  // virtual methods for local reduction phase
  virtual T2_UDFRet dafElem(const T2_QueryInfo& qinfo, T2_DSID id,
			    const void* dataElem, void* accElem) = 0;
  virtual T2_UDFRet dafAcc(const T2_QueryInfo& qinfo,
			   const T2_Cluster& input, const T2_Box& mbr,
			   T2_ClusterAuxInfoReader& meta,
			   T2_Dataset& dataset, T2_DSID dsid,
			   T2_IteratorID dsiterid,
			   const T2_Box& iqr, const T2_Region& tilereg,
			   T2_ProjectFuncObj& proj,
			   u_int acciterid, T2_Accumulator& acc);

  // virtual methods for global reduction phase
  // (by default, no global combine phase is needed)
  virtual bool needGlobalCombine(const T2_QueryInfo& qinfo, T2_Accumulator& acc)
							{ return false; };
  virtual T2_UDFRet fillAccMsgBuffer(const T2_QueryInfo& qinfo,
				     T2_Accumulator& acc,
				     T2_Array<T2_AccMsgBufferWriter>& msgbuf);
  virtual T2_UDFRet processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
				  T2_AccMsgBufferReader& msgbuf,
				  T2_Accumulator& acc);

  // virtual methods for output finalization phase
  virtual T2_UDFRet finalize(const T2_QueryInfo& QueryInfo, T2_Accumulator& acc,
			     T2_Output*& outp) = 0;

protected:
private:
};


inline T2_UDFRet
T2_AggregateFuncObj::aifElem(const T2_QueryInfo& qinfo, void *accElem)
{
  return T2_UDFRet_OK;
}

inline T2_UDFRet
T2_AggregateFuncObj::fillAccMsgBuffer(const T2_QueryInfo& qinfo,
				      T2_Accumulator& acc,
				      T2_Array<T2_AccMsgBufferWriter>& msgbuf)
{
  return T2_UDFRet_OK;
}

inline T2_UDFRet
T2_AggregateFuncObj::processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
				   T2_AccMsgBufferReader& msgbuf,
				   T2_Accumulator& acc)
{
  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// T2_AggregateFuncConstructor
//
////////////////////////////////////////////////////////////////////////
typedef T2_UDFRet (T2_AggregateFuncConstructor)
			(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			 T2_AggregateFuncObj*& afp);
typedef T2_UDFRet (*T2_AggregateFuncConstructorPtr)
			(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			 T2_AggregateFuncObj*& afp);



////////////////////////////////////////////////////////////////////////
//
// T2_AggregateFuncObjDenseAcc
//
////////////////////////////////////////////////////////////////////////
class T2_AggregateFuncObjDenseAcc {
public:
protected:
private:
};



#endif	/* _T2_AGGR_H_ */
