
/* $Id: t2_arrayindex.h,v 1.3 2000/01/08 17:09:52 chialin Exp $ */

/* t2_arrayindex.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/20/98
 *
 */


#ifndef _T2_ARRAYINDEX_H_
#define _T2_ARRAYINDEX_H_


#include	<iostream.h>
#include	"t2_header.h"
#include	"t2_tarray.h"


////////////////////////////////////////////////////////
//
// T2_ArrayIndex
//
////////////////////////////////////////////////////////
class T2_ArrayIndex {
public:
  T2_ArrayIndex(u_int l=0): index(l) {};
  T2_ArrayIndex(u_int l, const u_int *a);
  T2_ArrayIndex(const T2_ArrayIndex& a);
  ~T2_ArrayIndex() {};

  // const access methods
  u_int getNumberDimensions() const { return index.getNumberElements(); };
  u_int operator[] (u_int i) const { return index[i]; };

  // access methods
  void setNumberDimensions(u_int s) { index.clear(s); };
  u_int& operator[] (u_int i) { return index[i]; };

  // comparison operators
  bool compatible(const T2_ArrayIndex& a) const;
  bool operator ==(const T2_ArrayIndex& a) const;
  bool operator !=(const T2_ArrayIndex& a) const;
  bool operator <(const T2_ArrayIndex& a) const;
  bool operator <=(const T2_ArrayIndex& a) const;
  bool operator >(const T2_ArrayIndex& a) const;
  bool operator >=(const T2_ArrayIndex& a) const;

  // write
  ostream& write(ostream& s) const;

protected:
  T2_Array<u_int> index;

private:
};


inline
T2_ArrayIndex::T2_ArrayIndex(const T2_ArrayIndex& a) : index(a.index)
{
}

inline
T2_ArrayIndex::T2_ArrayIndex(u_int l, const u_int *a) : index(l)
{
  for (u_int i=0; i<l; i++)
    index[i] = a[i];
}

inline bool
T2_ArrayIndex::compatible(const T2_ArrayIndex& a) const
{
  return (getNumberDimensions() == a.getNumberDimensions())? true : false;
}

inline bool
T2_ArrayIndex::operator ==(const T2_ArrayIndex& a) const
{
  if (!compatible(a))
     return false;

  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (index[i] != a[i])
        return false;
  }
  return true;
}

inline bool
T2_ArrayIndex::operator !=(const T2_ArrayIndex& a) const
{
  if (!compatible(a))
     return true;

  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (index[i] != a[i])
        return true;
  }
  return false;
}

inline bool
T2_ArrayIndex::operator <(const T2_ArrayIndex& a) const
{
  if (!compatible(a))
     return false;

  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (index[i] >= a[i])
        return false;
  }
  return true;
}

inline bool
T2_ArrayIndex::operator <=(const T2_ArrayIndex& a) const
{
  if (!compatible(a))
     return false;

  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (index[i] > a[i])
        return false;
  }
  return true;
}

inline bool
T2_ArrayIndex::operator >(const T2_ArrayIndex& a) const
{
  if (!compatible(a))
     return false;

  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (index[i] <= a[i])
        return false;
  }
  return true;
}

inline bool
T2_ArrayIndex::operator >=(const T2_ArrayIndex& a) const
{
  if (!compatible(a))
     return false;

  for (u_int i=0; i<getNumberDimensions(); i++) {
     if (index[i] < a[i])
        return false;
  }
  return true;
}

inline ostream&
T2_ArrayIndex::write(ostream& os) const
{
  const u_int n = getNumberDimensions();
  if (n == 0) {
     os << "<>";
     return os;
  }

  os << "<";
  for (u_int i=0; i<n-1; i++)
     os << index[i] << ",";
  os << index[n-1] << ">";
  return os;
}

inline ostream&
operator<<(ostream& os, const T2_ArrayIndex& a)
{
  return a.write(os);
}




////////////////////////////////////////////////////////
//
// T2_ArrayIndexRange
//
////////////////////////////////////////////////////////
class T2_ArrayIndexRange {
public:
  T2_ArrayIndexRange(): low(0), high(0) {};
  T2_ArrayIndexRange(const T2_ArrayIndex& l, const T2_ArrayIndex& h);
  T2_ArrayIndexRange(const T2_ArrayIndexRange& r);

  // access methods
  const T2_ArrayIndex& getLow() const { return low; };
  T2_ArrayIndex& getLow() { return low; };
  const T2_ArrayIndex& getHigh() const { return high; };
  T2_ArrayIndex& getHigh() { return high; };

  // write
  ostream& write(ostream& os) const;

protected:
  T2_ArrayIndex low, high;

  // check if low <= high ?
  bool valid() const { return (low <= high)? true : false; };

private:
};


inline
T2_ArrayIndexRange::
T2_ArrayIndexRange(const T2_ArrayIndex& l, const T2_ArrayIndex& h)
		: low(l), high(h)
{
  t2_Assert (valid());
}

inline
T2_ArrayIndexRange::
T2_ArrayIndexRange(const T2_ArrayIndexRange& r) : low(r.low), high(r.high)
{
  t2_Assert (valid());
}

inline ostream&
T2_ArrayIndexRange::write(ostream& os) const
{
  os << "Low = " << getLow() << ", High = " << getHigh();
  return os;
}

inline ostream&
operator<<(ostream& os, const T2_ArrayIndexRange& r)
{
  return r.write(os);
}



////////////////////////////////////////////////////////
//
// T2_ArrayIndexSet
//
////////////////////////////////////////////////////////
class T2_ArrayIndexSet: public T2_VArray<T2_ArrayIndexRange> {
public:
  // n is the dimension of the array index, and l is the segment size
  T2_ArrayIndexSet(u_int n, u_int l=20);
  ~T2_ArrayIndexSet() {};

  // access methods
  u_int getNumberIndexRanges() const { return getNumberElements(); };

  // write
  ostream& write(ostream& os) const;

protected:
  u_int ndim;

  void growInit(T2_ArrayIndexRange&);

private:
};

inline
T2_ArrayIndexSet::T2_ArrayIndexSet(u_int l, u_int n)
			: T2_VArray<T2_ArrayIndexRange>(l), ndim(n)
{
}

inline void
T2_ArrayIndexSet::growInit(T2_ArrayIndexRange& a)
{
  a.getLow().setNumberDimensions(ndim);
  a.getHigh().setNumberDimensions(ndim);
}

inline ostream&
T2_ArrayIndexSet::write(ostream& os) const
{
  os << "T2_ArrayIndexSet has " << getNumberIndexRanges()
     << " index ranges" << endl;
  for (u_int i=0; i<getNumberIndexRanges(); i++)
     os << "    " << i << ": " << (*this)[i] << endl;
  return os;
}

inline ostream&
operator<<(ostream& os, const T2_ArrayIndexSet& s)
{
  return s.write(os);
}



#endif	/* _T2_ARRAYINDEX_H_ */
