
/* t2_bcache.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 6/18/98
 * a cache for block requests
 */


#ifndef _T2_BCACHE_H_
#define _T2_BCACHE_H_

#include	"t2_header.h"
#include	"t2_tset.h"


// forward declaration
class T2_PhyBlockRequest;


/////////////////////////////////////////////////////////////////////////
//
// T2_BlockCache
//
// The constructor takes an argument that specifies the cache size.
//
// Note that none of the methods of T2_BlockCache deal with the status
// nor the ref count of the block request. The caller must update them
// appropriately. The method "insert", however, does check to see if
// the block request has more than one static owners. Caching only
// makes sense if it does have more than one static owners. The method
// "insert" returns true if the block is cached (in which case the caller
// should set the status to t2_bCached), or false if the cache decides
// not to cache it but to kill it immediately (in which case the
// method T2_PhyBlockRequest::kill" would set the status to idle).
//
// The current implementation uses the easiest implementation -- a set.
// Probably does not make any sense at all, but it works. I'll change
// the implementation later.
//
/////////////////////////////////////////////////////////////////////////
class T2_BlockCache {
public:
  T2_BlockCache(u_int s): cache(s) {};
  ~T2_BlockCache() {};

  // query methods
  u_int getNumberBlockRequests() const { return cache.getNumberElements(); };
  u_int getCacheSize() const { return cache.getMaxSetSize(); };

  // resurrect a block by removing it from cache without killing it
  void resurrect(T2_PhyBlockRequest* b);

  // try to insert a block into cache; returns true if cache decides to keep
  // it, returns false if cache decides not to keep it and kill it immediately
  bool insert(T2_PhyBlockRequest* b);

protected:
  T2_TSet<T2_PhyBlockRequest*> cache;

  // search for a block request, return the index if found, return
  // getNumberBlockRequests() if not found
  u_int search(T2_PhyBlockRequest* b) const;

  // insert an available block into cache when the cache is full
  void overflow(T2_PhyBlockRequest* b);

private:
};


inline u_int
T2_BlockCache::search(T2_PhyBlockRequest* b) const
{
  t2_Assert (b != NULL);

  u_int i;
  const u_int nb = getNumberBlockRequests();
  for (i=0; i<nb; i++) {
     if (cache[i] == b)
	return i;
  }
  return nb;
}

inline void
T2_BlockCache::resurrect(T2_PhyBlockRequest* b)
{
  t2_Assert (b != NULL);
  u_int i = search(b);
  t2_Assert (i != getNumberBlockRequests());
  cache.out(i);
}




#endif	/* _T2_BCACHE_H_ */
