
/* $Id: t2_bits.h,v 1.10 2000/11/21 23:05:01 chialin Exp $ */

/* t2_bits.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/23/97
 * A bit vector implemented with a long integer. Code adapted from the part
 * of Titan code implemented by Carter T. Shock.
 */

#ifndef _T2_BIT_H_
#define _T2_BIT_H_

#include	<iostream.h>
#include	<fstream.h>
#include	"t2_header.h"
#include	"t2_bstream.h"


//typedef const char *const  BitsNames;

/////////////////////////////////////////////////////
//
// T2_Bits
// A bit vector represented by a long integer.
//
/////////////////////////////////////////////////////
class T2_Bits {
public:
  T2_Bits() { v = 0UL; };
  T2_Bits(u_long l) { v = l; };
  T2_Bits(ifstream &is) { is >> v;is.ignore(255,'\n'); };
  T2_Bits(const T2_Bits &b) { v = b.v; };

  static u_int getNumberBits() { return sizeof(u_long) * 8; };
  u_long getVec() const { return v; };
  u_long& getVec() { return v; };

  bool operator == (const T2_Bits &b) const { return (v == b.v)? true:false; };
  bool operator != (const T2_Bits &b) const { return (v != b.v)? true:false; };

  T2_Bits operator ~() const { return T2_Bits(~v); };

  T2_Bits operator ^(const T2_Bits &b) const { return T2_Bits(v ^ b.v); };
  T2_Bits operator |(const T2_Bits &b) const { return T2_Bits(v | b.v); };
  T2_Bits operator &(const T2_Bits &b) const { return T2_Bits(v & b.v); };

  T2_Bits & operator = (const T2_Bits &b) { v = b.v; return *this; };
  T2_Bits & operator ^= (const T2_Bits &b) { v ^= b.v; return *this; };
  T2_Bits & operator |= (const T2_Bits &b) { v |= b.v; return *this; };
  T2_Bits & operator &= (const T2_Bits &b) { v &= b.v; return *this; };

  char *sprintb(char *s) const;
  char *sprintb(char *s, const char *const *const names, 
		char term = '\n') const;
  ostream & write(ostream &os) const;

  // input/output between a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& ibf);

  T2_Bits & clrAll() { v = 0UL; return *this; };
  T2_Bits & setAll() { v = 0xFFFFFFFFUL; return *this; };
  T2_Bits & set(u_int i);
  T2_Bits & clr(u_int i);
  T2_Bits & tog(u_int i);

  bool isSet(u_int i) const;
  bool isClr(u_int i) const;
  bool allClr() const { if (v == 0) return true; return false; };

  // count number of bits set to one
  u_int getNumberBitsSet() const;

protected:
  u_long v;
private:
};


inline u_int
T2_Bits::getNumberBitsSet() const
{
  // use a static array to keep track of the number of bits set for
  // a given byte, ie byte=0 has 0 bit set to 1, byte=1 has 1 bit set to 1,
  // etc.
  const u_int num_ones[256] = {
		0, 1, 1, 2, 1, 2, 2, 3,
		1, 2, 2, 3, 2, 3, 3, 4,
		1, 2, 2, 3, 2, 3, 3, 4,
		2, 3, 3, 4, 3, 4, 4, 5,
		1, 2, 2, 3, 2, 3, 3, 4,
		2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5,
		3, 4, 4, 5, 4, 5, 5, 6,

		1, 2, 2, 3, 2, 3, 3, 4,
		2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5,
		3, 4, 4, 5, 4, 5, 5, 6,
		2, 3, 3, 4, 3, 4, 4, 5,
		3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6,
		4, 5, 5, 6, 5, 6, 6, 7,

		1, 2, 2, 3, 2, 3, 3, 4,
		2, 3, 3, 4, 3, 4, 4, 5,
		2, 3, 3, 4, 3, 4, 4, 5,
		3, 4, 4, 5, 4, 5, 5, 6,
		2, 3, 3, 4, 3, 4, 4, 5,
		3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6,
		4, 5, 5, 6, 5, 6, 6, 7,

		2, 3, 3, 4, 3, 4, 4, 5,
		3, 4, 4, 5, 4, 5, 5, 6,
		3, 4, 4, 5, 4, 5, 5, 6,
		4, 5, 5, 6, 5, 6, 6, 7,
		3, 4, 4, 5, 4, 5, 5, 6,
		4, 5, 5, 6, 5, 6, 6, 7,
		4, 5, 5, 6, 5, 6, 6, 7,
		5, 6, 6, 7, 6, 7, 7, 8
	};

  u_char *b = (u_char *) &v;
  u_int s = 0;
  for (u_int i=0; i<sizeof(v); i++)
     s += num_ones[(u_int) (b[i])];
  return s;
}

inline ostream & 
T2_Bits::write(ostream &os) const
{ 
  long oldbase;
  oldbase = os.setf(ios::hex,ios::basefield);
  os << "0x" << getVec(); 
  os.setf(oldbase, ios::basefield);
  return os; 
}

inline ostream & operator << (ostream &os, const T2_Bits &b)
{ return b.write(os); }

inline T2_OBStream & operator << (T2_OBStream &obf, const T2_Bits &b)
{ return b.bwrite(obf); }

inline T2_IBStream & operator >> (T2_IBStream &ibf, T2_Bits &b)
{ return b.bread(ibf); };



/////////////////////////////////////////////////////
//
// T2_BitVector
// A long bit vector consisting of multiple T2_Bits
//
/////////////////////////////////////////////////////
class T2_BitVector {
public:
  T2_BitVector(u_int nb=32);
  T2_BitVector(const T2_BitVector& bv);
  virtual ~T2_BitVector() { delete [] bits; };

  // methods to convert a bit vector into an array of u_longs
  u_int getNumberLongs() const { return veclen; };
  u_int getVec(u_long *l) const;

  // access the i-th long in the vec, where i < getNumberLongs();
  // returns 0 when i >= getNumberLongs()
  u_long getLong(u_int i) const;

  // methods to access the bit array
  const T2_Bits* getBitsArray() const { return bits; };
  T2_Bits* getBitsArray() { return bits; };
  u_int getBitsArraySize() const { return veclen; };

  // assign value to T2_BitVector and change the size accordingly
  T2_BitVector& assign(u_int n, u_long *v);
  T2_BitVector& assign(const T2_BitVector& bv);

  u_int getNumberBits() const { return nbits; };
  void setNumberBits(u_int nb);

  bool operator == (const T2_BitVector &b) const;
  bool operator != (const T2_BitVector &b) const;
  T2_BitVector operator ~() const;

  ///////////////////////////////////////////////////////////////////////
  //
  // If the two bit vectors are of different lengths, the output vector
  // will have the same size as the shorter bit vector operand.
  //
  ///////////////////////////////////////////////////////////////////////
  T2_BitVector operator ^(const T2_BitVector &b) const;
  T2_BitVector operator |(const T2_BitVector &b) const;
  T2_BitVector operator &(const T2_BitVector &b) const;

  ///////////////////////////////////////////////////////////////////////
  //
  // If the two bit vectors are of different lengths, the operator will
  // abort.
  //
  ///////////////////////////////////////////////////////////////////////
  T2_BitVector & operator = (const T2_BitVector &b);
  T2_BitVector & operator ^= (const T2_BitVector &b);
  T2_BitVector & operator |= (const T2_BitVector &b);
  T2_BitVector & operator &= (const T2_BitVector &b);

  virtual char *sprintb(char *s) const;
//  virtual char *sprintb(char *s, const char *const *const names,
//                       char term = '\n') const;
  virtual ostream & write(ostream &os) const;

  // input/output between a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& ibf);

  T2_BitVector & clrAll();
  T2_BitVector & setAll();
  T2_BitVector & set(u_int i);
  T2_BitVector & clr(u_int i);
  T2_BitVector & tog(u_int i);

  bool isSet(u_int i) const;
  bool isClr(u_int i) const;
  bool allClr() const;

  // count number of bits set to one
  // (caller must call clrExcessBits() first if excess bits may be 
  // accidentally set to one)
  u_int getNumberBitsSet() const;

  // clear all excess bits
  void clrExcessBits();

protected:
  T2_Bits *bits;
  u_int   veclen,		// number of T2_Bits (ie size of bits[])
          nbits;		// number of bits
private:
  u_long  excess_bits_mask;	// a mask used to set all the excess bits
				// in the last T2_Bits in bits[] to zero
};


inline u_int
T2_BitVector::getVec(u_long *l) const
{
  t2_Assert (veclen == 0 || l != NULL);
  for (u_int i=0; i<veclen; i++)
     l[i] = bits[i].getVec();
  return veclen;
}


inline u_long 
T2_BitVector::getLong(u_int i) const
{
  if (i >= getNumberLongs())
     return 0;
  return bits[i].getVec();
}


inline bool
T2_BitVector::operator == (const T2_BitVector &b) const
{
  for (u_int i=0; i<veclen; i++)
     if (bits[i] != b.bits[i])
	return false;
  return true;
}

inline bool
T2_BitVector::operator != (const T2_BitVector &b) const
{
  for (u_int i=0; i<veclen; i++)
     if (bits[i] == b.bits[i])
	return false;
  return true;
}

inline T2_BitVector
T2_BitVector::operator ~() const
{
  T2_BitVector bv(nbits);
  for (u_int i=0; i<veclen; i++)
     bv.bits[i] = ~(this->bits[i]);
  return bv;
}

inline T2_BitVector
T2_BitVector::operator ^(const T2_BitVector &b) const
{
  u_int l, n;
  if (veclen < b.veclen) {
     l = veclen; n = nbits;
  }
  else {
     l = b.veclen; n = b.nbits;
  }
  T2_BitVector bv(n);
  for (u_int i=0; i<l; i++)
     bv.bits[i] = bits[i] ^ b.bits[i];
  return bv;
}

inline T2_BitVector
T2_BitVector::operator |(const T2_BitVector &b) const
{
  u_int l, n;
  if (veclen < b.veclen) {
     l = veclen; n = nbits;
  }
  else {
     l = b.veclen; n = b.nbits;
  }
  T2_BitVector bv(n);
  for (u_int i=0; i<l; i++)
      bv.bits[i] = bits[i] | b.bits[i];
  return bv;
}

inline T2_BitVector
T2_BitVector::operator &(const T2_BitVector &b) const
{
  u_int l, n;
  if (veclen < b.veclen) {
     l = veclen; n = nbits;
  }
  else {
     l = b.veclen; n = b.nbits;
  }
  T2_BitVector bv(n);
  for (u_int i=0; i<l; i++)
     bv.bits[i] = bits[i] & b.bits[i];
  return bv;
}


inline T2_BitVector &
T2_BitVector::operator =(const T2_BitVector &b)
{
  t2_Assert(this->nbits == b.nbits);
  for (u_int i=0; i<veclen; i++)
      bits[i] = b.bits[i];
  return *this;
}

inline T2_BitVector &
T2_BitVector::operator ^= (const T2_BitVector &b)
{
  t2_Assert(this->nbits == b.nbits);
  for (u_int i=0; i<veclen; i++)
     bits[i] ^= b.bits[i];
  return *this;
}

inline T2_BitVector &
T2_BitVector::operator |= (const T2_BitVector &b)
{
  t2_Assert(this->nbits == b.nbits);
  for (u_int i=0; i<veclen; i++)
     bits[i] |= b.bits[i];
  return *this;
}

inline T2_BitVector &
T2_BitVector::operator &= (const T2_BitVector &b)
{
  t2_Assert(this->nbits == b.nbits);
  for (u_int i=0; i<veclen; i++)
     bits[i] &= b.bits[i];
  return *this;
}

inline T2_BitVector&
T2_BitVector::assign(u_int n, u_long *v)
{
  if (veclen != n) {		// change size accordinly
     delete [] bits;
     t2_Alloc (bits, n, T2_Bits);
     veclen = n;
  }
  nbits = veclen * T2_Bits::getNumberBits();
  for (u_int i=0; i<veclen; i++)
     bits[i] = T2_Bits(v[i]);
  return *this;
}

inline T2_BitVector&
T2_BitVector::assign(const T2_BitVector &b)
{
  if (veclen != b.veclen) {	// change size accordinly
     delete [] bits;
     veclen = b.veclen;
     t2_Alloc (bits, veclen, T2_Bits);
  }
  nbits = b.nbits;
  for (u_int i=0; i<veclen; i++)
     bits[i] = b.bits[i];
  return *this;
}

inline ostream&
T2_BitVector::write(ostream &os) const
{
  for (u_int i=0; i<veclen; i++)
     os << bits[i] << " ";
  return os;
}


inline T2_BitVector &
T2_BitVector::clrAll()
{
  for (u_int i=0; i<veclen; i++)
     bits[i].clrAll();
  return *this;
}

inline T2_BitVector & 
T2_BitVector::setAll()
{
  for (u_int i=0; i<veclen; i++)
     bits[i].setAll();
  return *this;
}

inline T2_BitVector &
T2_BitVector::set(u_int i)
{
  u_int a = i / T2_Bits::getNumberBits(),
  b = i % T2_Bits::getNumberBits();
  bits[a].set(b);
  return *this;
}

inline T2_BitVector &
T2_BitVector::clr(u_int i)
{
  u_int a = i / T2_Bits::getNumberBits(),
	b = i % T2_Bits::getNumberBits();
  bits[a].clr(b);
  return *this;
}

inline T2_BitVector &
T2_BitVector::tog(u_int i)
{
  u_int a = i / T2_Bits::getNumberBits(),
	b = i % T2_Bits::getNumberBits();
  bits[a].tog(b);
  return *this;
}

inline bool
T2_BitVector::isSet(u_int i) const
{
  u_int a = i / T2_Bits::getNumberBits(),
	b = i % T2_Bits::getNumberBits();
  return bits[a].isSet(b);
}

inline bool
T2_BitVector::isClr(u_int i) const
{
  u_int a = i / T2_Bits::getNumberBits(),
	b = i % T2_Bits::getNumberBits();
  return bits[a].isClr(b);
}

inline bool
T2_BitVector::allClr() const
{
  for (u_int i=0; i<veclen; i++)
     if (bits[i].allClr() == false)
	return false;
  return true;
}

inline u_int
T2_BitVector::getNumberBitsSet() const
{
  if (veclen == 0)
     return 0;
  t2_Assert (bits != NULL);

  u_int s = 0;
  for (u_int i=0; i<veclen; i++)
     s += bits[i].getNumberBitsSet();
  return s;
}

inline void
T2_BitVector::clrExcessBits()
{
  if (excess_bits_mask == 0 || veclen == 0)		// no excess bits
     return;

  t2_Assert (veclen > 0);
  bits[veclen-1].getVec() &= excess_bits_mask;
}

inline
ostream& operator<<(ostream& os, const T2_BitVector& b)
{ return b.write(os); }


inline T2_OBStream &
operator << (T2_OBStream &obf, const T2_BitVector &b)
{ return b.bwrite(obf); }

inline T2_IBStream &
operator >> (T2_IBStream &ibf, T2_BitVector &b)
{ return b.bread(ibf); };



///////////////////////////////////////////////////////
//
// T2_ProcBit
// A long bit vector for the back-end nodes, one bit
// per back-end node.
//
///////////////////////////////////////////////////////
class T2_ProcBits : public T2_BitVector {
public:
  T2_ProcBits() : T2_BitVector(T2_MAX_NUM_BE_PROCS) {};
protected:
private:
};

#endif		/* _T2_BITS_H_ */

