
/* t2_clusterinfo.h */


#ifndef _T2_CLUSTERINFO_H_
#define _T2_CLUSTERINFO_H_


#include	"t2_header.h"
#include	"t2_clusterreq.h"
#include	"t2_shape.h"
#include	"t2_qenv.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_ClusterInfoList
//
// A class used to access a T2_ClusterRequestList (ie a list of
// T2_ClusterRequest's) as a list of bounding boxes.
//
//////////////////////////////////////////////////////////////////////////
class T2_ClusterInfoList {
public:
  T2_ClusterInfoList(const T2_QEnv& q): qenv(q) {};
  ~T2_ClusterInfoList() {};

  // access methods -- only need constant onces
  u_int getNumberDatasets() const { return qenv.getNumberDatasets(); };
  T2_DSID getDatasetID(u_int d) const;
  u_int getNumberClusters(u_int d) const;
  const T2_Box& getBoundingBox(u_int d, u_int i) const;
  bool getMetaData(u_int d, u_int i, T2_UsrArgReader& reader) const;
  bool getClusterInfo(u_int d, u_int i, T2_Box& box, T2_UsrArgReader& reader)
								const;

protected:
  const T2_QEnv& qenv;

private:
  // do not use copy constructor and operator =
  T2_ClusterInfoList(const T2_ClusterInfoList& clist);
  T2_ClusterInfoList& operator =(const T2_ClusterInfoList& clist);

};


inline T2_DSID
T2_ClusterInfoList::getDatasetID(u_int d) const
{
  t2_Assert (d < getNumberDatasets());
  return qenv.getDatasetEnv(d).getDatasetID();
}

inline u_int
T2_ClusterInfoList::getNumberClusters(u_int d) const
{
  t2_Assert (d < getNumberDatasets());
  return
     qenv.getDatasetEnv(d).getClusterRequestList().getNumberClusterRequests();
}

inline const T2_Box&
T2_ClusterInfoList::getBoundingBox(u_int d, u_int i) const
{
  t2_Assert (d < getNumberDatasets());
  t2_Assert (i < getNumberClusters(d));
  const T2_ClusterRequestList& reqlist =
			qenv.getDatasetEnv(d).getClusterRequestList();
  return reqlist[i].getMBR();
}

inline bool
T2_ClusterInfoList::getMetaData(u_int d, u_int i, T2_UsrArgReader& reader) const
{
  if (d < getNumberDatasets() && i < getNumberClusters(d)) {
     const T2_ClusterRequestList& reqlist =
			qenv.getDatasetEnv(d).getClusterRequestList();
     reader.open(reqlist[i].getMetaData());
     return true;
  }
  return false;
}

inline bool
T2_ClusterInfoList::getClusterInfo(u_int d, u_int i,
				   T2_Box& box, T2_UsrArgReader& reader) const
{
  if (d < getNumberDatasets() && i < getNumberClusters(d)) {
     const T2_ClusterRequestList& reqlist =
			qenv.getDatasetEnv(d).getClusterRequestList();
     const T2_ClusterRequest& req = reqlist[i];

     const T2_Box& b = req.getMBR();
     box.setNumberDimensions(b.getNumberDimensions());
     box = b;

     reader.open(req.getMetaData());
     return true;
  }
  return false;
}

inline
T2_ClusterInfoList::T2_ClusterInfoList(const T2_ClusterInfoList& clist):
							qenv(clist.qenv)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_ClusterInfoList copy constructor");
}

inline T2_ClusterInfoList&
T2_ClusterInfoList::operator =(const T2_ClusterInfoList& clist)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_ClusterInfoList operator =");
  return *this;
}




#endif	/* _T2_CLUSTERINFO_H_ */
