
/* $Id: t2_diskassign.h,v 1.2 2000/03/14 23:35:15 chialin Exp $ */

/* t2_diskassign.h */

#ifndef _T2_DISKASSIGN_H_
#define _T2_DISKASSIGN_H_

#include	"t2_header.h"
#include	"t2_txtfile.h"
#include	"t2_bits.h"
#include	"t2_idrange.h"


////////////////////////////////////////////////////////////////////////////
//
// T2_DiskAssignmentFile
//
// A file used to access a disk assignment file, which is used to
// reassign logical processor ids or logical disk ids to local processor,
// by either hostname or rank.
//
////////////////////////////////////////////////////////////////////////////
class T2_DiskAssignmentFile {
public:
  enum assignment_type {
	assign_error = 0,		// error
	assign_disk_ids,		// assign logical disk ids
	assign_proc_ids			// assign logical processor ids
  };

  T2_DiskAssignmentFile();
  ~T2_DiskAssignmentFile();

  // access methods
  assignment_type getAssignmentType() const { return type; };
  const T2_BitVector& getDiskIDAssignment() const { return diskmap; };
  const T2_BitVector& getProcIDAssignment() const { return procmap; };

  // query methods
  operator const void* () const { return (type == assign_error)? NULL:this; };

  // open and close an assignment file
  void open(const char* hostname, u_int rank, const char* fname);
  void close();

  // output to ostream for debugging purpose
  ostream& write(ostream& os) const;

protected:
  assignment_type type;
  T2_BitVector diskmap,		// only valid when type is assign_disk_ids
	       procmap;		// only valid when type is assign_proc_ids

private:
  // various read methods, which eventually calls the generic read method
  bool readHostname2Diskid(T2_TxtFile& file, const char* hostname);
  bool readHostname2Procid(T2_TxtFile& file, const char* hostname);
  bool readRank2Diskid(T2_TxtFile& file, const u_int rank);
  bool readRank2Procid(T2_TxtFile& file, const u_int rank);

  // a generic read function that handle all cases
  bool read(T2_TxtFile& file, const char* hostname, const u_int rank,
	    const char* id_name, T2_BitVector& bitmap);

};


inline
T2_DiskAssignmentFile::T2_DiskAssignmentFile()
		: type(assign_error), diskmap(0), procmap(0)
{
}

inline void
T2_DiskAssignmentFile::close()
{
  type = assign_error;
  diskmap.setNumberBits(0);
  procmap.setNumberBits(0);
}

inline
T2_DiskAssignmentFile::~T2_DiskAssignmentFile()
{
  close();
}

inline bool
T2_DiskAssignmentFile::
readHostname2Diskid(T2_TxtFile& file, const char* hostname)
{
  if (hostname == NULL) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: null hostname.");
     return false;
  }
  type = assign_disk_ids;
  return read(file, hostname, 0, "disk", diskmap);
}

inline bool
T2_DiskAssignmentFile::
readHostname2Procid(T2_TxtFile& file, const char* hostname)
{
  if (hostname == NULL) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: null hostname.");
     return false;
  }
  type = assign_proc_ids;
  return read(file, hostname, 0, "processor", procmap);
}

inline bool
T2_DiskAssignmentFile::
readRank2Diskid(T2_TxtFile& file, const u_int rank)
{
  type = assign_disk_ids;
  return read(file, NULL, rank, "disk", diskmap);
}

inline bool
T2_DiskAssignmentFile::
readRank2Procid(T2_TxtFile& file, const u_int rank)
{
  type = assign_proc_ids;
  return read(file, NULL, rank, "processor", procmap);
}

inline ostream&
operator << (ostream& os, const T2_DiskAssignmentFile& f)
{
  return f.write(os);
}



#endif	/* ! _T2_DISKASSIGN_H_ */
