
/* t2_dsmap.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/8/98
 * A class for an associative map for datasets.
 */


#ifndef _T2_DSMAP_H_
#define _T2_DSMAP_H_


#include	"t2_header.h"
#include	"t2_map.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_DSFileMap
//
// T2_DSMap is a class using the data file id as the key and the struct
// that contains a file descriptor and a global disk number as the value
// for a T2 dataset.
//
//////////////////////////////////////////////////////////////////////////
class T2_DSFileMap {
public:
  T2_DSFileMap(u_int l = 100): map(l) {};

  // quey methods
  bool isEmpty() { return map.isEmpty(); };
  u_int getNumberElements() const { return map.getNumberElements(); };
  bool getFileInfo(T2_DSFileID i, int& fd, u_int& d) const;

  // access methods based on position in map (for iterating through map)
  T2_DSFileID getFileIDByPosition(u_int i) const;
  int getFileDescriptorByPosition(u_int i) const;
  u_int getDiskByPosition(u_int i) const;

  // insert an dataset id, file descriptor pair into map
  void insert(T2_DSFileID id, int f, u_int d) { map.insert(id,FileDisk(f,d)); };

  // empty map
  void emptyMap() { map.emptyMap(); };

  // of max map size ...
  u_int getMaxMapSize() const { return map.getMaxMapSize(); };
  void growMaxMapSize(u_int newlen) { map.growMaxMapSize(newlen); };

protected:
  struct FileDisk { int fd; u_int disk; 
		    FileDisk(): fd(-1), disk(0) {};
		    FileDisk(int f, u_int d): fd(f), disk(d) {};
		  };
  T2_PMap<T2_DSFileID, FileDisk> map;

private:
};


inline bool
T2_DSFileMap::getFileInfo(T2_DSFileID i, int& f, u_int& d) const
{
  FileDisk fds;
  if (map.search(i, fds)) {
     f = fds.fd;
     d = fds.disk;
     return true;
  }
  return false;
}

inline T2_DSFileID
T2_DSFileMap::getFileIDByPosition(u_int i) const
{
  return map.getKeyByPosition(i);
}

inline int
T2_DSFileMap::getFileDescriptorByPosition(u_int i) const
{
  return map.getValueByPosition(i).fd;
}

inline u_int
T2_DSFileMap::getDiskByPosition(u_int i) const
{
  return map.getValueByPosition(i).disk;
}




#endif	/* _T2_DSMAP_H_ */
