
/* $Id: t2_output.h,v 1.4 2000/02/04 00:19:36 chialin Exp $ */

/* t2_output.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/12/98
 * The base class for final output of a T2 query
 */


#ifndef _T2_OUTPUT_H_
#define _T2_OUTPUT_H_


#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_outbuf.h"
#include	"t2_queryinfo.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_OutputDataPtrList
//
///////////////////////////////////////////////////////////////////////////
class T2_OutputDataPtrList {
public:
  T2_OutputDataPtrList(u_int l=50): ptrs(l) {};
  ~T2_OutputDataPtrList() {};

  // const access methods
  u_int getNumberPointers() const { return ptrs.getNumberElements(); };
  const char* getDataPointer(u_int i) const;
  size_t getDataSize(u_int i) const;

  // non-const access methods
//  char* getDataPointer(u_int i);
  void insertDataPointer(const char* p, size_t s) { ptrs << Buf(p, s); };

  // clear all pointers
  void removeAllPointers() { ptrs.clear(); };

protected:
  struct Buf { const char* data; size_t nbytes;
		Buf() { data = NULL; nbytes = 0; };
		Buf(const char* p, size_t s) { data = p; nbytes = s; };
		~Buf() {};
  };

  T2_VArray<Buf> ptrs;

private:
};


inline const char*
T2_OutputDataPtrList::getDataPointer(u_int i) const
{
  if (i < getNumberPointers())
     return ptrs[i].data;
  return NULL;
}

inline size_t
T2_OutputDataPtrList::getDataSize(u_int i) const
{
  if (i < getNumberPointers())
     return ptrs[i].nbytes;
  return 0;
}

//inline char*
//T2_OutputDataPtrList::getDataPointer(u_int i)
//{
//  if (i < getNumberPointers())
//     return ptrs[i].data;
//  return NULL;
//}




///////////////////////////////////////////////////////////////////////////
//
// T2_Output
//
///////////////////////////////////////////////////////////////////////////
class T2_Output {
public:
  T2_Output() {};
  virtual ~T2_Output() {};

  // virtual methods
  virtual size_t getOutputBufferSize(const T2_QueryInfo& qinfo) = 0;
  virtual T2_UDFRet flushOutput(const T2_QueryInfo& qinfo, T2_OutputBuffer& buf,
				T2_OutputDataPtrList& ptr) = 0;

protected:
private:
};



#endif	/* _T2_OUTPUT_H_ */
