
/* $Id: t2_system.h,v 1.4 2000/03/16 00:46:24 chialin Exp $ */

/* t2_system.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 2/9/98
 * T2 run-time system
  */


#ifndef _T2_SYSTEM_H_
#define _T2_SYSTEM_H_


#include	"t2_header.h"


////////////////////////////////////////////////////////////
//
// T2_System
//
// A class to store information about the T2 system.
//
////////////////////////////////////////////////////////////
class T2_System {
public:
  T2_System();
  ~T2_System() {};

  // operator =
  T2_System& operator =(const T2_System& s);

  // access the processor information
  u_int getNumberNodes() const { return numnodes; };
  T2_ProcID getMyNode() const { return mynode; };
  u_char getBackEndEndian() const { return endian; };
  void setProcessorInfo(u_int nn, T2_ProcID me) { numnodes = nn; mynode = me; };

  // access the program number for meta-chaos
  const char* getMCProgramName() const { return mcpname; };
  const char* setMCProgramName(const char* n) { return (mcpname = n); };

protected:
  u_int numnodes;
  T2_ProcID mynode;
  u_char endian;			// endian-ness of the local machine
  const char* mcpname;			// program name for meta-chaos

private:
};


inline
T2_System::T2_System()
{
  numnodes = 1;
  mynode = 0;
  endian = t2_checkEndianess();
}

inline T2_System&
T2_System::operator =(const T2_System& s)
{
  numnodes = s.numnodes;
  mynode = s.mynode;
  return *this;
}



#endif	/* _T2_SYSTEM_H_ */
