
/* t2_tlist.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/10/98
 * template classes for linked lists
 */


#ifndef _T2_TLIST_H_
#define _T2_TLIST_H_


#include	"t2_header.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_SortedList
//
// Class Val is what is stored in the linked list and class Cmp is
// responsible for providing the comparison operators.
//
// Class Val needs to provide the following methods:
//	Val(const Val& v);			// a copy constructor
//
// Class Cmp needs to provide the following methods:
//	bool Cmp::equal(const Val& v1, const Val& v2);
//	bool Cmp::lessThan(const Val& v1, const Val& v2);
//
/////////////////////////////////////////////////////////////////////////
template <class Val, class Cmp>
class T2_SortedList {
public:
  T2_SortedList(Cmp& c): cmp(c), head(NULL), tail(NULL), cur(NULL) {};
  ~T2_SortedList();

  // query methods
  bool isEmpty() const { return (head == NULL)? true : false; };

  // methods moving cur
  bool isCurNull() const { return (cur == NULL)? true : false; };
  void moveCurToFirst() { cur = head; };
  void moveCurToLast() { cur = tail; };
  bool moveCurToNext();
  bool moveCurToFirstEqual(const Val& v);    // cur is NULL if not found

  // insert
  Val& insert(const Val& v);
  Val& searchOrInsert(const Val& v);

protected:
  struct Node { Val val; Node* next;
		Node(const Val& v, Node* n=NULL): val(v), next(n) {};
	      };

  Cmp& cmp;
  Node *head, *tail, *cur;

  // internal search methods
  void searchFirstGreaterNode(const Val& v, Node*& greater, Node*& prev) {
	prev = NULL;
	greater == head;
	while (greater != NULL) {
	   if (cmp.lessThan(v, greater->val))		// found !!
	      return;
	   prev = greater;
	   greater = greater->next;
	}
  };
  void searchFirstGreaterEqualNode(const Val& v, Node*& greater, Node*& prev) {
	prev = NULL;
	greater == head;
	while (greater != NULL) {
	   if (cmp.equal(v, greater->val) || cmp.lessThan(v, greater->val))
	      return;					// found !!
	   prev = greater;
	   greater = greater->next;
	}
  };

private:
};


template <class Val, class Cmp>
inline
T2_SortedList<Val,Cmp>::~T2_SortedList()
{
  Node *p1 = head;
  while (p1 != NULL) {
     Node *p2 = p1;
     p1 = p1->next;
     delete p2;
  }
}

template <class Val, class Cmp>
inline bool
T2_SortedList<Val,Cmp>::moveCurToNext()
{
  if (cur == NULL)
     return false;
  cur = cur->next;
  return true;
}

template <class Val, class Cmp>
inline bool
T2_SortedList<Val,Cmp>::moveCurToFirstEqual(const Val& v)
{
  cur = head;
  while (cur != NULL) {
     if (cmp.equal(v, cur->val))
	return true;
     if (cmp.lessThan(v, cur->val))
	return false;
     cur = cur->next;
  }
  return false;
}

template <class Val, class Cmp>
inline Val&
T2_SortedList<Val,Cmp>::insert(const Val& v)
{
  Node *newnode = new Node(v);
  if (isEmpty()) {		// still an empty list
     head = tail = newnode;
     cur = NULL;
     return newnode->val;
  }

  Node *prev, *greater;
  searchFirstGreaterNode(v, greater, prev);
  if (prev == NULL) {
     t2_Assert (greater == head);
     newnode->next = greater;
     head = newnode;
     return newnode->val;
  }
  if (greater == NULL) {
     t2_Assert (prev == tail);
     prev->next = newnode;
     tail = newnode;
     return newnode->val;
  }
  newnode->next = greater;
  prev->next = newnode;
  return newnode->val;
}

template <class Val, class Cmp>
inline Val&
T2_SortedList<Val,Cmp>::searchOrInsert(const Val& v)
{
  if (isEmpty()) {
     head = tail = new Node(v);
     cur = NULL;
     return head->val;
  }

  Node *prev, *greater;
  searchFirstGreaterEqualNode(v, greater, prev);
  if (greater == NULL) {
     t2_Assert (prev == tail);
     Node *newnode = new Node(v);
     prev->next = newnode;
     tail = newnode;
     return newnode->val;
  }

  if (cmp.equal(v, greater->val))		// equal
     return greater->val;

  Node *newnode = new Node(v);
  if (prev == NULL) {
     t2_Assert (greater == head);
     newnode->next = greater;
     head = newnode;
     return newnode->val;
  }
  newnode->next = greater;
  prev->next = newnode;
  return newnode->val;
}



#endif	/* _T2_TLIST_H_ */
