
/* clusterreq.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/10/98
 */


#include	<iostream.h>
#include	"t2_clusterreq.h"


//static const u_int nOwnersPerBlockRequest = 5;	// segment size of owners[]


//////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockOwnerInfo methods
//
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockRequest static members
//
//////////////////////////////////////////////////////////////////////
//T2_PhyBlockRequest::IBVCmp T2_PhyBlockRequest::cmp;
T2_BlockCache* T2_PhyBlockRequest::bcache = NULL;


//////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockRequest methods
//
//////////////////////////////////////////////////////////////////////
T2_PhyBlockRequest::T2_PhyBlockRequest():
		fd(-1), offset(0), nbytes(0),
		buf(), status(t2_bIdle),
		ownerinfo(), refc(0)
//		owners(nOwnersPerBlockRequest),
//		iterations(T2_PhyBlockRequest::cmp)
{
}


T2_PhyBlockRequest::T2_PhyBlockRequest(int i, off_t f, size_t s):
		fd(i), offset(f), nbytes(s),
		buf(), status(t2_bIdle),
		ownerinfo(), refc(0)
//		owners(nOwnersPerBlockRequest),
//		iterations(T2_PhyBlockRequest::cmp)
{
}


T2_PhyBlockRequest::T2_PhyBlockRequest(const T2_PhyBlockRequest& r):
		fd(r.fd), offset(r.offset), nbytes(r.nbytes),
		buf(), status(t2_bIdle),
		ownerinfo(), refc(0)
//		owners(nOwnersPerBlockRequest),
//		iterations(T2_PhyBlockRequest::cmp)
{
  // make sure that both owners and iterations are both empty
  t2_Assert (r.hasNoBuffer());
  t2_Assert (hasNoBuffer());
  t2_Assert (r.getStatus() == t2_bIdle);
  t2_Assert (getStatus() == t2_bIdle);
//  t2_Assert (owners.getNumberElements() == 0);
//  t2_Assert (r.owners.getNumberElements() == 0);
//  t2_Assert (iterations.isEmpty());
//  t2_Assert (r.iterations.isEmpty());
}


T2_PhyBlockRequest&
T2_PhyBlockRequest::operator =(const T2_PhyBlockRequest& r)
{
  fd = r.fd;
  offset = r.offset;
  nbytes = r.nbytes;
  status = r.status;

  // make sure that both owners and iterations are both empty
  t2_Assert (r.hasNoBuffer());
  t2_Assert (hasNoBuffer());
  t2_Assert (r.getStatus() == t2_bIdle);
//  t2_Assert (owners.getNumberElements() == 0);
//  t2_Assert (r.owners.getNumberElements() == 0);
//  t2_Assert (iterations.isEmpty());
//  t2_Assert (r.iterations.isEmpty());

  return *this;
}


T2_PhyBlockRequestStatus
T2_PhyBlockRequest::touch(T2_ClusterRequest* p)
{
  t2_Assert (p != NULL);
  switch (status) {
     case t2_bIdle:
		t2_Assert (hasNoBuffer() == true);
		t2_Assert (refc == 0);

		// try to allocate buffer
		if (allocBuffer() == 0) {	// allocBuffer failed
		   return t2_bIdle;		// remain idle
		}
		else {				// allocBuffer succeeded
		   // add p to the list of current owners
		   ownerinfo.addCurrentOwner(p);

		   // change status to t2_bActivated
		   return (status = t2_bActivated);
		}
		break;

     case t2_bActivated:
		t2_Assert (hasNoBuffer() == false);
		t2_Assert (refc == 0);
		t2_Assert (ownerinfo.getNumberCurrentOwners() == 1);
		t2_Assert (p == ownerinfo.getCurrentOwner(0));
		return t2_bActivated;		// remain activated
		break;

     case t2_bPending:
		t2_Assert (hasNoBuffer() == false);
		t2_Assert (refc == 0);
		t2_Assert (ownerinfo.getNumberCurrentOwners() > 0);

		// add p to the list of current owners
		ownerinfo.addCurrentOwner(p);
		return t2_bPending;		// remain pending
		break;

     case t2_bAvailable:
		t2_Assert (hasNoBuffer() == false);
		t2_Assert (refc > 0);
		refc++;
		return t2_bAvailable;		// remain available

     case t2_bCached:
		t2_Assert (hasNoBuffer() == false);
		t2_Assert (refc == 0);

		// inform cache of its new status here !!!
		t2_Assert (bcache != NULL);
		bcache->resurrect(this);

		refc++;
		return (status = t2_bAvailable);	// become available
		break;
  }

  t2_err_quit(__FILE__, __LINE__, "touch detects unexpected status = %d",
	     (int) status);
  return t2_bIdle;
}




//int
//T2_PhyBlockRequest::findOwner(T2_ClusterRequest* p)
//{
//  u_int cnt = 0;
//  owners.getFirstSegment();
//  while (!owners.currentIsNull()) {
//     const u_int n = owners.getNumberElementsInCurrentSegment();
//     for (u_int i=0; i<n; i++, cnt++) {
//	if (owners.getCurrentSegmentElement(i) == p)	// found !!
//	   return cnt;
//     }
//     owners.getNextSegment();
//  }
//  return -1;		// not found !!
//}
//
//
//void
//T2_PhyBlockRequest::setOwner(T2_ClusterRequest* p)
//{
//  // first make sure that p is not already in the list
//  if (findOwner(p) >= 0) {
//     t2_err_ret(__FILE__, __LINE__,
//	       "T2_PhyBlockRequest::setOwners find pointer %d already in the list when invoked",
//	       (int) p);
//	   return;
//  }
//  owners << p;
//}
//
//
//bool
//T2_PhyBlockRequest::setIteration(T2_ClusterRequest* p, T2_Iteration i)
//{
//  // look for p in owners[]
//  int n = findOwner(p);
//  if (n < 0) {
//     t2_err_ret(__FILE__, __LINE__,
//		"T2_PhyBlockRequest::setIteration cannot find owner pointer %d",
//		(int) p);
//     return false;
//  }
//
//  // look for the bit vector for i in iterations
//  IterationBitVector& ibv = iterations.searchOrInsert(IterationBitVector(i));
//  T2_BitVector& bv = ibv.bitvector;
//  const u_int nb = owners.getNumberElements();
//  if (bv.getNumberBits() < nb) {
//     t2_Assert (bv.getNumberBits() == 0);
//     bv.setNumberBits(nb);
//  }
//  t2_Assert (bv.isClr(n));
//  bv.set(n);
//
//  return true;
//}



//////////////////////////////////////////////////////////////////////
//
// T2_PhyBlockRequestList methods
//
//////////////////////////////////////////////////////////////////////
T2_PhyBlockRequest&
T2_PhyBlockRequestList::touch(int fdesc, off_t offset, size_t nbytes)
{
  // search and see if a block request already exists
  reqs.getFirstSegment();
  while (!reqs.currentIsNull()) {
    const u_int n = reqs.getNumberElementsInCurrentSegment();
    for (u_int i=0; i<n; i++) {
	T2_PhyBlockRequest &breq = reqs.getCurrentSegmentElement(i);
	if (breq.getFileDescriptor() == fdesc &&
	    breq.getOffset() == offset &&
	    breq.getNumberBytes() == nbytes)
	{					// found !!
	   return breq;
	}
    }
    reqs.getNextSegment();
  }
  reqs << T2_PhyBlockRequest(fdesc, offset, nbytes);
  return reqs[reqs.getNumberElements()-1];
}


int
T2_PhyBlockRequestList::search(int fdesc, off_t offset, size_t nbytes)
{
  int cnt = 0;
  reqs.getFirstSegment();
  while (!reqs.currentIsNull()) {
    const u_int n = reqs.getNumberElementsInCurrentSegment();
    for (u_int i=0; i<n; i++, cnt++) {
	T2_PhyBlockRequest &breq = reqs.getCurrentSegmentElement(i);
	if (breq.getFileDescriptor() == fdesc &&
	    breq.getOffset() == offset &&
	    breq.getNumberBytes() == nbytes)
	{					// found !!
	   return cnt;
	}
    }
    reqs.getNextSegment();
  }
  return -1;
}

void
T2_PhyBlockRequestList::initBlockRequests()
{
  int cnt = 0;
  reqs.getFirstSegment();
  while (!reqs.currentIsNull()) {
    const u_int n = reqs.getNumberElementsInCurrentSegment();
    for (u_int i=0; i<n; i++, cnt++) {
	T2_PhyBlockRequest &breq = reqs.getCurrentSegmentElement(i);
	breq.resetCurrentOwnerSet();
    }
    reqs.getNextSegment();
  }
}



//////////////////////////////////////////////////////////////////////
//
// T2_ClusterRequest methods
//
//////////////////////////////////////////////////////////////////////
T2_ClusterRequest::T2_ClusterRequest(u_int l): blocks(l), metadata()
{
  refc = 0;
  datasetenv = NULL;
}



//////////////////////////////////////////////////////////////////////
//
// T2_ClusterRequestList methods
//
//////////////////////////////////////////////////////////////////////
