
/* constructors.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/16/98
 * An example code to show how to include various constructors into an
 * T2_RTS object.
 */


#include	"t2_header.h"
#include	"t2_rts.h"


//////////////////////////////////////////////////////////////////////////
//
// This is an example code for including various constructors (ie for
// the accumulator meta objects, the aggregation function objects,
// the projection function objects, the dataset objects and the index
// objects) into a T2_RTS object. Ideally, it would be nice if these
// constructors can be included into T2 automatically. But before I
// know how to do that, we'll just have to do it by hand.
//
// There is a function for including each type of constructors into a
// list. Each of those functions consists of two steps:
// 1) tell the list how many constructors will be included,
// 2) add those constructors to the list one by one.
//
//////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////
//
// to include constructors for accumulator meta objects
//
//////////////////////////////////////////////////////////////////////////
void
initAccMetaConstructors(T2_Array<T2_AccMetaConstructorPtr>& acc)
{
  T2_AccMetaConstructor t2_defAccMetaConstructor;
  // specify how many constructors are expected (1 in this example)
  acc.clear(1);

  // now add the constructors one by one
  acc[0] = t2_defAccMetaConstructor;
}



//////////////////////////////////////////////////////////////////////////
//
// to include constructors for aggregation function objects
//
//////////////////////////////////////////////////////////////////////////
void
initAggregateFuncConstructors(T2_Array<T2_AggregateFuncConstructorPtr>& aggr)
{
  T2_AggregateFuncConstructor t2_defAggrFuncConstructor;

  // specify how many constructors are expected (1 in this example)
  aggr.clear(1);

  // now add the constructors one by one
  aggr[0] = t2_defAggrFuncConstructor;
}



//////////////////////////////////////////////////////////////////////////
//
// to include constructors for projection function objects
//
//////////////////////////////////////////////////////////////////////////
void
initProjectFuncConstructors(T2_Array<T2_ProjectFuncConstructorPtr>& prj)
{
  T2_ProjectFuncConstructor t2_defPrjFuncConstructor;

  // specify how many constructors are expected (1 in this example)
  prj.clear(1);

  // now add the constructors one by one
  prj[0] = t2_defPrjFuncConstructor;
}



//////////////////////////////////////////////////////////////////////////
//
// to include constructors for dataset objects
//
//////////////////////////////////////////////////////////////////////////
void
initDatasetConstructors(T2_Array<T2_DatasetConstructorPtr>& ds)
{
  T2_DatasetConstructor t2_defDatasetConstructor;

  // specify how many constructors are expected (1 in this example)
  ds.clear(1);

  // now add the constructors one by one
  ds[0] = t2_defDatasetConstructor;
}



//////////////////////////////////////////////////////////////////////////
//
// to include constructors for index objects
//
//////////////////////////////////////////////////////////////////////////
void
initIndexConstructors(T2_Array<T2_IndexConstructorPtr>& index)
{
  T2_IndexConstructor t2_defIndexConstructor;

  // specify how many constructors are expected (1 in this example)
  index.clear(1);

  // now add the constructors one by one
  index[0] = t2_defIndexConstructor;
}



#ifdef T2_METACHAOS
//////////////////////////////////////////////////////////////////////////
//
// to include constructors for T2_MetaChaosUser objects
//
//////////////////////////////////////////////////////////////////////////
void
initMetaChaosUserConstructors(T2_Array<T2_MetaChaosUserConstructorPtr>& mcu)
{
  T2_MetaChaosUserConstructorPtr t2_defMetaChaosUserConstructor;

  // specify how many constructors are expected (1 in this example)
  mcu.clear(1);

  // now add the constructors one by one
  mcu[0] = t2_defMetaChaosUserConstructor;
}

#endif
