
/* $Id: prj.C,v 1.4 2000/02/04 00:20:11 chialin Exp $ */

/* prj.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/25/98
 * projection methods
 */


#include	<iostream.h>
#include	"t2_prj.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_ProjectFuncObj methods
//
//////////////////////////////////////////////////////////////////////////
T2_UDFRet
T2_ProjectFuncObj::projectBox(const T2_System& system, const T2_Box& inbox,
			      bool& succeed, T2_Box& outbox)
{
  const u_int input_ndim = getNumberInputDimensions();
  const u_int output_ndim = getNumberOutputDimensions();

  t2_Assert (input_ndim == inbox.getNumberDimensions());

  T2_Box iqr(input_ndim);
  t2_setFullBox(iqr);

  T2_Region oqw(output_ndim, 1);
  oqw << T2_Box(output_ndim);
  t2_setFullBox(oqw[0]);

  T2_Region rg1(output_ndim),
	    rg2(output_ndim);
  bool b1, b2;
  T2_UDFRet rv1 = project(system, iqr, oqw, inbox.getLow(), b1, rg1),
            rv2 = project(system, iqr, oqw, inbox.getHigh(), b2, rg2);
  if (rv1 != T2_UDFRet_OK || rv2 != T2_UDFRet_OK)
     return T2_UDFRet_ERROR;

  if (b1 == false || b2 == false) {
     succeed = false;
     return T2_UDFRet_OK;
  }

  T2_Box mbr1, mbr2;
  rg1.getMBR(mbr1);		// mbr of rg1
  rg2.getMBR(mbr2);		// mbr of rg2

  outbox.setNumberDimensions(output_ndim);
  T2_Point& low = outbox.getLow();
  T2_Point& high = outbox.getHigh();
  for (u_int d=0; d<output_ndim; d++) {
     low[d] = t2_min((mbr1.getLow())[d], (mbr2.getLow())[d]);
     high[d] = t2_max((mbr1.getHigh())[d], (mbr2.getHigh())[d]);
  }

  t2_say5(cout, "T2_ProjectFuncObj projects " << inbox
		<< endl << "into " << outbox << endl);

  succeed = true;
  return T2_UDFRet_OK;
}
