
/* $Id: register-constructors.C,v 1.7 2000/11/10 02:01:10 chialin Exp $ */

/* register-constructors.C */


#include	<unistd.h>
#include	<sys/types.h>
#include	<regex.h>
#include	<strings.h>
#include	<string.h>
#include	<errno.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_custom.h"
#include	"t2_conscat.h"


//////////////////////////////////////////////////////////////////////////
//
// prototypes
//
//////////////////////////////////////////////////////////////////////////
int
match(const char *string, char *pattern);
bool
process_files(T2_UdfConsRegistrationFile& regfile,
	      T2_ConsCatalogWriter& conscat, T2_UdfConstructorFile& consfile);
u_int
search_name_add(const char* target, T2_VArray<char*>& list);

void
free_names(T2_VArray<char*>& list);


//////////////////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  if (argc < 4) {
     cout << "Usage: " << argv[0]
	  << " <cons registration file> <cons catalog file> <C++ file>"
	  << endl
	  << "where" << endl
	  << "<cons registration file> is the input constructor registration file"
	  << endl
	  << "<cons catalog file> is the output constructor catalog file" << endl
	  << "<C++ file> is the output constructor C++ code to be linked with the back-end"
	  << endl;
     return -1;
  }

  // try to open the constructor registration file
  T2_UdfConsRegistrationFile regfile(argv[1]);
  if (!regfile) {
      cout << "ERROR: cannot open '" << argv[1] << "'" << endl;
      return -1;
  }

  // try to open the constructor catalog file
  T2_ConsCatalogWriter conscat(argv[2]);
  if (!conscat) {
      cout << "ERROR: cannot open '" << argv[2] << "'" << endl;
      regfile.close();
      return -1;
  }

  // try to open the constructor C++ file
  T2_UdfConstructorFile consfile(argv[3]);
  if (!consfile) {
      cout << "ERROR: cannot open '" << argv[3] << "'" << endl;
      regfile.close();
      conscat.close();
      return -1;
  }

  // go through the constructor registration file and create a constructor
  // catalog file and a constructor C++ file
  bool r = process_files(regfile, conscat, consfile);
  regfile.close();
  conscat.close();
  consfile.close();

  if (r == false) {
     // some thing wrong happens ...
     // assuming that conscat and consfile would both rename the original
     // files as filename.bak, here I am just gonna remove the newly created
     // files
     if (unlink(argv[2]) != 0) {
	ifstream f(argv[2]);
	if (f)			// no complaint if argv[2] does not exist
	   t2_err_ret(__FILE__, __LINE__, "WARNING: cannot remove '%s'",
			argv[2]);
     }
     if (unlink(argv[3]) != 0) {
	ifstream f(argv[3]);
	if (f)			// no complaint if argv[3] does not exist
	   t2_err_ret(__FILE__, __LINE__, "WARNING: cannot remove '%s'",
			argv[3]);
     }
  }

  return 0;
}



//////////////////////////////////////////////////////////////////////////
//
// match
//
// A function that matches a regular expression against a target string.
// It returns whatever regexec returns (ie 0 means match, REG_NOMATCH
// for no match, REG_BADPAT for invalid regular expression, etc.).
//
//////////////////////////////////////////////////////////////////////////
int
match(const char *string, char *pattern)
{
  int  status;
  regex_t   re;

  if ((status = regcomp(&re, pattern, REG_EXTENDED|REG_NOSUB|REG_ICASE)) != 0) {
       return status;	/* report error */
  }
  status = regexec(&re, string, (size_t) 0, NULL, 0);
  regfree(&re);
  return status;
}


//////////////////////////////////////////////////////////////////////////
//
// process_files
//
//////////////////////////////////////////////////////////////////////////
bool
process_files(T2_UdfConsRegistrationFile& regfile,
	      T2_ConsCatalogWriter& conscat, T2_UdfConstructorFile& consfile)
{
  t2_Assert (regfile);
  t2_Assert (conscat);
  t2_Assert (consfile);

  // varialbe to remember all the function names to enforce unique names
  T2_VArray<char*> dataset_fnames(10),
		   index_fnames(10),
		   accmeta_fnames(10),
		   aggr_fnames(10),
		   prj_fnames(10),
		   metachaos_fnames(10);

// NOT ANY MORE
//  // id=0 is always reserved for default constructor
//  u_int i=0;
//  if (consfile.addEntry(T2_UDF_Dataset, "t2_defDatasetConstructor", i)
//	== false || 
//      consfile.addEntry(T2_UDF_Index, "t2_defIndexConstructor", i)
//	== false ||
//      consfile.addEntry(T2_UDF_AccMeta, "t2_defAccMetaConstructor", i)
//	== false ||
//      consfile.addEntry(T2_UDF_Aggregation, "t2_defAggrFuncConstructor", i)
//	== false ||
//      consfile.addEntry(T2_UDF_Projection, "t2_defPrjFuncConstructor", i)
//	== false ||
//      consfile.addEntry(T2_UDF_MetaChaos, "t2_defMetaChaosUserConstructor", i)
//	== false)
//  {
//      t2_err_ret(__FILE__, __LINE__,
//		"ERROR: fail to register default constructors with the constructor file");
//      return false;
//  }

  int eflag;
  u_int id;
  T2_UdfConsRegistrationEntry *entry;
  while ((entry = regfile.getNextEntry(eflag)) != NULL) {
     if (entry->type == T2_UDF_Unknown) {
	delete entry;
	free_names(dataset_fnames);
	free_names(index_fnames);
	free_names(accmeta_fnames);
	free_names(aggr_fnames);
	free_names(prj_fnames);
	free_names(metachaos_fnames);
	return false;
     }

     // check for duplicated name, and if not found, add it to the list
     u_int search_result;
     switch (entry->type) {
	case T2_UDF_Dataset:
		search_result = search_name_add(entry->name, dataset_fnames);
		break;
	case T2_UDF_Index:
		search_result = search_name_add(entry->name, index_fnames);
		break;
	case T2_UDF_AccMeta:
		search_result = search_name_add(entry->name, accmeta_fnames);
		break;
	case T2_UDF_Aggregation:
		search_result = search_name_add(entry->name, aggr_fnames);
		break;
	case T2_UDF_Projection:
		search_result = search_name_add(entry->name, prj_fnames);
		break;
	case T2_UDF_MetaChaos:
		search_result = search_name_add(entry->name, metachaos_fnames);
		break;
	default:
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: invalid constructor function type %d",
			(int) entry->type);
		search_result = 2;
     }
     switch (search_result) {
	case 0:	// did not find duplicate name
		break;
	case 1:	// found duplicate name
		t2_err_ret(__FILE__, __LINE__,
		    "ERROR: %s constructor function name '%s' is not a unique name",
		    t2_UDFType2shortstring(entry->type), entry->name);
		break;
	case 2:	// some error
		break;
	default: // invalid value
		t2_err_quit(__FILE__, __LINE__,
			"ERROR: invalid value for search_result: %d",
			search_result);
		break;
     }
     if (search_result > 0) {		// some error
	delete entry;
	free_names(dataset_fnames);
	free_names(index_fnames);
	free_names(accmeta_fnames);
	free_names(aggr_fnames);
	free_names(prj_fnames);
	free_names(metachaos_fnames);
	return false;
     }

     if (consfile.addEntry(entry->type, entry->function, id) == false) {
	delete entry;
	free_names(dataset_fnames);
	free_names(index_fnames);
	free_names(accmeta_fnames);
	free_names(aggr_fnames);
	free_names(prj_fnames);
	free_names(metachaos_fnames);
	return false;
     }

     if (conscat.addEntry(entry->type, id, entry->name, entry->description,
			  entry->iterators)
	 == false)
     {
	delete entry;
	free_names(dataset_fnames);
	free_names(index_fnames);
	free_names(accmeta_fnames);
	free_names(aggr_fnames);
	free_names(prj_fnames);
	free_names(metachaos_fnames);
	return false;
     }

     delete entry;
  }			// while ((entry = regfile.getNextEntry(eflag)) != NULL)

  // is there at least one for every type of construction function?
  errno = 0;
  if (dataset_fnames.getNumberElements() == 0)
     t2_err_ret(NULL, 0,
		"WARNING: no dataset constructor function was registered");
  if (index_fnames.getNumberElements() == 0)
     t2_err_ret(NULL, 0,
		"WARNING: no index constructor function was registered");
  if (accmeta_fnames.getNumberElements() == 0)
     t2_err_ret(NULL, 0,
		"WARNING: no accmulator meta-data constructor function was registered");
  if (aggr_fnames.getNumberElements() == 0)
     t2_err_ret(NULL, 0,
		"WARNING: no aggregation constructor function was registered");
  if (prj_fnames.getNumberElements() == 0)
     t2_err_ret(NULL, 0,
		"WARNING: no projection constructor function was registered");
# ifdef T2_METACHAOS
     if (metachaos_fnames.getNumberElements() == 0)
        t2_err_ret(NULL, 0,
		"WARNING: no meta-chaos constructor function was registered");
# endif

  free_names(dataset_fnames);
  free_names(index_fnames);
  free_names(accmeta_fnames);
  free_names(aggr_fnames);
  free_names(prj_fnames);
  free_names(metachaos_fnames);

  if (eflag > 0) {	// error
     return false;
  }

  return true;
}



//////////////////////////////////////////////////////////////////////////
//
// seach_name_add
//
// Look for a target name, and if not found, just add it to the list.
// Returns 0 if target not found and add succeeded,
//         1 if target found,
//         2 some error occurred
//
//////////////////////////////////////////////////////////////////////////
u_int
search_name_add(const char* target, T2_VArray<char*>& list)
{
  if (target == NULL)
     return 2;

  for (u_int i=0; i<list.getNumberElements(); i++) {
     t2_Assert (list[i] != NULL);
     if (strcmp(target, list[i]) == 0)		// a match !
	return 1;
  }

  // target not found, so add targete to the list
  char*& p = list.appendElement();
  size_t len = strlen(target)+1;
  t2_Alloc(p, len, char);
  memcpy(p, target, len);
  return 0;
}


//////////////////////////////////////////////////////////////////////////
//
// free_names
//
// Release memory from the given array.
//
//////////////////////////////////////////////////////////////////////////
void
free_names(T2_VArray<char*>& list)
{
  for (u_int i=0; i<list.getNumberElements(); i++) {
     t2_Assert (list[i] != NULL);
     delete [] list[i];
  }
}
