
/* $Id: register-datasets.C,v 1.10 2000/03/15 00:36:46 chialin Exp $ */

/* register-datasets.C */


#include	<stdio.h>
#include	<unistd.h>
#include	<strings.h>
#include	<string.h>
#include	<errno.h>
#include	"t2_header.h"
#include	"t2_dscat.h"
#include	"t2_idxcat.h"
#include	"t2_conscat.h"
#include	"t2_dsinfo.h"


//////////////////////////////////////////////////////////////////////////
//
// prototypes
//
//////////////////////////////////////////////////////////////////////////
bool
process_files(T2_BEConfigFile& beconfig, T2_DSRegistrationFileReader& dsreg,
	      T2_ConsCatalogReader& conscat, T2_DSInfoFileWriter& dsinfo,
	      T2_DSCatalogWriter& dscat, T2_IndexCatalogWriter& idxcat);
u_int
search_name_add(const char* target, T2_VArray<T2_CharPtr>& list);


//////////////////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  if (argc < 7) {
     cout << "Usage: " << argv[0]
	  << " <backend config file> <dataset registration file>"
	  << " <cons catalog file>"
	  << endl
	  << "	<dataset info file> <dataset catalog> <index catalog>"
	  << endl
	  << "where" << endl
	  << "<backend config file> is the input back-end configuration file"
	  << endl
	  << "<dataset registration file> is the input dataset registration file"
	  << endl
	  << "<cons catalog file> is the input constructor catalog file" << endl
	  << "<dataset info file> is the output dataset information file to be used by the front-end"
	  << endl
	  << "<dataset catalog> is the output dataset catalog file to be used "
	  << endl
	  << "	by the back-end"
	  << endl
	  << "<index catalog> is the output index catalog file to be used by the back-end"
	  << endl;
     return -1;
  }

  bool r = true;
  u_int ac = 0;
  // try to open the back-end registration file and verify it
  T2_BEConfigFile beconfig(argv[++ac]);
  if (!beconfig) {
      return -1;
  }
  if (beconfig.check_content_correct() == false) {
     return -1;
  }

  // try to open the dataset registration file
  T2_DSRegistrationFileReader dsreg(argv[++ac]);
  if (!dsreg) {
      beconfig.close();
      return -1;
  }

  // try to open the constructor catalog file
  T2_ConsCatalogReader conscat(argv[++ac]);
  if (!conscat) {
      beconfig.close();
      dsreg.close();
      return -1;
  }

  // try to open the dataset info file
  T2_DSInfoFileWriter dsinfo(argv[++ac]);
  if (!dsinfo) {
      beconfig.close();
      dsreg.close();
      conscat.close();
      return -1;
  }

  // try to open the dataset catalog file
  T2_DSCatalogWriter dscat;
  dscat.open(argv[++ac], beconfig);
  if (!dscat) {
      beconfig.close();
      dsreg.close();
      conscat.close();
      dsinfo.close();
      return -1;
  }

  // try to open the dataset catalog file
  T2_IndexCatalogWriter idxcat(argv[++ac]);
  if (!idxcat) {
      beconfig.close();
      dsreg.close();
      conscat.close();
      dsinfo.close();
      dscat.close();
      return -1;
  }

  // go through the dataset registration file and create a dataset info
  // file
  if (r == true)
     r = process_files(beconfig, dsreg, conscat, dsinfo, dscat, idxcat);
  beconfig.close();
  dsreg.close();
  conscat.close();
  dsinfo.close();
  dscat.close();
  idxcat.close();

  if (r == false) {
     // some thing wrong happens ...
     // assuming that dsinfo, dscat and idxcat would rename the original file
     // as filename.bak in case of name conflict, here I am just gonna remove 
     // the newly created files
     for (u_int k=4; k<7; k++) {
        if (unlink(argv[k]) != 0) {
	   ifstream f(argv[k]);
	   if (f)		// no complaint if argv[k] does not exist
	      t2_err_ret(__FILE__, __LINE__, "WARNING: cannot remove '%s'",
			 argv[k]);
	}
     }
  }

  return 0;
}


//////////////////////////////////////////////////////////////////////////
//
// process_files
//
//////////////////////////////////////////////////////////////////////////
bool
process_files(T2_BEConfigFile& beconfig, T2_DSRegistrationFileReader& dsreg,
	      T2_ConsCatalogReader& conscat, T2_DSInfoFileWriter& dsinfo,
	      T2_DSCatalogWriter& dscat, T2_IndexCatalogWriter& idxcat)
{
  t2_Assert (beconfig);
  t2_Assert (dsreg);
  t2_Assert (conscat);
  t2_Assert (dsinfo);
  t2_Assert (dscat);

  // remember dataset names to enforce uniqueness
  T2_VArray<T2_CharPtr> ds_names(20);

  T2_DSRegistrationEntry ds_e;
  T2_ConsCatalogEntry conscat_e, idx_conscat_e;
  T2_DSID ds_id;			// dataset id
  T2_Array<T2_IndexID> idx_ids;		// index id's
  T2_Array<u_int> idx_cons_ids;		// index constructor id's
  while (dsreg.getNextEntry(ds_e) == true) {
     t2_Assert (ds_e.dataset_name != NULL);
     t2_Assert (ds_e.dsconstructor_name != NULL);

     // check if entry content is valid
     if (ds_e.check_content_correct(beconfig.getMinInvalidLogicalDiskID())
         == false)
     {
	errno = 0;
	t2_err_ret(NULL, 0,
		"ERROR: invalid entry is printed below:");
	cerr << ds_e;
	return false;
     }

     // check if dataset name is unique
     if (search_name_add(ds_e.dataset_name, ds_names) > 0) {
	errno = 0;
	t2_err_ret(NULL, 0, "ERROR: dataset name '%s' is not a unique name",
		   ds_e.dataset_name);
	return false;
     }

     u_int nindexes = ds_e.indexes.getNumberElements();
     t2_Assert (nindexes > 0);
     idx_cons_ids.clear(nindexes);

     // search for dataset constructor
     if (conscat.fetchExactMatch(conscat_e, ds_e.dsconstructor_name,
				 T2_UDF_Dataset)
	== false)
     {
        errno = 0;
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot find dataset constructor '%s'",
		ds_e.dsconstructor_name);
	return false;
     }

     // find constructors for all indexes
     for(u_int j=0; j<nindexes; j++) {
	if (conscat.fetchExactMatch(idx_conscat_e,
				    ds_e.indexes[j].index_constructor_name,
				    T2_UDF_Index)
	   == false)
	{
	   errno = 0;
	   t2_err_ret(__FILE__, __LINE__,
		     "ERROR: cannot find index constructors '%s'",
		      ds_e.indexes[j].index_constructor_name);
	   return false;
	}
	idx_cons_ids[j] = idx_conscat_e.id;
     }

     // add the new dataset entry into dataset info file, and get
     // assigned id's for dataset and its indexes
     t2_Assert (ds_e.indexes.getNumberElements() > 0);
     idx_ids.clear(ds_e.indexes.getNumberElements());
     if (dsinfo.addEntry(ds_e, ds_id, idx_ids, conscat_e.iterators) == false) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot add dataset '%s' into dataset info file",
		ds_e.dataset_name);
	return false;
     }

     // add the new dataset entry into dataset catalog file
     if (dscat.addEntry(ds_e, ds_id, conscat_e.id) == false) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot add dataset '%s' into dataset catalog file",
		ds_e.dataset_name);
	return false;
     }

     // add the new index entries into index catalog file
     if (idxcat.addEntry(ds_e, idx_ids, idx_cons_ids) == false) {
	t2_err_ret(__FILE__, __LINE__,
	  "ERROR: cannot add indexes of dataset  '%s' into index catalog file",
	  ds_e.dataset_name);
	return false;
     }
  }			// while (dsreg.getNextEntry(ds_e) == true)

  return true;
}



//////////////////////////////////////////////////////////////////////////
//
// search_name_add
//
// Search for a target name from a list, and if not found, add it to
// the list.
// Return 0 if not found, 1 otherwise.
//
//////////////////////////////////////////////////////////////////////////
u_int
search_name_add(const char* target, T2_VArray<T2_CharPtr>& list)
{
  t2_Assert (target != NULL);

  for (u_int i=0; i<list.getNumberElements(); i++) {
     if (strcmp(target, list[i]) == 0)		// a match !!
	return 1;
  }

  // not found, so add to list
  char*& p = (char *&) list.appendElement();
  size_t len = strlen(target)+1;
  t2_Alloc(p, len, char);
  memcpy(p, target, len);
  return 0;
}

