
/* $Id: t2_custom.h,v 1.6 2000/03/14 22:55:12 chialin Exp $ */

/* t2_custom.h */

#ifndef _T2_CUSTOM_H_
#define _T2_CUSTOM_H_

#include	<iostream.h>
#include	<fstream.h>
#include	"t2_header.h"
#include	"t2_txtfile.h"
#include	"t2_tarray.h"
#include	"t2_charptr.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_UdfConsRegistrationEntry
//
// A class that contains an entry from the user-defined constructor file
// to be described below.
//
///////////////////////////////////////////////////////////////////////////
class T2_UdfConsRegistrationEntry {
public:
  T2_UdfConsRegistrationEntry();
  ~T2_UdfConsRegistrationEntry();

  ostream& write(ostream&) const;

  T2_UDFType type;			// constructor type
  char *name,				// unique constructor name
       *function,			// unique function name for the 
					// constructor (ie the one to be put
					// into constructors.C)
       *description;			// optional text description for the
					// constructor
  T2_VArray<T2_CharPtr> iterators;	// only valid for dataset
					// constructor functions
					// (ie type = T2_UDF_Dataset)

protected:
private:
};


inline
T2_UdfConsRegistrationEntry::T2_UdfConsRegistrationEntry()
			: type(T2_UDF_Unknown), name(NULL),
			  function(NULL), description(NULL), iterators(5)
{
}

inline
T2_UdfConsRegistrationEntry::~T2_UdfConsRegistrationEntry()
{
  delete [] name;
  delete [] function;
  delete [] description;
}

inline ostream&
T2_UdfConsRegistrationEntry::write(ostream& os) const
{
  os << "type: " << t2_UDFType2string(type) << endl
     << "name: " << name << endl
     << "function: " << function << endl;
  if (description == NULL)
     os << "description: NULL" << endl;
  else
     os << "description: " << description << endl;
  if (iterators.getNumberElements() == 0)
     os << "iterators: none" << endl;
  else {
     os << "iterators: ";
     for (u_int i=0; i<iterators.getNumberElements(); i++)
	os << " , " << t2_stringNullWrap(iterators[i]);
     os << endl;
  }
  return os;
}

inline ostream&
operator << (ostream& os, const T2_UdfConsRegistrationEntry& e)
{
  return e.write(os);
}


///////////////////////////////////////////////////////////////////////////
//
// T2_UdfConsRegistrationFile
//
// A class to access a user-defined constructor file, which is a file that 
// contains information about user-defined customization constructor
// functions, such as for the datasets, the indexes, the projection functions,
// the accumulator meta objects, the aggregation functions, and the
// MetaChaos functions.
//
// See comments in T2_UdfConsRegistrationFile::getNextEntry(int& eflag)
// for format of the file.
//
///////////////////////////////////////////////////////////////////////////
class T2_UdfConsRegistrationFile {
public:
  T2_UdfConsRegistrationFile(const char* = NULL);
  ~T2_UdfConsRegistrationFile() { close(); };

  // open and close file
  void open(const char* fn) { close(); if (fn != NULL) file.open(fn, true); };
  void close() { if (file) file.close(); };

  // query methods
  operator const void*() const { return (const void *) file; };

  // get next entry from file;
  // if eof, returns NULL and set eflag to 0
  // if error, returns NULL and set eflag to 1
  T2_UdfConsRegistrationEntry* getNextEntry(int& eflag);

  // return true if funcname is a valid C++ function name (eg doesn't
  // contain comma, semicolon, etc), false otherwise
  static bool check_function_name(const char* funcname);

protected:
  T2_TxtFile file;
  const char comment;		// char used as comment delimiter

private:
};



///////////////////////////////////////////////////////////////////////////
//
// T2_UdfConstructorFile
//
// A class to create a C++ file, which will be linked with the T2 back-end
// program. The file contains functions that would store all the
// user-defined constructors into several arrays of function pointers.
// See back-end/src/constructors.C as an example.
//
///////////////////////////////////////////////////////////////////////////
class T2_UdfConstructorFile {
public:
  T2_UdfConstructorFile(const char* f = NULL);
  ~T2_UdfConstructorFile() { close(); delete [] files; }

  // open and close file
  void open(const char*);
  void close();

  // query methods
  operator const void*() const { return (valid == true)? this:NULL; };

  // insert a new entry
  bool addEntry(T2_UDFType t, const char* funcname, u_int& id);

protected:
  bool valid;
  fstream* files;				// a set of temporary files
  T2_Array<u_int> counts;
  char tmpfname[T2_NUMBER_UDFTypes][256];	// filenames for temporary files

  // append file src to file dest, with total number of ids = n
  bool appendFile(fstream& dest, fstream& src, u_int n);

  // print prologue and epilogue into a file of type i
  void printPrologue(fstream& f, u_int i);
  void printEpilogue(fstream& f, u_int i);


private:
};



#endif	/* ! _T2_CUSTOM_H_ */
